/* 
 * \file        ComplexFiltersLiu.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * The implementation of ComplexFiltersLiu (Liu2010) class
 *
 * 
 */


#include "ComplexFiltersLiu.h"
#include "NewPoincareSingularPoints.h"

#include "MemUtil.hpp"
#include "ImgUtil.h"
#include "FingerprintImg.h"

#include "ParametersLiu.h"

ComplexFiltersLiu::ComplexFiltersLiu() {
    img = 0;
    h = -1;
    w = -1;
    ySin = 0;
    xCos = 0;
    O = 0;
    filteredAngleCore = 0;
    filteredMagnitudeCore = 0;
    filteredAngleDelta = 0;
    filteredMagnitudeDelta = 0;
    featureVector = 0;
    
    coherence = 0;
    blockMean = 0;
    blockVariance = 0;
    segmentation = 0;
}


void ComplexFiltersLiu::initialize(float **image, int h, int w) {
       
    img = MemUtil::copy2DMatrix<float>(image, h, w);
    
    this->h = h;
    this->w = w;
}

ComplexFiltersLiu::~ComplexFiltersLiu() {

    MemUtil::freeMatrix(&img);    
    MemUtil::freeMatrix(&O);    
    MemUtil::freeMatrix(&xCos);    
    MemUtil::freeMatrix(&ySin);    
    MemUtil::freeMatrix(&coherence);    
    MemUtil::freeMatrix(&blockMean);    
    MemUtil::freeMatrix(&blockVariance);    
    MemUtil::freeMatrix(&segmentation);    
    MemUtil::freeMatrix(&filteredMagnitudeCore);    
    MemUtil::freeMatrix(&filteredAngleCore);    
    MemUtil::freeMatrix(&filteredMagnitudeDelta);    
    MemUtil::freeMatrix(&filteredAngleDelta); 
    MemUtil::freeMatrix(&featureVector);
    
       
    for (int i = 0; i < nScales; i++)
        MemUtil::freeMatrix(&(complexFilter[i]));
    delete[] complexFilter; // no se puede memUtil porque intenta liberar todo
    
}


float *ComplexFiltersLiu::run(string file, PCASYSManager *pcasys, bool segmentar){
    FingerprintImg fi;
    
    if (ImgUtil::fAdjust != 1){
        
        fi.setPCASYSManager(pcasys);
        fi.readFingerprintImage(file);       
        fi.setSegmentedPCASYS(fi.getImage(), fi.getH(), fi.getW());

    }
    else{


        fi.setPCASYSManager(pcasys);        
        fi.readFingerprintImage(file);
        fi.segment();
    }
    
    initialize(fi.getSegmentedImageFloat(), fi.getH(), fi.getW());
    
    // se asigna a si mismo...
    featureVector = computeFeatureVector(segmentar);

    dimFV = getNScales() * 16;
    
    float *fvOut = MemUtil::createVector<float>(dimFV);
    for (int i = 0; i < getNScales(); i++) {
        for (int j = 0; j < 16; j++) {
            fvOut[i * 16 + j] = featureVector[i][j];
        }
    }
    cout << endl;

    MemUtil::freeMatrix(&featureVector);
    return fvOut;
}

float **ComplexFiltersLiu::computeFeatureVector(bool segmentationPoincare) {
    
    // AQUí va todo el proceso
    createOrientationImage(blockSizeLiu * ImgUtil::fAdjust); // block size = 16 x 16
    smoothOrientationImage(sMaxLiu, constThresholdLiu); // (4, 0.5)

    float *w = MemUtil::createVector<float>(3);
    w[0] = defCoherenceLiu; // coherence
    w[1] = defMeanLiu; // mean
    w[2] = defVarianceLiu; // variance
    createSegmentationImage(w);
    MemUtil::freeMatrix(&w);
    
    if (segmentationPoincare) {
        NewPoincareSingularPoints *npsp = new NewPoincareSingularPoints(img, this->h, this->w);
        npsp->setOrW(orW);
        npsp->setOrH(orH);
        npsp->generateSegmentation(img, this->h, this->w);
        npsp->combineWithSegmentation(this->segmentation);
        MemUtil::freeMatrix(&segmentation);
        segmentation = MemUtil::copy2DMatrix<bool>(npsp->getSegmentation(), orH, orW);
        
        delete npsp;
    }

    createFilters(orderKLiu, defNumScalesLiu);
    filterImage();
    vector< vector<int> > cores = findSingularPoints(filteredMagnitudeCore, maxWindowSizeLiu, multiScaleWLiu, coreThresholdLiu, true); // CORES
    vector< vector<int> > deltas = findSingularPoints(filteredMagnitudeDelta, maxWindowSizeLiu, multiScaleWLiu, deltaThresholdLiu, false); // multi-scale find window size 13 x 13 and threshold core and delta

    computeFeatures(cores, deltas);

    return featureVector;
}



void ComplexFiltersLiu::createOrientationImage(int blockW) {
    
    this->blockW = blockW;
    orH = h / blockW;
    orW = w / blockW;
    
    O = MemUtil::create2DMatrix<float>(orH,orW);

    xCos = MemUtil::create2DMatrix<float>(orH,orW);
    ySin = MemUtil::create2DMatrix<float>(orH,orW);
    float **Vx = MemUtil::create2DMatrix<float>(orH,orW);
    float **Vy = MemUtil::create2DMatrix<float>(orH,orW);
        
    coherence = MemUtil::create2DMatrix<float>(orH,orW);
    blockMean = MemUtil::create2DMatrix<float>(orH,orW);
    blockVariance = MemUtil::create2DMatrix<float>(orH,orW);   
           
    float **gradX = MemUtil::create2DMatrix<float>(h,w);
    float **gradY = MemUtil::create2DMatrix<float>(h,w);
    float **grad = MemUtil::create2DMatrix<float>(h,w);

    /* compute SOBEL */
    for (int i = 1; i < h - 1; i++) {
        for (int j = 1; j < w - 1; j++) {
            gradX[i][j] = (img[i + 1][j + 1] + 2.0 * img[i][j + 1] + img[i - 1][j + 1] - 
                   img[i + 1][j - 1] - 2.0 * img[i][j - 1] - img[i - 1][j - 1]) ;
            
            gradY[i][j] = -(img[i + 1][j - 1] + 2.0 * img[i + 1][j] + img[i + 1][j + 1] -
                    img[i - 1][j - 1] - 2.0 * img[i - 1][j] - img[i - 1][j + 1]) ;
            
            grad[i][j] = sqrt(gradX[i][j] * gradX[i][j] + gradY[i][j] * gradY[i][j]);
        }
    }
       
    float Gsx, Gsy;
    /* Compute Orientations */
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {
            
            for (int u = i * blockW; u < (i + 1) * blockW; u++) {
                for (int v = j * blockW; v < (j + 1) * blockW; v++) {
                    Gsx = 2.0 * gradX[u][v] * gradY[u][v];
                    Gsy = (gradX[u][v] * gradX[u][v] - gradY[u][v] * gradY[u][v]);
                    Vx[i][j] += Gsx;
                    Vy[i][j] += Gsy;
                    coherence[i][j] += sqrt(Gsx * Gsx + Gsy * Gsy);
                    blockMean[i][j] += img[u][v];
                }
            }
            // Coherence
            coherence[i][j] = sqrt(Vx[i][j] * Vx[i][j] + Vy[i][j] * Vy[i][j]) / coherence[i][j];
            blockMean[i][j] /= (float) (blockW * blockW);
            
            for (int u = i * blockW; u < (i + 1) * blockW; u++) {
                for (int v = j * blockW; v < (j + 1) * blockW; v++) {
                    blockVariance[i][j] += (blockMean[i][j] - img[u][v]) * (blockMean[i][j] - img[u][v]);
                }
            }
            blockVariance[i][j] /= (float) (blockW * blockW);
            
            float ang = 0.5 * atan2(Vx[i][j], Vy[i][j]);
            if (ang <= 0)
                O[i][j] = ang + 0.5 * PI;
            else
                O[i][j] = ang - 0.5 * PI;
                
            xCos[i][j] = cos(2.0 * O[i][j]) * sqrt(Vx[i][j] * Vx[i][j] + Vy[i][j] * Vy[i][j]);
            ySin[i][j] = sin(2.0 * O[i][j]) * sqrt(Vx[i][j] * Vx[i][j] + Vy[i][j] * Vy[i][j]);
        }
    }
    if (ImgUtil::DEBUG) {
        string fileName2 = "OrientationSinCos";
        ImgUtil::writeOrImg(xCos, ySin, orW, orH, fileName2);
    }
    MemUtil::freeMatrix(&Vx);
    MemUtil::freeMatrix(&Vy);
    
    MemUtil::freeMatrix(&gradX);
    MemUtil::freeMatrix(&gradY);
    MemUtil::freeMatrix(&grad);
}

void ComplexFiltersLiu::smoothOrientationImage(int sMax2, float consThreshold){
   
    
    float **xCosSmooth = MemUtil::create2DMatrix<float>(orH,orW);
    float **ySinSmooth = MemUtil::create2DMatrix<float>(orH,orW);
    float **Osmooth = MemUtil::create2DMatrix<float>(orH,orW);
    
    int i, j, s;
    float cons, consAnt;
    float val;
    for (i = 0; i < orH; i++) {
        for (j = 0; j < orW; j++) {
            s = 1;
            cons = consBlock(i, j, s, &val);
            do{
                s++;
                consAnt = cons;
                cons = consBlock(i, j, s, &val);    
            } while(((cons < consThreshold) || (cons < consAnt)) && (s <= sMax2));
            if (s > sMax2){
                s = 1;
                cons = consBlock(i, j, s, &val);
            }
            Osmooth[i][j] = val;
            xCosSmooth[i][j] = cos(2.0 * Osmooth[i][j]);
            ySinSmooth[i][j] = sin(2.0 * Osmooth[i][j]);
        }
    }
    
    if (ImgUtil::DEBUG) {
        string fileName2 = "OrientationSinCosSmooth";
        ImgUtil::writeOrImg(xCosSmooth, ySinSmooth, orW, orH, fileName2);
    }
    
   MemUtil::freeMatrix(&xCos);
   MemUtil::freeMatrix(&ySin);
   MemUtil::freeMatrix(&O);
   
   ySin = ySinSmooth;
   xCos = xCosSmooth;
   O = Osmooth;
}


float ComplexFiltersLiu::consBlock(int i, int j, int s, float *val){
    int n = 0;
    float sumCos = 0;
    float sumSin = 0;
    for (int x = max(i - s, 0); x < min(i + s + 1, orH); x++) {
        for (int y = max(j - s, 0); y < min(j + s + 1, orW); y++) {
            if ( (i != x) && (j != y)  ) {
                sumCos += xCos[x][y];
                sumSin += ySin[x][y];
                n++;
            }
        }
    }
    *val = 0.5 * atan2(sumSin, sumCos);
    return sqrt((sumCos * sumCos) + (sumSin * sumSin)) / (float) n;
}


void ComplexFiltersLiu::createSegmentationImage(float *w) {
    
    segmentation = MemUtil::create2DMatrix<bool>(orH,orW);
    //segmentation initialization
    for (int i = 0; i < orH; i++) 
        for (int j = 0; j < orW; j++) 
            if ((coherence[i][j] > w[0]) && (blockMean[i][j] < w[1]) && (blockVariance[i][j] > w[2]))
                segmentation[i][j] = true;
        
        
    if (ImgUtil::DEBUG) {
        string fileName = "segmentedLiuInitial";
        ImgUtil::writeGrayScaleImage(segmentation, orH, orW, fileName);
    }
    bool **segmentationDilatated = ImgUtil::imDilate(segmentation, orH, orW);
    MemUtil::freeMatrix(&segmentation);
    
    if (ImgUtil::DEBUG) {
        string fileName = "segmentedLiuDilated";
        ImgUtil::writeGrayScaleImage(segmentationDilatated, orH, orW, fileName);
    }
    segmentation = ImgUtil::imErode(segmentationDilatated, orH, orW);
    MemUtil::freeMatrix(&segmentationDilatated);
    
    if (ImgUtil::DEBUG) {
        string fileName = "segmentedLiu";
        ImgUtil::writeGrayScaleImage(segmentation, orH, orW, fileName);
    }
}



void ComplexFiltersLiu::createFilters(int m, int nScales) {
    
    float ang = 0;
    // a complex filter value of order m for each scale 1, 2, 3,..., nScales and value x, y
    complexFilter =  MemUtil::createVector<float**>(nScales); //new float**[nScales];
    for (int k = 0; k < nScales; k++) {
        int size = (k + 1) * 2 + 1;
        complexFilter[k] = MemUtil::create2DMatrix<float>(size, size);
//        complexFilter[k] = new float*[size];
//        for (int i = 0; i < size; i++) {
//            complexFilter[k][i] = new float[size];
//        }
    }
    
    for (int k = 0; k < nScales; k++) {
        int size = (k + 1) * 2 + 1;
        for (int i = 0; i < size; i++) {
            for (int j = 0; j < size; j++) {
                ang = atan2((float)j - (k + 1),(float)i - (k + 1));
                
                complexFilter[k][size - j - 1][i] = m * ang;
            }
        }
    }
    
    if (ImgUtil::DEBUG) {
        for (int k = 0; k < nScales; k++) {
            int size = (k + 1) * 2 + 1;
            for (int i = 0; i < size; i++) {
                for (int j = 0; j < size; j++) {
                    cout << complexFilter[k][i][j] *180./PI<< " ";
                }
                cout << endl;
            }
             cout << endl;
        }   
    }
    
    this->m = m;
    this->nScales = nScales;
    
}




void ComplexFiltersLiu::filterImage() {

    filteredMagnitudeCore = MemUtil::create3DMatrix<float>(nScales,orH,orW);
    filteredAngleCore = MemUtil::create3DMatrix<float>(nScales,orH,orW);
    filteredMagnitudeDelta = MemUtil::create3DMatrix<float>(nScales,orH,orW);
    filteredAngleDelta = MemUtil::create3DMatrix<float>(nScales,orH,orW);  
    
    int wSize = 0;
    float aCore, bCore;
    float aDelta, bDelta;
    float angCore = 0;
    float angDelta = 0;
    int p, q;
    for (int k = 0; k < nScales; k++) {
        wSize = (k + 1) * 2 + 1;
        aCore = 0;
        bCore = 0;
        aDelta = 0;
        bDelta = 0;
        for (int i = (k + 1); i < orH - (k + 1) ; i++) {
            for (int j = (k + 1); j < orW - (k + 1) ; j++) {
                    for (int u = i - (k + 1); u < i + (k + 1) + 1; u++) {
                        for (int v = j - (k + 1); v < j + (k + 1) + 1; v++) {

                            p = u - i + (k + 1);
                            q = v - j + (k + 1);

                            /* Cores */
                            angCore = 2. * O[u][v] - complexFilter[k][p][q];    
                            aCore += cos(angCore);
                            bCore += sin(angCore);
                            /* Deltas */
                            angDelta = 2. * O[u][v] + complexFilter[k][p][q];    
                            aDelta += cos(angDelta);
                            bDelta += sin(angDelta);
                        }
                    }
                    /* Cores */
                    aCore /= (float) (wSize * wSize);
                    bCore /= (float) (wSize * wSize);
                    /* Deltas */
                    aDelta /= (float) (wSize * wSize);
                    bDelta /= (float) (wSize * wSize);

                    /* Cores */
                    filteredMagnitudeCore[k][i][j] = sqrt(aCore * aCore + bCore * bCore);
                    filteredAngleCore[k][i][j] = atan2(bCore, aCore); 
                    /* Deltas */
                    filteredMagnitudeDelta[k][i][j] = sqrt(aDelta * aDelta + bDelta * bDelta);
                    filteredAngleDelta[k][i][j] = atan2(bDelta, aDelta); 
            }
        }        
    }
    
    if (ImgUtil::DEBUG) {
        
        for (int k = 0; k < nScales; k++) {
            string fileName2 = "filteredMagnitude";
            char str[20];
            sprintf(str, "%d", k);
            fileName2.append(str);
            ImgUtil::writeGrayScaleImage(filteredMagnitudeCore[k], orH, orW, fileName2, 255);
            fileName2.append("delta");
            ImgUtil::writeGrayScaleImage(filteredMagnitudeDelta[k], orH, orW, fileName2, 255);
        }
    }    
    
}


vector< vector<int> > ComplexFiltersLiu::findSingularPoints(float ***magnitude, int s, int multiScaleW1, float threshold, bool core) { 
    // multi-scale find window size 13 x 13 and threshold core and delta

    vector<int> localMaximums; //(10);
    // buscar máximos en la más gorda que es nScales - 1
    
    // solo podemos guardar dos localMaximums, los de mayor magnitud dentro de la huella
    
    bool uniqueCore = false;
    /* Comprobar que como máximo hay 2 SPs, eliminar por la segmentada y por la magnitud */
    float max1 = -1, max2 = -1;
    for (int i = s; i < orH - s; i++) {
        for (int j = s; j < orW - s; j++) {
            
            if ( (magnitude[nScales - 1][i][j] >= threshold) && (segmentation[i][j]) ) {
                bool notMax = false;
                for (int u = i - s; u < i + s + 1 && !notMax; u++) {
                    for (int v = j - s; v < j + s + 1 && !notMax; v++) {
                        if (magnitude[nScales - 1][u][v] > magnitude[nScales - 1][i][j]) {
                            notMax = true;
                            break;
                        }
                    }
                }
                if (!notMax) {
                    if (ImgUtil::DEBUG)
                        cout << "max in " << i << ", " << j << " magnitude = " << magnitude[nScales - 1][i][j] << endl;
                    
                    if  (max1 == -1) { 
                        max1 = magnitude[nScales - 1][i][j];
                        if(core)
                            uniqueCore = true;
                        localMaximums.push_back(i);
                        localMaximums.push_back(j);
                    }
                    else if (max2 == -1) {
                        max2 = magnitude[nScales - 1][i][j];
                        localMaximums.push_back(i);
                        localMaximums.push_back(j);
                        uniqueCore = false;
                        if (max2 > max1) {
                            float aux = max2;
                            max2 = max1;
                            max1 = aux;
                            int pos = localMaximums[0];
                            localMaximums[0] = localMaximums[2];
                            localMaximums[2] = pos;
                            
                            pos = localMaximums[1];
                            localMaximums[1] = localMaximums[3];
                            localMaximums[3] = pos;
                            
                        }
                    }
                    else {
                        float mag = magnitude[nScales - 1][i][j];
                        
                        if (mag > max1) {
                            max2 = max1;
                            max1 = mag;
                            localMaximums[2] = localMaximums[0];
                            localMaximums[3] = localMaximums[1];
                            localMaximums[0] = i;
                            localMaximums[1] = j;
                        }
                        else if (mag > max2) {
                            max2 = mag;
                            localMaximums[2] = i;
                            localMaximums[3] = j;
                        }
                    }
                }
            }
        }
    }
    
    /* si es core y hay 0 SPs, buscar el máximo y usarlo*/
    if (core && localMaximums.empty()) {
        // cout << "core = " << core << " localMax empty = " << localMaximums.empty() << endl;
        
        int iMax, jMax;
        iMax = 0;
        jMax = 0;
        for (int i = s; i < orH - s; i++) {
            for (int j = s; j < orW - s; j++) {
                if ( (magnitude[nScales - 1][i][j] > magnitude[nScales - 1][iMax][jMax])  && (segmentation[i][j])) {
                    iMax = i;
                    jMax = j;
                }
                
            }
        }
        
        localMaximums.push_back(iMax);
        localMaximums.push_back(jMax);
        uniqueCore = true;
        if (ImgUtil::DEBUG)
                cout << "absolute max in " << iMax << ", " << jMax << endl;
    }
    
    vector< vector<int> > singularPoints;
    while (!localMaximums.empty()) {
        vector<int> sp;
        int j = localMaximums.back();
        localMaximums.pop_back();
        int i = localMaximums.back();
        localMaximums.pop_back();
        sp.push_back(i);
        sp.push_back(j);
        for (int k = nScales - 2; k >= 0; k--) {
            int iMax = i;
            int jMax = j;
            for (int u = max(i - multiScaleW1, 0); u < min(i + multiScaleW1 + 1, orH); u++) {
                for (int v = max(j - multiScaleW1, 0); v < min(j + multiScaleW1 + 1, orW); v++) {
                    if (magnitude[k][u][v] > magnitude[k][iMax][jMax]) {
                        iMax = u;
                        jMax = v;
                    }

                }
            }
            i = iMax;
            j = jMax;
            
            //cout << "uniqueCore = "<<  uniqueCore << " mag = " << magnitude[k][i][j] << endl;
            if (!uniqueCore && (magnitude[k][i][j] < threshold) ) {
                sp.clear();
                break;
            }
            else {
                sp.push_back(i);
                sp.push_back(j);
            }
        }
        
        if (!sp.empty()) {
            bool spExists = false;
            for (int p = 0; p < singularPoints.size(); p++) {
                vector<int> sp2 = singularPoints.at(p);

                int i2 = sp2.at(sp2.size() - 1);
                int j2 = sp2.at(sp2.size() - 2);

                if ((i2 == i) && (j2 == j)) {
                    spExists = true;
                    break;
                }            
            }
            if (!spExists) {
                singularPoints.push_back(sp);
                if (ImgUtil::DEBUG)
                    cout << "hay un SP en " << i << ", " << j << " es core ? = " << core <<  endl;
            }
        }
        
    }
    
    if (core && singularPoints.empty()) {
        if (ImgUtil::DEBUG)
            cout << "necesito menor umbral para detectar 1 core " << endl;
        return findSingularPoints(magnitude, s, multiScaleWLiu, threshold - 0.1, core); // CORES
    }
    
    return singularPoints;
    
}

void ComplexFiltersLiu::computeFeatures(vector< vector<int> > cores, vector< vector<int> > deltas) {
    
    featureVector = MemUtil::create2DMatrix<float>(nScales,16);

    /* create for each scale a feature vector */
    /* order: core1 core2 delta1 delta2; 0's if they do not exist; core1 must always exist */

    /* Core 1 */
    vector<int> core1 = cores.back();
    cores.pop_back();

    /* Core 2 */
    vector<int> core2;
    if (!cores.empty()) {
        core2 = cores.back();
        cores.pop_back();
        
        if (core2.back() < core1.back()) { // compare y positions
            vector<int> aux = core1;
            core1 = core2;
            core2 = aux;
        }
    }
  
    /* Delta 1 */
    vector<int> delta1;
    if (!deltas.empty()) {
        delta1 = deltas.back();
        deltas.pop_back();
    }

    /* Delta 2 */
    vector<int> delta2;
    if (!deltas.empty()) {
        delta2 = deltas.back();
        deltas.pop_back();
        
        float d1 = sqrt(pow(delta1.back() - core1.back(), 2) + 
                pow(delta1.at(delta1.size() - 2) - core1.at(delta1.size() - 2), 2));
        float d2 = sqrt(pow(delta2.back() - core1.back(), 2) + 
                pow(delta2.at(delta2.size() - 2) - core1.at(delta1.size() - 2), 2));
        
        if (d2 < d1) {
            vector<int> aux = delta1;
            delta1 = delta2;
            delta2 = aux;
        }
        
    }
        
    for (int i = 0; i < nScales; i++) {
        // el back en el vector es la escala más fina!
        int y = core1.back();
        core1.pop_back();
        int x = core1.back();
        core1.pop_back();
        
        float ang = filteredAngleCore[i][x][y];
        float mag = filteredMagnitudeCore[i][x][y];
        //cout <<  "core 1 angle " << filteredAngleCore[i][x][y] * 180. / PI << endl;
        // cout << x << " " << y  << endl;
        
        featureVector[i][0] = y;
        featureVector[i][1] = x;
        featureVector[i][2] = ang;
        featureVector[i][3] = mag;
        int y2, x2;
        float rc, lineAng, beta, alfa;
        
        if (!core2.empty()) {
            y2= core2.back();
            core2.pop_back();
            x2 = core2.back();
            core2.pop_back();

            rc = sqrt( (x - x2) * (x - x2) + (y - y2) * (y - y2) );

            lineAng = atan2((x - x2), (y2 - y));

            beta = ang - lineAng;
            beta = (beta > 2. * PI) ? beta = beta - 2. * PI : beta;
            beta = (beta < -2. * PI) ? beta = beta + 2. * PI : beta;

       // cout <<  "core 2 angle " << filteredAngleCore[i][x2][y2] * 180. / PI << endl;
            alfa = filteredAngleCore[i][x2][y2] - ang;
            alfa = (alfa > 2. * PI) ? alfa = alfa - 2. * PI : alfa;
            alfa = (alfa < -2. * PI) ? alfa = alfa + 2. * PI : alfa;

            mag = filteredMagnitudeCore[i][x2][y2];

            featureVector[i][4] = rc;
            featureVector[i][5] = beta;
            featureVector[i][6] = alfa;
            featureVector[i][7] = mag;
        }
        
        if (!delta1.empty()) {
            y2 = delta1.back();
            delta1.pop_back();
            x2 = delta1.back();
            delta1.pop_back();
 //cout << x2 << " " << y2  << endl;
           // cout <<  "delta 1 angle " << filteredAngleDelta[i][x2][y2] * 180. / PI << endl;
            rc = sqrt( (x - x2) * (x - x2) + (y - y2) * (y - y2) );
            lineAng = atan2((x - x2), (y2 - y));

            beta = ang - lineAng;
            beta = (beta > 2. * PI) ? beta = beta - 2. * PI : beta;
            beta = (beta < -2. * PI) ? beta = beta + 2. * PI : beta; 
           // cout <<  "delta 1 lineAng " << lineAng * 180. / PI << endl;
           // cout <<  "delta 1 beta " << beta * 180. / PI << endl;

            alfa = filteredAngleDelta[i][x2][y2] - ang;
            alfa = (alfa > 2. * PI) ? alfa = alfa - 2. * PI : alfa;
            alfa = (alfa < -2. * PI) ? alfa = alfa + 2. * PI : alfa;

            
           // cout <<  "delta 1 alfa " << alfa * 180. / PI << endl;
            mag = filteredMagnitudeDelta[i][x2][y2];

            featureVector[i][8] = rc;
            featureVector[i][9] = beta;
            featureVector[i][10] = alfa;
            featureVector[i][11] = mag;
        
            if (!delta2.empty()) {
                y2 = delta2.back();
                delta2.pop_back();
                x2 = delta2.back();
                delta2.pop_back();

            //    cout <<  "delta 2 angle " << filteredAngleDelta[i][x2][y2] * 180. / PI << endl;
                rc = sqrt( (x - x2) * (x - x2) + (y - y2) * (y - y2) );
                lineAng = atan2((x - x2), (y2 - y));

                beta = ang - lineAng;
                beta = (beta > 2. * PI) ? beta = beta - 2. * PI : beta;
                beta = (beta < -2. * PI) ? beta = beta + 2. * PI : beta; 

                alfa = filteredAngleDelta[i][x2][y2] - ang;
                alfa = (alfa > 2. * PI) ? alfa = alfa - 2. * PI : alfa;
                alfa = (alfa < -2. * PI) ? alfa = alfa + 2. * PI : alfa;

                mag = filteredMagnitudeDelta[i][x2][y2];

                featureVector[i][12] = rc;
                featureVector[i][13] = beta;
                featureVector[i][14] = alfa;
                featureVector[i][15] = mag;
            }
        }
        
        
    }
    
}