/* 
 * \file        ComplexFiltersLiu.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * Header file for ComplexFiltersLiu (Liu2010) class
 *
 * 
 */


#ifndef COMPLEXFILTERSLIU_H
#define	COMPLEXFILTERSLIU_H


#include <vector>
#include "PCASYSManager.h"


using namespace std;


/**
 * \class ComplexFiltersLiu
 *
 * The ComplexFiltersLiu class implements Liu2010 algorithm based on complex
 * filters to obtain the singular points of a fingerprint image
 */
class ComplexFiltersLiu {
public:
    
    /** Default constructor */
    ComplexFiltersLiu();
    
    /**
     * It constructs an instance of the class to be processed with the following
     * parameters
     * \param image fingerprint image to extract the feature vector
     * \param h height of the image
     * \param w width of the image
     */
    void initialize(float **image, int h, int w);
    
    /** Default destructor */
    ~ComplexFiltersLiu();
    
    /**
     * Executes the Complex Filters Liu method from the fingerprint in file
     * \param file fingerprint
     * \param pcasys pcasys manager
     * \param segmentar indicates if Poincare's segmentation must be done
     * \return fingerprint fingerprint feature vector
     */
    float *run(string file, PCASYSManager *pcasys, bool segmentar = false);
    
    /**
     * Main method of the class. It processes the fingerprint to obtain the 
     * feature vector for classification purposes
     * \return the feature vector of size 16 x nScales
     */
    float **computeFeatureVector(bool segmentationPoincare = false);
    //TODO: parameterized the method with nScales, blockW, consThreshold
    
    /**
     * Computes the orientation map for the fingerprint image in blocks of size
     * blockW x blockW.It uses sobel's method for gradient computation
     * The orientation map is stored in O and xCos, ySin
     * \param blockW the size of each orientation block
     */
    void createOrientationImage(int blockW);
    
    /**
     * Smooths the obtained orientaiton map. It implements the smoothing algorithm in
     * Fingerprint Reference-Point Detection, EURASIP Journal on Applied 
     * Signal Processing 2005:4, 498–509
     * The smoothed orientation map is stored in O and xCos, ySin
     * \param sMax Maximum window size for the smoothing
     * \param consThreshold Minimum threshold to consider the smoothing as valid
     */
    void smoothOrientationImage(int sMax, float consThreshold);
    
    /**
     * Computes the Cons function for the smoothOrientationImage (this is the 
     * consistency of the smoothed orientation block)
     * \param i i-position of the block to be smoothed
     * \param j j-position of the block to be smoothed
     * \param s size of the neighborhood
     * \param val returning value of the smoothed orientation (in radians)
     * \return the consistency of the smoothed block in the s-neighborhood
     */
    float consBlock(int i, int j, int s, float *val);
    
    /**
     * Segments the fingerprint image using the variance, the mean and the 
     * coherence. It mostly implements the paper: Segmentation of Fingerprint 
     * Images, Asker M. Bazen and Sabih H. Gerez, ProRISC 2001 Workshop on 
     * Circuits, Systems and Signal Processing, Veldhoven, The Netherlands, 
     * November 2001.
     * \param w the thresholds used to segment the image (coherence, mean, variance)
     */
    void createSegmentationImage(float *w);
    
    /**
     * Creates the complex filters to be convolved with the orientation map
     * \param k order of the filter (with positive value, both + and - are created
     * \param nScales number of scales (filters) to be constructed
     */
    void createFilters(int k, int nScales); 
    
    /**
     * Filters the orientation map with the previously created complex filters.
     * It stores the magnitude responses and the angle of each block in each scale
     * for cores (positive order k) and deltas (negative order k).
     * Modified attributes: filteredMagnitudeCore, filteredMagnitudeDelta,
     * filteredAngleCore and filteredAngleDelta
     */
    void filterImage();
    
    /**
     * It finds the singular points from the magnitude responses of the filters
     * using the given threshold and the multi-scale approach.
     * \param magnitude Magnitude responses of the complex filtered image
     * \param s size of the window to find the local maximums (1 = 3x3 window)
     * \param multiScaleW window size for the multi-scale maximum search
     * \param threshold threshold to consider the filter response as singular point
     * \param core whether to find a core (or delta = false) 
     * \return the singular points found in each scale
     */    
    vector< vector<int> > findSingularPoints(float ***magnitude, int s, int multiScaleW, float threshold, bool core); // multi-scale find window size 13 x 13 and threshold core and delta
    
    /**
     * Computes the features for each core and delta in each scale
     * \param cores the detected cores (positions)
     * \param deltas the detected deltas (positions)
     */
    void computeFeatures(vector< vector<int> > cores, vector< vector<int> > deltas);

    /**
     * Returns the number os scales used in Liu method.
     * \return nScales
     */
    int getNScales();
    
    /**
     * It returns the feature vector's size
     * \return feature vector's size
     */
    int getDimFV() const;
    
private:
    float **img;        //!< Reference to the fingerprint image data
    int h;              //!< Height of the image
    int w;              //!< Width of the image
    
    int orH;            //!< Height of the orientation map
    int orW;            //!< Width of the orientation map
    int blockW;         //!< Block size used to compute the orientation map
    
    float **O;          //!< Orientation of each block (in radians)
    float **xCos;       //!< cos(2 O) value of the cosine of the orientation (x-direction)
    float **ySin;       //!< sin(2 O) value of the sine of the orientation (y-direction)
    
    float **coherence;          //!< Coherence value of each orientation block
    float **blockMean;          //!< Mean of each block
    float **blockVariance;      //!< Variance of each block
    bool **segmentation;        //!< Segmentation of the fingerprint image (whether it is foreground or background)
    
    float ***complexFilter;             //!< Complex filters to be used
    float ***filteredMagnitudeCore;     //!< Magnitude response of the filter for the core detection
    float ***filteredAngleCore;         //!< Angle detected for each block by the complex filter (core)
    float ***filteredMagnitudeDelta;    //!< Magnitude response of the filter for the delta detection
    float ***filteredAngleDelta;        //!< Angle detected for each block by the complex filter (delta)
    
    int m;                      //!< Order of the complex filter (positive, negative is automatically ocmputed)
    int nScales;                //!< Number of scales to be used in for the complex filtering
    
    int dimFV;                          //!< Dimension of feature vector
    
    float **featureVector;      //!< Feature vector computed from the cores and deltas detected by the complex filtering
};

inline int ComplexFiltersLiu::getDimFV() const { return dimFV;}
inline int ComplexFiltersLiu::getNScales() { return nScales;}
#endif	/* COMPLEXFILTERSLIU_H */

