/* 
 * \file        CovarianceBasedCoreDetector.h
 * \author      Carlos Lopez Molina (carlos.lopez@unavarra.es)
 * \version     1.0
 * 
 * \section     Description
 * 
 * Header file for CovarianceBasedCoreDetector class
 *
 * 
 */

#ifndef COVARIANCEBASEDCOREDETECTOR_H
#define	COVARIANCEBASEDCOREDETECTOR_H


#include "ParametersConvarianceBasedCoreDetector.h"

/**
 * @class CovarianceBasedCoreDetector
 *
 * The CovarianceBasedCoreDetector class implements the covariance-based fingerprint core detection, as explained in
 * 
 * A.K. Jain, S. Prabhakar, L. Hong
 * "A multichannel approach to fingerprint classification," 
 * IEEE Transactions on Pattern Analysis and Machine Intelligence, 
 * pp. 349-359, April, 1999
 * 
 */
class CovarianceBasedCoreDetector {
public:
    
    /**
     * Default constructor
     */
    CovarianceBasedCoreDetector();
    
    /**
     * Recommmended constructor
     * \param xCos  Horizontal (x) component of the orientations
     * \param ySin  Vertical (y) components of the orientations
     * \param h     Height of the orientations image
     * \param w     Width of the orientations image
     */
    CovarianceBasedCoreDetector(float **xCos, float **ySin, int h, int w);
    
    /**
     * Using the segmentation in the constructor
     * \param xCos  Horizontal (x) component of the orientations
     * \param ySin  Vertical (y) components of the orientations
     * \param segmentation Segmentation image
     * \param h     Height of the orientations image
     * \param w     Width of the orientations image
     */
    CovarianceBasedCoreDetector(float **xCos, float **ySin, bool **segmentation, int h, int w);
    
    /**
     * Destructor
     */
    virtual ~CovarianceBasedCoreDetector();

    /**
     * Computes the core of the fingerprint using the Covariance method.
     * \param marginSize Width of the banned strip around the image limits. Positions less than marginside tiles away from the limits are not considered.
     *              
     */
    void computeCore(int marginSize=12);
    
    /**
     * Computes the core of the fingerprint using the Covariance method.
     * \param marginSize Width of the banned strip around the image limits. Positions less than marginside tiles away from the limits are not considered.
     *                      The order is [N S E W].
     *              
     */
    void computeCore(int *fourMarginSizes);
    
    /**
     * Computes the max eigenvalue of the covariances matrix of each orientations neighbourhood
     * \param halfWindowSize    Integer part of half of the size of the neighbourhood size
     * \param marginSize Width of the banned strip around the image limits. Positions less than marginside tiles away from the limits are assigned 0 values. The order is [N S E W].
     */
    void createMaxEigenvalueMap(int halfWindowSize,int *marginSizes);
    
    /**
     * Thresholds the Eigenvalue map, selecting those positions with values at least normalizedPercentage times the maximum value in the map.
     * \param normalizedPercentage  Percentage of the max value necessary for being selected.
     */
    void binarizeMaxEigenvalueMap(float normalizedPercentage);
    
    /**
     * Assigns a numerical label to each connected region in the binarized eigenvalue map.
     */
    void labelCandidatesMap();
    
    /**
     * Assigns a numerical label to those positions connected to (posI,posJ). 8-connectitivy is considered.
     * \param posI      Vertical position
     * \param posJ      Horizontal position
     * \param label     Label the connected positions are to be assigned.
     */
    void fillRegion(int posI,int posJ,int label);
    
    /**
     * Calculates the centroid of each labelled region
     */
    void calculateRegionCentroids();
    
    /**
     * The position of the closest centroid to the center of the image
     */
    void selectBestCandidateRegion();

    /**
     * \return the core found the the aforementioned procedure.
     */
    float *getCore();
    
private:

    float **xCos;       //!< cos(2 O) value of the cosine of the orientation (x-direction)
    float **ySin;       //!< sin(2 O) value of the sine of the orientation (y-direction)
    
    int orH;            //!< Height of the orientation map
    int orW;            //!< Width of the orientation map
    
    float **maxEigValMap;
    int **binEigValMap;
    int **labelMap;
    
    float *candidateCore;
    
    int maxLabel;
    float **regionCentroids;
    
    bool **segmentation;
    
};

#endif	/* COVARIANCEBASEDCOREDETECTOR_H */

