/* 
 * \file        DirectionImagePartitioningCapelli.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * Header file for DirectionImagePartitioningCapelli (Capelli99) class
 *
 * 
 */

#ifndef DIRECTIONIMAGEPARTITIONINGCAPELLI_H
#define	DIRECTIONIMAGEPARTITIONINGCAPELLI_H

#include "DynamicMask.h"
#include "PCASYSManager.h"

/**
 * \class DirectionImagePartitioningCapelli
 *
 * The DirectionImagePartitioningCapelli class implements Cappelli1999
 * algorithm based the partitioning of the orientation map guided by
 * dynamic masks
 */
class DirectionImagePartitioningCapelli {
public:
    /* Default constructor */
    DirectionImagePartitioningCapelli();

    /**
     * Initializer of the class
     * \param xCos      The x-component of the orientation map
     * \param ySin      The y-component of the orientation map
     * \param orH       The height of the orientation map
     * \param orW       The width of the orientation map
     */
    void initialize(float **xCos, float **ySin, int orH, int orW);

    /* Default destructor */
    ~DirectionImagePartitioningCapelli();

    /**
     * It creates the left loop mask
     */
    void createLeftLoopMask();
    
    /**
     * It creates the left right mask
     */
    void createRightLoopMask();
    
    /**
     * It creates the arch mask
     */
    void createArchMask();
    
    /**
     * It creates the tented arch mask
     */
    void createTentedArchMask();

    /**
     * It creates the whorl mask
     */
    void createWhorlMask();

    /**
     * It optimizes the mask to the given orientation map
     * \return the normalized costs in the following order: A L R T W
     */
    float *fitMasks();

    /**
     * It executes all methods 
     * \param file fingerprint image
     * \param pcasys pcasys manager
     * \return feature vector
     */
    float *run(string file, PCASYSManager *pcasys);
    
    /**
     * It returns the feature vector's size
     * \return feature vector's size
     */
    int getDimFV() const;
    
private:

    DynamicMask *archMask;              //!< The arch mask
    DynamicMask *leftLoopMask;          //!< The left loop mask
    DynamicMask *rightLoopMask;         //!< The right loop mask
    DynamicMask *tentedArchMask;        //!< The tented arch mask
    DynamicMask *whorlMask;             //!< The whorl mask

    float **ySin;                       //!< The sine factor of the orientation map to be partitioned
    float **xCos;                       //!< The cosin factor of the orientation map to be partitioned

    int orH;                            //!< The height of the orientation map
    int orW;                            //!< The widht of the orientation map

    int dimFV;                          //!< Dimension of feature vector
};  

inline int DirectionImagePartitioningCapelli::getDimFV() const { return dimFV;}

#endif	/* DIRECTIONIMAGEPARTITIONINGCAPELLI_H */

