/* 
 * \file        DynamicMask.cpp
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * The implementation of DynamicMask (Capelli99) class
 *
 * 
 */

#include <iostream>
#include <cmath>

#include "MemUtil.hpp"
#include "DynamicMask.h"
#include "ParametersCappelli.h"

DynamicMask::DynamicMask() {
}


DynamicMask::DynamicMask(vector<MaskVertex *> vertices, vector<MaskRegion *> regions, vector<MaskRegionsRelation *> relations) {
    this->vertices = vertices;
    this->regions = regions;
    this->relations = relations;
    
    regionMap = MemUtil::create2DMatrix<int>(MaskVertex::getOrH(), MaskVertex::getOrW(), -1);
}

DynamicMask::~DynamicMask() {
    
    // HAY QUE BORRAR TODO LO QUE TIENEN DENTRO!!!!
    for (int i = 0; i < vertices.size(); i++)
        delete vertices[i];
    
    for (int i = 0; i < regions.size(); i++)
        delete regions[i];
    
    for (int i = 0; i < relations.size(); i++)
        delete relations[i];
    
    vertices.clear();
    regions.clear();
    relations.clear();
    
    MemUtil::freeMatrix(&regionMap);
    
}


void DynamicMask::alterGlobalRotationDisplacement(int deltaX, int deltaY, float ang) {
    
    for (int i = 0; i < vertices.size(); i++) 
        vertices[i]->alterGlobalRotationDisplacement(deltaX, deltaY, ang);
    
    for (int i = 0; i < vertices.size(); i++) 
        vertices[i]->correctVertexPosition(true);        
    
}

void DynamicMask::restoreGlobalRotationDisplacement() {
    
    alterGlobalRotationDisplacement(0, 0, 0);
}

int DynamicMask::regionOf(int x, int y) {
    
    int region = -1;
    int i = 0;

    i = 0;
    while ( (region == -1) && (i < regions.size()) ) {
        if (regions[i]->isInRegion(x, y))
            region = i;
        i++;
    }

    return region;
}

void DynamicMask::fillRegions() {
    
    int orH = MaskVertex::getOrH();
    int orW = MaskVertex::getOrW();
    
    int halfOrH = (MaskVertex::getOrH() - 1) / 2;
    int halfOrW = (MaskVertex::getOrW() - 1) / 2;
    // first pass
    for (int i = 0; i < orH; i++)
        for (int j = 0; j < orW; j++) 
            regionMap[i][j] = regionOf(j, (orH - 1) - i);
    
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {
            int y = ((orH - 1) - i);
            int x =  j;
            float u = x - halfOrW;
            float v = y - halfOrH;

            float xF, yF;
            xF = x;
            yF = y;
            float xIni = xF;
            float yIni = yF;
            float ang = atan2(v, u) + PI;
            if ((u == 0) && (v == 0))
                ang = PI / 2.;
            int n = 0;
            while ((regionMap[i][j] == -1)&& (n < 1000)) {

                xF += cos(ang);
                yF += sin(ang);

                x = round(xF);
                y = round(yF);
                n++;
                if ((n % (orW * orH / 2) == 0) || (x < 0) || (y < 0) || (x > orW - 1) || (y > orH - 1)) {
                    ang += PI / 12.;
                    xF = xIni;
                    yF = yIni;
                    x = round(xF);
                    y = round(yF);
                }
                
                if (regionMap[y][(orH - 1) - x] != -1)
                    regionMap[i][j] = regionMap[(orH - 1) -y][ x];
            //    cout << "new x,y = " << x << ", " << y << endl;
            // si la región es -1 cojo la recta al centro y relleno con la región más cercana.
            }
        }
    }
    
}

void DynamicMask::computeVariance(float **xCos, float **ySin) {
        
    int orH = MaskVertex::getOrH();
    int orW = MaskVertex::getOrW();
    int Ri;
    
    float *v2 = MemUtil::createVector<float>( regions.size() );
    float *vvx = MemUtil::createVector<float>(regions.size());
    float *vvy = MemUtil::createVector<float>(regions.size());
    float *sx = MemUtil::createVector<float>(regions.size());
    float *sy = MemUtil::createVector<float>(regions.size());
    
    
    fillRegions();
    
    float mod, mod2;
    
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {
            
            // Ri = regionOf(j, (orW - 1) - i);
            Ri = regionMap[i][j]; // = Ri;
            
            if (Ri != -1) {
                mod2 = xCos[i][j] * xCos[i][j] + ySin[i][j] * ySin[i][j];
                mod = sqrt(mod2);

                v2[Ri] += mod2;
                vvx[Ri] += mod * xCos[i][j];
                vvy[Ri] += mod * ySin[i][j];
                sx[Ri] += xCos[i][j];
                sy[Ri] += ySin[i][j];
            }
            else {
                cout << j << ", " << (orW - 1) - i << " no pertenece a ninguna region" << endl;
            }
        }
    }
    
    float variance = 0;
    float regionAverageO = 0;
    float aux;
    for (int i = 0; i < regions.size(); i++) {
        aux = sqrt(sx[i] * sx[i] + sy[i] * sy[i]);
        if (aux == 0) {
            variance = 0;
            regionAverageO = 0;
        }
        else {
            variance = 2. * (v2[i] - (sx[i] * vvx[i] + sy[i] * vvy[i]) / aux );
            regionAverageO = 0.5 * atan2(sy[i], sx[i]);
        }
        //cout << "Variance of region " << i << " = " << variance << " Average orientation = " << regionAverageO * 180. / PI << endl;
        regions[i]->setAverageO(regionAverageO);
        regions[i]->setVariance(variance);
    }
    
    MemUtil::freeMatrix(&v2);
    MemUtil::freeMatrix(&vvx);
    MemUtil::freeMatrix(&vvy);
    MemUtil::freeMatrix(&sx);
    MemUtil::freeMatrix(&sy);
    
    
}


float DynamicMask::computeCost(float **xCos, float **ySin) {
    
    float cost = 0;
    
    computeVariance(xCos, ySin);
    
    float firstTerm = 0;
    for (int i = 0; i < regions.size(); i++) {
        firstTerm += C0Capelli + regions[i]->getVariance();
    }
    
    float secondTerm = 0;
    MaskRegion *R1, *R2;
    float averageOR1, averageOR2;
    float deltaDir, deltaAbs;
    for (int i = 0; i < relations.size(); i++) {
        R1 = relations[i]->getFirstRegion();
        R2 = relations[i]->getSecondRegion();
        averageOR1 = R1->getAverageO();
        averageOR2 = R2->getAverageO();
        // compute dir differences, deltaDir
        deltaDir = averageOR1 - averageOR2;
        if (deltaDir < -PI / 2.)
            deltaDir += PI;
        else if (deltaDir > PI / 2.)
            deltaDir -= PI;
        
        // compute deltaAbs
        deltaAbs = abs(deltaDir - relations[i]->getRelationAngle());
        if (deltaAbs > PI / 2.)
            deltaAbs = abs( deltaAbs - PI );
        
        
        secondTerm += deltaAbs * 180. / PI;
    }
    
    cost = firstTerm + muCappelli / (float)relations.size() * secondTerm ;
    
    //cout << "Region cost = " << cost << " firstTerm = " << firstTerm << " secondTerm = " << secondTerm << endl;
    return cost;
    
}

float DynamicMask::findBestGlobalRotationDisplacement(float **xCos, float **ySin) {
    
    // Probar exhaustivamente todas y dejar la más pequeña    
    float cost;
    float minCost = 1000000;
    float deltaX, deltaY, ang;
    
    for (int i = displacementLowerBoundCappelli; i <= displacementUpperBoundCappelli; i++) {
        for (int j = displacementLowerBoundCappelli; j <= displacementUpperBoundCappelli; j++) {
            for (float a = rotationLowerBoundCappelli; a <= rotationUpperBoundCappelli; a += rotationStepCappelli) {
               //cout << "Probando... deltaX = " << i << " deltaY = " << j << " ang = " << a << endl;
                alterGlobalRotationDisplacement(i, j, a);
                
                cost = computeCost(xCos, ySin);
                if (cost < minCost) {
                    minCost = cost;
                    deltaX = i;
                    deltaY = j;
                    ang = a;                   
                }
                //cout << " coste = " << cost << endl;
            }
        }
    }
    alterGlobalRotationDisplacement(deltaX, deltaY, ang);
    if (ImgUtil::DEBUG)
        cout << "Mejor coste global = " << minCost << " obtenido con deltaX = " << deltaX << " deltaY = " << deltaY << " ang = " << ang * 180. / PI << endl;
    
    return minCost;
}

float DynamicMask::findBestLocalDisplacement(float **xCos, float **ySin) {
    
//    for (int i = 0; i < regions.size(); i++) {
//        cout << "Region " << i << endl;
//        regions[i]->print();
//    }
//    cout << endl;
    
    /* para cada vértice móvil encontrar la mejor displacement */
    int halfMaxDeltaX, halfMaxDeltaY;
    float minCost, cost; 
    int deltaX, deltaY;
    minCost = 1000000;
    for (int i = 0; i < vertices.size(); i++) {
        if (vertices[i]->isMobile()) {
            halfMaxDeltaX = vertices[i]->getMaxDeltaX() / 2;
            halfMaxDeltaY = vertices[i]->getMaxDeltaY() / 2;
            for (int u = -halfMaxDeltaX; u <= halfMaxDeltaX; u++) {
                for (int v = -halfMaxDeltaY; v <= halfMaxDeltaY; v++) {
                //cout << "Probando vertice " << i << "... deltaX = " << u << " deltaY = " << v;
                    vertices[i]->alterLocalDisplacement(u, v);
                    cost = computeCost(xCos, ySin);
                    if (cost < minCost) {
                        minCost = cost;
                        deltaX = u;
                        deltaY = v;
                    }
                    //cout << " coste = " << cost << endl;
                    
                }
            }
            if (ImgUtil::DEBUG)
                cout << "Mejor coste local vertice " << i << " = " << minCost << " obtenido con deltaX = " << deltaX << " deltaY = " << deltaY << endl;
            vertices[i]->alterLocalDisplacement(deltaX, deltaY);
        }
    }    
    return minCost;
    
}

float DynamicMask::optimizeMask(float **xCos, float **ySin) {
    
    float cost, costLocal;
    
    cost = findBestGlobalRotationDisplacement(xCos, ySin);
   
     costLocal = findBestLocalDisplacement(xCos, ySin);
    if (costLocal < cost)
        cost = costLocal;
    
    return cost;
}