/* 
 * \file        DynamicMask.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * Header file for DynamicMask (Capelli99) class
 *
 * 
 */


#ifndef DYNAMICMASK_H
#define	DYNAMICMASK_H

#include "MaskRegionsRelation.h"

/**
 * \class DynamicMask
 *
 * The DynamicMask class implements the structure of a dynamic
 * mask used in Cappelli1999's algorithm
 */
class DynamicMask {
public:
    /* Default constructor */
    DynamicMask();
    
    /**
     * It constructs a dynamic mask with the given vertices, regions and relations between regions
     * \param vertices  Vertices of the dynamic mask
     * \param regions   Regions formed by the vertices
     * \param relations Relations between regions of the mask
     */
    DynamicMask(vector<MaskVertex *> vertices, vector<MaskRegion *> regions, vector<MaskRegionsRelation *> relations);
    
    /* Default destructor */
    ~DynamicMask();
    
    /**
     * It modifies the global rotation-displacement of the dynamic mask
     * \param deltaX    Displacement in the x-coordinate
     * \param deltaY    Displacement in the y-coordinate
     * \param ang       Rotation angle (0 is upward direction)
     */
    void alterGlobalRotationDisplacement(int deltaX, int deltaY, float ang);
    
    /**
     * It restores the original global positioning of the dynamic mask
     */
    void restoreGlobalRotationDisplacement();
    
    /**
     * It returs the region to which the given point belongs
     * \param x The x-position of the point
     * \param y The y-position of the point
     * \return The region number to which the point belongs
     */
    int regionOf(int x, int y);
    
    /**
     * It fills the regionMap given the vertices, regions and 
     * their global/local displacements/rotations
     */
    void fillRegions();
    
    /**
     * It computes the variance of each region given the actual positioning 
     * (after fillRegions() is called)
     * \param xCos      The x-component of the orientation map
     * \param ySin      The y-component of the orientation map
     */
    void computeVariance(float **xCos, float **ySin); // devuelve array con las var de cada region
    
    /**
     * It computes the global cost of the actual partitioning
     * \param xCos      The x-component of the orientation map
     * \param ySin      The y-component of the orientation map
     * \return The costs of the mask for the given orientation map
     */
    float computeCost(float **xCos, float **ySin);
    
    /**
     * It finds the best global rotation-displacement for the given orientation map
     * \param xCos      The x-component of the orientation map
     * \param ySin      The y-component of the orientation map
     * \return The cost of the best partitioning
     */
    float findBestGlobalRotationDisplacement(float **xCos, float **ySin);
    
    /**
     * It finds the best local displacement of all the mobile vertices
     * of the mask for the given orientation map
     * \param xCos      The x-component of the orientation map
     * \param ySin      The y-component of the orientation map
     * \return The cost of the best partitioning
     */
    float findBestLocalDisplacement(float **xCos, float **ySin);

    /**
     * It optimizes the mask for the given orientation map. This consists in finding
     * the best global and local rotation/positioning of the mask/points
     * \param xCos      The x-component of the orientation map
     * \param ySin      The y-component of the orientation map
     * \return The cost of the best mask obtained
     */
    float optimizeMask(float **xCos, float **ySin);
    
    /**
     * It returns the number of regions in the mask
     * \return The number of regions in the mask
     */
    int getNRegions() const;
    
    /**
     * It returns the region map of the mask (which is filled by fillRegions())
     * \return The region map of the mask
     */
    int **getRegionMap() const;
    
private:
    vector<MaskVertex *> vertices;              //!< Vertices that composed the mask
    vector<MaskRegion *> regions;               //!< Regions that are formed by the given vertices
    vector<MaskRegionsRelation *> relations;    //!< The relations between the regions in the mask
    
    int **regionMap;                            //!< The map which stores the region to which each block belongs
};

inline int DynamicMask::getNRegions() const { return regions.size(); }

inline int **DynamicMask::getRegionMap() const { return regionMap; }

#endif	/* DYNAMICMASK_H */

