/* 
 * \file        FeedbackBasedLineDetector.cpp
 * \author      Jose Antonio Sanz Delgado (joseantonio.sazn@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * The implementation of FeedbackBasedLineDetector for Shah method
 *
 * 
 */

#include "MathUtil.h"
#include "ImgUtil.h"
#include "MemUtil.hpp"
#include "FingerprintImg.h"
#include "FeedbackBasedLineDetector.h"
#include "ParametersShah.h"
#include "Thinning.h"


FeedbackBasedLineDetector::FeedbackBasedLineDetector() {
    img = 0;
    h = -1;
    w = -1;
}

void FeedbackBasedLineDetector::initialize(float **image, int h, int w) {
    
    this->h = h;
    this->w = w;
    img = MemUtil::copy2DMatrix<float>(image, h, w);
    Ibin = MemUtil::create2DMatrix<float>(h, w);
    hist = MemUtil::createVector<float>(256);
    //S = MemUtil::create3DMatrix<float>(8, h, w);
    S = new float**[8];
    for (int i = 0; i < 8; i++)
        S[i] = MemUtil::create2DMatrix<float>(h, w);
    maximos = MemUtil::create3DMatrix<int>(8, h, w);
    Sfinal = MemUtil::create2DMatrix<float>(h, w);
    Lmean = MemUtil::create2DMatrix<float>(h, w);
    L = MemUtil::create3DMatrix<float>(8, h, w);
    OIk = MemUtil::create3DMatrix<float>(8, h, w);
    //E = MemUtil::create3DMatrix<float>(8, 7, 7);
    E = MemUtil::create3DMatrix<float>(8, MShah*2+1, MShah*2+1, -1.0);
    redImg = MemUtil::create2DMatrix<int>(h/16, w/16);
    F1 = MemUtil::createVector<float>(96);
    F2 = MemUtil::createVector<float>(128);
    dimFV = 96 + 128;
}

FeedbackBasedLineDetector::~FeedbackBasedLineDetector() {
    MemUtil::freeMatrix(&img); 
    MemUtil::freeMatrix(&Ibin); 
    MemUtil::freeMatrix(&hist);    
    for (int i = 0; i < 8; i++)
        MemUtil::freeMatrix(&S[i]);
    delete[] S;
    MemUtil::freeMatrix(&maximos);
    MemUtil::freeMatrix(&Sfinal);
    MemUtil::freeMatrix(&L);
    MemUtil::freeMatrix(&Lmean);
    MemUtil::freeMatrix(&E);
    MemUtil::freeMatrix(&OIk);
    MemUtil::freeMatrix(&redImg);
    MemUtil::freeMatrix(&F1);
    MemUtil::freeMatrix(&F2);
}

float *FeedbackBasedLineDetector::run(string file, PCASYSManager *pcasys) {
        FingerprintImg fi;
     
        if (ImgUtil::fAdjust != 1){  
            fi.setPCASYSManager(pcasys);    
            fi.readFingerprintImage(file);    
            fi.setSegmentedPCASYS(fi.getImage(), fi.getH(), fi.getW());
        }
        else{
            fi.setPCASYSManager(pcasys);
            fi.readFingerprintImage(file);
            fi.segment();
        }
        
        float *FV1 = MemUtil::createVector<float>(96);
        float *FV2 = MemUtil::createVector<float>(128);
        float *FV3 = MemUtil::createVector<float>(224);

        initialize(fi.getSegmentedImageFloat(), fi.getH(), fi.getW());
        binarizeImage();
        computeLineOutput();
        Thinning *thin = new Thinning(getH(), getW());
        thin->slimmerNSEW(getSfinal());
        thin->thinner2();
        thin->clean();
        thin->purge2();
        thin->cleanNoise();
        FV1 = computeF1(thin->getImg());
        for (int i = 0; i < 96 ; i++){
//            cout << FV1[i] << " ";
            FV3[i] = FV1[i];
        }
        FV2 = computeF2(thin->getImg());
        for (int i = 0; i < 128 ; i++){
//            cout << FV2[i] << " ";
            FV3[i+96] = FV2[i];
        }
        
        delete thin;
//        delete fi;
        
        cout << endl;
        return FV3;
}
    
    /** This function thresholds the initial image based on the histogram information of the whole image*/
    void FeedbackBasedLineDetector::imageThresholding(){
        float max1,max2=0.0; 
        float threshold;
        int pos1, pos2;
        float **auxCopy;
        
        /*We compute the histogram*/
        hist = ImgUtil::computeHistogram(img, h, w);
        /*We compute the maximum in the range [0, 127]*/
        for (int i = 0; i <= 127; i++){
            if (hist[i]>max1){
                max1 = hist[i];
                pos1 = i;
            }
        }
        /*We compute the maximum in the range [128, 255]*/
        for (int i = 128; i <= 255; i++){
            if (hist[i]>max2){
                max2 = hist[i];
                pos2 = i;
            }
        }
        
        /*The thrshold is the averge between both maximums*/
        threshold = (pos1+pos2)/2.0;
        
        if (ImgUtil::DEBUG)
                cout << "Primer maximo: " << max1 << "en posicion: " << pos1 << "Segundo maximo: " << max2 << "en posicion: " << pos2 << "Umbral: " << threshold << endl;
        
        /*We segment the image using the above computed threshold*/
        auxCopy = ImgUtil::thresholdImage(img, h, w, threshold);
        Ibin = MemUtil::copy2DMatrix(auxCopy, h, w);
        
        if (ImgUtil::DEBUG){
                ImgUtil::writeGrayScaleImage(Ibin, h, w, "LineDetectionInputImage", 255);
        }
        
        MemUtil::freeMatrix(&auxCopy);
    }
    
    /* This function segments the image based on the histogram information but using square blocks of size W*/
    void FeedbackBasedLineDetector::binarizeImage() {
        float max1, max2, threshold;
        int pos1, pos2, aux;
        int N;
        float mean;
        
        int wshah = WShah * ImgUtil::fAdjust;

        int halfW = wshah / 2;
        
        for (int i = 0; i < h; i += wshah) {
            for (int j = 0; j < w; j += halfW) {
                
                /*We compute both the histogram and the mean of the block*/
                mean = 0.0;
                N = 0;
                for (int u = i; u < min(i + wshah, h); u++) {
                    for (int v = j; v < min(j + wshah, w); v++) {
                        aux = img[u][v];
                        hist[aux]++;
                        mean += aux;                    
                        N++;
                    }
                }
                mean /= (float)N;
                
                
                /*We compute the maximum in the range [0, 127]*/
                max1 = 0.0;
                pos1 = 0;
                /*127 could be substitud by the mean just commenting the following line*/
                mean = 127;
                for (int ii = 0; ii <= mean; ii++){
                    if (hist[ii]>max1){
                        max1 = hist[ii];
                        pos1 = ii;
                    }
                    hist[ii] = 0.0;
                }
                /*We compute the maximum in the range [128, 255]*/
                max2 = 0.0;
                pos2 = mean;
                for (int ii = (mean+1); ii <= 255; ii++){
                    if (hist[ii]>max2){
                        max2 = hist[ii];
                        pos2 = ii;
                    }
                    hist[ii] = 0.0;
                }
                
                /*The threshold is the averge between both maximums*/
                threshold = (pos1+pos2)/2.0;
               
                for (int u = i; u < min(i + wshah, h); u++) 
                    for (int v = j; v < min(j + halfW, w); v++) 
                        Ibin[u][v] = img[u][v] > threshold ? 1 : 0;
            }
    }
    
    if (ImgUtil::DEBUG) {
        string fileName2 = "binarizedImg";
        ImgUtil::writeGrayScaleImage(Ibin, h, w, fileName2, 255);
    }
    
}
 
     int FeedbackBasedLineDetector::getH(){
         return h;
     }
    
    int FeedbackBasedLineDetector::getW(){
        return w;
    }
    
    float **FeedbackBasedLineDetector::getImg(){
        return Ibin;
    }
    
    float **FeedbackBasedLineDetector::getSfinal(){
        return Sfinal;
    }
    
    void FeedbackBasedLineDetector::computeLineOutput(){
        int it = 0;
        float aux, suma, max;
        float **auxCopy;
        int posMax;
        bool stop = true;
        
        float ***Saux = new float**[8];
        for (int i = 0; i < 8; i++)
            Saux[i] = MemUtil::create2DMatrix<float>(h, w);
        
       
        createEdgeMasks();/*We compute the bipolar edge orientation masks*/
        float sumaAnt = 1000000;
        while (stop) {
            /*We compute the feedbock for each pixel in the current iteration*/
            computeFeedback2(it, maximos);
            /*We compute the edge strength in each orientation*/
            for (int k = 0; k < 8; k++){
                for (int i = MShah; i < h-(MShah+1); i++)
                    for (int j = MShah; j < w-(MShah+1); j++){
                        if (Ibin[i][j] != 0) {
                            S[k][i][j] = 0.0;
                            S[k][i][j] = computeEdgeStrength(i, j, k);
                        }
                    }
            }
            
            /*We compute the line strength in each orientation*/
            for (int i = MShah; i < h-(MShah+1); i++)
               for (int j = MShah; j < w-(MShah+1); j++){
                       Lmean[i][j] = 0.0;
                       max = 0.0;
                       posMax = 0;
                       for (int k = 0; k < 8; k++){
                            L[k][i][j] = 0.0;
                            maximos[k][i][j] = 0;
                            aux = computeLineStrength(i, j, k);
                            L[k][i][j] = aux; 
                            Lmean[i][j] += aux;//To compute the average L
                            if (aux>max){
                                max = aux;
                                posMax = k;
                            }
                            maximos[k][i][j] = posMax;//the maximum value for speeding up the feedback computation
                       }
                       Lmean[i][j] = Lmean[i][j]/8.0;//Averaging L to seed up the feedback computation
               }

            
          /*We compute the differnce with the previous iteraration to check when the method estabilishes*/
          suma = 0.0;
          for (int k = 0; k < 8; k++){
                auxCopy = ImgUtil::thresholdImage(S[k], h, w, 0.0);
                MemUtil::freeMatrix(S[k]);
                S[k] = auxCopy;
                aux = MathUtil::computeMatrixDifferences<float>(S[k], Saux[k], h, w);
                suma += aux; 
                MemUtil::freeMatrix(Saux[k]);
                Saux[k] = MemUtil::copy2DMatrix<float>(S[k], h, w);
            }
            //cout << "La diferencia en la iteracion " << it << " es: " << suma << endl;
            
            /*We check whether the method has estabilished*/
            if ((abs(suma - sumaAnt) < 10000) || (suma > sumaAnt))
                stop = false;
            
            sumaAnt = suma;
            it += 1;
        }
        
        /*We generate the final output image*/
        for (int i = 0; i < h; i++){
            for (int j = 0; j < w; j++){
                Sfinal[i][j] = 0.0;
                for (int k = 0; k < 8; k++){
                    if ((S[k][i][j] != 0.0)&&(OIk[k][i][j]==1.0)){
                        Sfinal[i][j] = 1.0;
                        //Sfinal[i][j] = (k + 1) * 255. / 8; //to show the image in gray scale
                        break;
                    }
                }
            }
        }   
        
        if (ImgUtil::DEBUG){
            ImgUtil::writeGrayScaleImage(Sfinal, h, w, "LineDetectionFinalOutputImage", 255);
            /*In case of gray scale*/
            //ImgUtil::writeGrayScaleImage(Sfinal, h, w, "LineDetectionFinalOutputImage", 1);
        }
        
        for (int i = 0; i < 8; i++)
            MemUtil::freeMatrix(&Saux[i]);
        delete[] Saux;
    }
    
    
    
    float FeedbackBasedLineDetector::computeEdgeStrength(int posI, int posJ, int orientation){
        float edgeStrength;
        float feedback = 1.0;
        
        /*We compute the edge strength of a given pixel using its neighbourhood*/
        edgeStrength = 0.0;
        for (int i = (-1)*MShah; i <= MShah; i++){
            for (int j = (-1)*MShah; j <= MShah; j++){
                feedback = OIk[orientation][posI+i][posJ+j];
                edgeStrength += Ibin[posI+i][posJ+j]*E[orientation][MShah+i][MShah+j]*feedback;
            }
        }
        edgeStrength -= thresholdEdgeShah;
        
        if (edgeStrength<0.0){
            edgeStrength = 0.0;
        }

        return edgeStrength;
    }

    
    void FeedbackBasedLineDetector::computeFeedback2(int iteration, int ***maximos){
        int km = 0;//inicialized at zero because it is the first orientation
              
        if (iteration == 0){ // con un solo memset listo estaría!
            for (int i = 0; i < h; i++)
                for (int j = 0; j < w; j++)
                    for (int k = 0; k < 8; k++)
                        OIk[k][i][j] = 1.0;
        }else{
            /*We compute the orientation having the greatest line strength*/
            for (int i = 0; i < h; i++)
                for (int j = 0; j < w; j++){
                    km = maximos[7][i][j];
                    for (int k = 0; k < 8; k++){
                        OIk[k][i][j] = 0.0;
                        if ((L[km][i][j] - Lmean[i][j])>thresholdFeedbackShah){
                            if (L[km][i][j] == L[k][i][j])
                                OIk[k][i][j] = 1.0;
                            else
                                OIk[k][i][j] = 0.0;
                        }
                        else{
                            OIk[k][i][j] = 1.0;
                        }
                    }
                }
        }     
    }
    
    float FeedbackBasedLineDetector::computeLineStrength(int posI, int posJ, int orientation){
        float sum;
        
        /*We compute the line strength by summing the esge strength of the neighbourhood of the pixel*/
        sum = 0.0;
        for (int i = (-1)*MShah; i <= MShah; i++)
            for (int j = (-1)*MShah; j <= MShah; j++){
                sum += S[orientation][posI+i][posJ+j];
            }
        
        return sum;
    }

    
    void FeedbackBasedLineDetector::createEdgeMasks(){
        
        /*Zero degrees*/
        for (int i = 2; i < 5; i++)
            for (int j = 0; j < 7; j++)
                E[0][i][j] = 1.0;
        
        /*90 degrees*/
        for (int i = 0; i < 7; i++)
            for (int j = 2; j < 5; j++)
                E[4][i][j] = 1.0;
        
        /*22.5 degrees*/
        E[1][3][0] = 1.0;
        E[1][4][0] = 1.0;
        E[1][5][0] = 1.0;
        E[1][3][1] = 1.0;
        E[1][4][1] = 1.0;
        E[1][5][1] = 1.0;
        E[1][3][2] = 1.0;
        E[1][4][2] = 1.0;
        E[1][5][2] = 1.0;
        E[1][2][3] = 1.0;
        E[1][3][3] = 1.0;
        E[1][4][3] = 1.0;
        E[1][1][4] = 1.0;
        E[1][2][4] = 1.0;
        E[1][3][4] = 1.0;
        E[1][1][5] = 1.0;
        E[1][2][5] = 1.0;
        E[1][3][5] = 1.0;
        E[1][1][6] = 1.0;
        E[1][2][6] = 1.0;
        E[1][3][6] = 1.0;
        
        /*45 degrees*/
        E[2][5][0] = 1.0;
        E[2][6][0] = 1.0;
        E[2][3][1] = 1.0;
        E[2][4][1] = 1.0;
        E[2][5][1] = 1.0;
        E[2][6][1] = 1.0;
        E[2][3][2] = 1.0;
        E[2][4][2] = 1.0;
        E[2][5][2] = 1.0;
        E[2][2][3] = 1.0;
        E[2][3][3] = 1.0;
        E[2][4][3] = 1.0;
        E[2][1][4] = 1.0;
        E[2][2][4] = 1.0;
        E[2][3][4] = 1.0;
        E[2][0][5] = 1.0;
        E[2][1][5] = 1.0;
        E[2][2][5] = 1.0;
        E[2][3][5] = 1.0;
        E[2][0][6] = 1.0;
        E[2][1][6] = 1.0;
        
        /*67.5 degrees*/
        E[3][4][1] = 1.0;
        E[3][5][1] = 1.0;
        E[3][6][1] = 1.0;
        E[3][3][2] = 1.0;
        E[3][4][2] = 1.0;
        E[3][5][2] = 1.0;
        E[3][6][2] = 1.0;
        E[3][0][3] = 1.0;
        E[3][1][3] = 1.0;
        E[3][2][3] = 1.0;
        E[3][3][3] = 1.0;
        E[3][4][3] = 1.0;
        E[3][5][3] = 1.0;
        E[3][6][3] = 1.0;
        E[3][0][4] = 1.0;
        E[3][1][4] = 1.0;
        E[3][2][4] = 1.0;
        E[3][3][4] = 1.0;
        E[3][0][5] = 1.0;
        E[3][1][5] = 1.0;
        E[3][2][5] = 1.0;
        
        /*112.5 degrees*/
        E[5][0][1] = 1.0;
        E[5][1][1] = 1.0;
        E[5][2][1] = 1.0;
        E[5][0][2] = 1.0;
        E[5][1][2] = 1.0;
        E[5][2][2] = 1.0;
        E[5][3][2] = 1.0;
        E[5][0][3] = 1.0;
        E[5][1][3] = 1.0;
        E[5][2][3] = 1.0;
        E[5][3][3] = 1.0;
        E[5][4][3] = 1.0;
        E[5][5][3] = 1.0;
        E[5][6][3] = 1.0;
        E[5][3][4] = 1.0;
        E[5][4][4] = 1.0;
        E[5][5][4] = 1.0;
        E[5][6][4] = 1.0;
        E[5][4][5] = 1.0;
        E[5][5][5] = 1.0;
        E[5][6][5] = 1.0;
        
        /*135 degrees*/
        E[6][0][0] = 1.0;
        E[6][1][0] = 1.0;
        E[6][0][1] = 1.0;
        E[6][1][1] = 1.0;
        E[6][2][1] = 1.0;
        E[6][3][1] = 1.0;
        E[6][1][2] = 1.0;
        E[6][2][2] = 1.0;
        E[6][3][2] = 1.0;
        E[6][2][3] = 1.0;
        E[6][3][3] = 1.0;
        E[6][4][3] = 1.0;
        E[6][3][4] = 1.0;
        E[6][4][4] = 1.0;
        E[6][5][4] = 1.0;
        E[6][3][5] = 1.0;
        E[6][4][5] = 1.0;
        E[6][5][5] = 1.0;
        E[6][6][5] = 1.0;
        E[6][5][6] = 1.0;
        E[6][6][6] = 1.0;
        
        /*167.5 degrees*/
        E[7][1][0] = 1.0;
        E[7][2][0] = 1.0;
        E[7][3][0] = 1.0;
        E[7][1][1] = 1.0;
        E[7][2][1] = 1.0;
        E[7][3][1] = 1.0;
        E[7][1][2] = 1.0;
        E[7][2][2] = 1.0;
        E[7][3][2] = 1.0;
        E[7][2][3] = 1.0;
        E[7][3][3] = 1.0;
        E[7][4][3] = 1.0;
        E[7][3][4] = 1.0;
        E[7][4][4] = 1.0;
        E[7][5][4] = 1.0;
        E[7][3][5] = 1.0;
        E[7][4][5] = 1.0;
        E[7][5][5] = 1.0;
        E[7][3][6] = 1.0;
        E[7][4][6] = 1.0;
        E[7][5][6] = 1.0;
        
//        if (ImgUtil::DEBUG){
//            MemUtil::showMatrix(E,8,7,7);
//        }
        
    }
    
    float *FeedbackBasedLineDetector::computeF1(float **img){
        int cuadranteI, cuadranteJ, posAbsI, posAbsJ, auxPos, octante, zone, firstMax, secMax, numFirst, numSec;
        float rPoint;
        
        int nRow = h/16;//number of blocks (height)
        int nCol = w/16;//number of blocks (width)
        

        //we compute the coarse approximation of the orientation information
        float ***sum = MemUtil::create3DMatrix<float>(8, nRow, nCol);
        //the orientation strengths are summed for each zone
        for (int i = 0; i < h; i++){
            for (int j = 0; j < w; j++){
                //if the pixel is white
                if (img[i][j] == 1){
                    for (int k = 0; k < 8; k++){
                        //if the orientation is one we sum one to this orientation in the corresponding zone
                        if (OIk[k][i][j] == 1.0){
                            sum[k][i/16][j/16] += 1.0;
                        }
                    }
                }
            }
        }

        //we compute the orientation with the greatest strength for each zone
        for (int i = 0; i < nRow; i++){
            for (int j = 0; j < nCol; j++){
                firstMax = 0;
                numFirst = 0;
                for (int k = 0; k < 8; k++){
                    if (sum[k][i][j]>numFirst){
                        numFirst = sum[k][i][j];
                        firstMax = k;
                    }
                }
                redImg[i][j] = firstMax;
            }
        }
        MemUtil::freeMatrix(&sum);

        //MemUtil::showMatrix(redImg, nRow, nCol);
        
        float **contador = MemUtil::create2DMatrix<float>(24, 8);
        float *totales = MemUtil::createVector<float>(24);
        //we compute the histogram of orientations in each of the 24 zones
        for (int i = 0; i < nRow; i++){
            for (int j = 0; j < nCol; j++){
                cuadranteI = i/(nRow/2);// nRow/2 because the center is the center of the image
                cuadranteJ = j/(nCol/2);// nCol/2 because the center is the center of the image
                //absolute position with respect to the center of the image
                posAbsI = abs(i-(nRow/2));
                posAbsJ = abs(j-(nCol/2));
                rPoint = sqrt(pow(posAbsI, 2) + pow(posAbsJ, 2));//distance to the center of the image
                //we check the zone of the pixel according to the radius
                if (rPoint<6)
                    auxPos = 0;
                else if (rPoint>12)
                    auxPos = 2;
                else
                    auxPos = 1;
                // we chech the zone of the pixel according to the eigth zones defined without the radius
                if ((cuadranteI == 0)&&(cuadranteJ == 0))
                    if (posAbsI<posAbsJ)
                        octante = 0;
                    else
                        octante = 1;
                else if ((cuadranteI == 0)&&(cuadranteJ == 1))
                    if (posAbsI<posAbsJ)
                        octante = 3;
                    else
                        octante = 2;
                else if ((cuadranteI == 1)&&(cuadranteJ == 0))
                    if (posAbsI<posAbsJ)
                        octante = 7;
                    else
                        octante = 6;
                else
                    if (posAbsI<posAbsJ)
                        octante = 4;
                    else
                        octante = 5;
                //the zone of the pixel
                zone = octante*3+auxPos;
                //we update the histogram of orientation for the specified zone
                contador[zone][redImg[i][j]] += 1;
                totales[zone] += 1;
            }
        }

        //we compute the two most frequent orientations and their percentage of occurrence
        for (int i = 0; i < 24; i++){
            firstMax = -1;
            secMax = -1;
            numFirst = -1;
            numSec = -1;
            for (int j = 0; j < 8; j++){
                if (contador[i][j]>numFirst){
                    numSec = numFirst;
                    secMax = firstMax;
                    numFirst = contador[i][j];
                    firstMax = j;
                }else if (contador[i][j]>numSec){
                    numSec = contador[i][j];
                    secMax = j;
                }
            }
            //We assign four position of the feature vector: the two most frequent orientations 
            //    with their corresponding pocentage of occurrence
            if (totales[i] == 0){
                F1[i*4] = 0;
                F1[i*4+1] = 0;
                F1[i*4+2] = 0;
                F1[i*4+3] = 0;
            }
            else{
                F1[i*4] = firstMax;
                F1[i*4+1] = numFirst/totales[i];
                F1[i*4+2] = secMax;
                F1[i*4+3] = numSec/totales[i];
            }
        }
        
        MemUtil::freeMatrix(&contador);
        MemUtil::freeMatrix(&totales);
        
        //cout << "Vector F1: " << endl;
        //MemUtil::showMatrix(F1, 96);
//        for (int i = 0; i < 96 ; i++){
//            cout << F1[i] << " ";
//        }
        return F1;

    }

float *FeedbackBasedLineDetector::computeF2(float **img){
    float entr, maxEntropy, posI, posJ, numFirst, numSec;
    int cont, firstMax, secMax;
    
    //We compute the 120x120 window with the greattest entropy
    maxEntropy = 0.0;
    posI = 0;
    posJ = 0;
    for (int i = 0; i < (h-120); i+=5){
        for (int j = 0; j < (w -120); j+=5){
            entr = computeEntropy(img, i, j);
            if (entr > maxEntropy){
                maxEntropy = entr;
                posI = i;
                posJ = j;
            } 
        }
    }

    //This is to write the window with the greatest entropy and its 15x15 blocks
    float **aux = MemUtil::copy2DMatrix(img, h, w);
    for (int i = posI; i <= posI+120; i += 15){
        for (int j = posJ; j <= posJ+120; j++){
            aux[i][j] = 1.0;
        }
    }
    for (int j = posJ; j <= posJ+120; j += 15){
        for (int i = posI; i <= posI+120; i++){
            aux[i][j] = 1.0;
        }
    }
    if (ImgUtil::DEBUG){
        ImgUtil::writeGrayScaleImage(aux, h, w, "VentanaMaxEntropia", 255);
    }
    MemUtil::freeMatrix(&aux);

    //We compute the histogram of occurrences
    float ***histogram = MemUtil::create3DMatrix<float>(8, 8, 8);
    int ii, jj;
    for (int i = posI; i < posI+120; i++){
        for (int j = posJ; j < posJ+120; j++){
            if (img[i][j] == 1){
                ii = (i-posI)/15;
                jj = (j-posJ)/15;
                for (int k = 0; k < 8; k++){
                    if (OIk[k][i][j] == 1.0){
                        histogram[k][ii][jj] += 1;
                    }
                }
            }
        }
    }

    cont = 0;
    for (int i = 0; i < 8; i++){
        for (int j = 0; j < 8; j++){
            firstMax = -1;
            secMax = -1;
            numFirst = -1;
            numSec = -1;
            for (int k = 0; k < 8; k++){
                if (histogram[k][i][j]>numFirst){
                    numSec = numFirst;
                    secMax = firstMax;
                    numFirst = histogram[k][i][j];
                    firstMax = k;
                }else if (histogram[k][i][j]>numSec){
                    numSec = histogram[k][i][j];
                    secMax = k;
                }
            }
            //We assign two positions of the feature vector: the two most frequent orientations
            F2[cont*2] = firstMax;
            F2[cont*2+1] = secMax;
            cont++;
        }
    }

    MemUtil::freeMatrix(&histogram);

    //cout << "Vector F2: " << endl;
    //MemUtil::showMatrix(F2, 128);
    return F2;

}

float FeedbackBasedLineDetector::computeEntropy(float **img, int posI, int posJ){
    float total = 0.0;
    float total2 = 0.0;
    float pi;
    float entropy = 0.0;
    float cont = 0.0;
    
    
    float *numRidgeOrientation2 = MemUtil::createVector<float>(8);
    //We compute the number of occurrences of each orientation
    for (int i = posI; i < posI+120; i++){
        for (int j = posJ; j < posJ+120; j++){
            if (img[i][j] == 1.0){
                total = 0.0;
                for (int k = 0; k < 8; k++){
                    if (OIk[k][i][j] == 1.0){
                          //we compute the number of active orientations for the pixel
                          total++;
                    }
                }
                //we check how many pixels have a unique orientation
                if (total == 1.0){
                    cont++;
                }
                numRidgeOrientation2[maximos[7][i][j]]++;
                total2++;
            }
        }
    }
    //if we have enough edge pixels we compute the entropy, in other case we panlize it by assigning zero entropy
    if (total2 > 120.0){
        //we compute the entropy of the window
        for (int i = 0; i < 8; i++){
            pi = numRidgeOrientation2[i]/total2;
            entropy += pi*(1-pi);
        }
        //we weight the entropy by the number of pixels having a unique active orientation
        entropy *= (cont/total2);
    }
    
    MemUtil::freeMatrix(&numRidgeOrientation2);
    
    return entropy;
}
