/* 
 * \file        FeedbackBasedLineDetector.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \author      Jose Antonio Sanz Delgado (joseantonio.sanz@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * Header file for FeedbackBasedLineDetector class
 *
 * 
 */


#ifndef FEEDBACKBASEDLINEDETECTOR_H
#define	FEEDBACKBASEDLINEDETECTOR_H

#include "PCASYSManager.h"

class FeedbackBasedLineDetector {
public:
    
    /** Default constructor */
    FeedbackBasedLineDetector();
    
    /**
     * It initializes the class to be processed with the following
     * parameters
     * \param image fingerprint image to extract the feature vector
     * \param h height of the image
     * \param w width of the image
     */
    void initialize(float **image, int h, int w);
    
    /** Default destructor */
    virtual ~FeedbackBasedLineDetector();

    /**
     * Executes the Shah and Sastry method from the fingerprint in file
     * \param file fingerprint
     * \param pcasys pcasys manager
     * \returns the feature vector
     */
    float *run(string file, PCASYSManager *pcasys);
    
    /**
     * It returns the feature vector's size
     * \return feature vector's size
     */
    int getDimFV() const;
    
    /** 
     * This function thresholds the intial image based on the histogram information of the whole image
     */
    void imageThresholding();
    /** 
     * This function thresholds the intial image based on the histogram information using blocks of size W
     */
    void binarizeImage();
    
    /** 
     * This function performs the line detection process
     */
    void computeLineOutput();
    
    /*
     * This function computes the edge strength at position [posX, posY] with certain orientation
     * \param posX Position in the height component
     * \param posY Position in the width component
     * \param orientation It specifies the orientation to compute the edge strength
     * \return The edge strength at position [posX, posY] in the specified orientation
     */
    float computeEdgeStrength(int posX, int posY, int orientation);
    
    /*
     * This function computes the feedback of every pixel
     * \param iteration The current iteration of the algorithm
     * \param maximos It specifies the maximum edge strength in the 7th layer
     * \return The feedback for every pixel
     */
    void computeFeedback2(int iteration, int ***maximos);
    
    /*
     * This function computes the line strength at position [posX, posY] with certain orientation
     * \param posX Position in the height component
     * \param posY Position in the width component
     * \param orientation It specifies the orientation to compute the edge strength
     * \return The edge strength at position [posX, posY] in the specified orientation
     */
    float computeLineStrength(int posX, int posY, int orientation);
    
    /*
     * This function returns the height of the image
     */
    int getH();
    
    /*
     * This function returns the width of the image
     */
    int getW();
    
    /*
     * This function returns the binarized image which is the input of the process
     */
    float **getImg();
    
    /*
     * This function returns the final image after the line detection process
     */
    float **getSfinal();
    
    /*
     * This function creates the bipolar oriented edge masks
     */
    void createEdgeMasks();
    
    /*
     * This function compoutes the feature vector following the first approach
     * \param img The line detection image after the thinning process
     * \return This function returns de feature vector F1
     */
    float *computeF1(float **img);
    
    /*
     * This function compoutes the feature vector following the second approach
     * \param img The line detection image after the thinning process
     * \return This function returns de feature vector F2
     */
    float *computeF2(float **img);
    
private:
    
    float **img;        //!< Reference to the fingerprint image data
    int h;              //!< Height of the image
    int w;              //!< Width of the image
    float **Ibin;        //!< Initial thresholded image
    float *hist;         //!< Histogram of the fingerprint image
    float ***S;   //!<  The edge strength matrix
    float **Sfinal;   //!<  The final output of the line detection method
    float ***L;   //!<  The line strength matrix
    float ***E;   //!<  The orientation mask
    float **Lmean;  //!< The average L in the orientation direction
    float ***OIk;  //!< OIk: the feedback of the line detection process
    int **redImg;        //!< Reduced image (16x16 blocks) to compute the F1 vector
    int ***maximos;    //!< matrix to store the orientation having the maximum value of S
    float *F1;    // The feature vector following the first method (centered in the image)
    float *F2;    // The feature vector following the second method (contered in the true center of the fingerprint)
    int dimFV;                          //!< Dimension of feature vector

    /*
     * This function computes the entropy of a portion of an image
     * \para img The image used to compute its entropy
     * \param posI The starting point (height) to obtain the portion of the image
     * \param posJ The starting point (width) to obtain the portion of the image
     * \return This function returns the value of the entropy in the portion of the image
     */
    float computeEntropy(float **img, int posI, int posJ);
    
};

inline int FeedbackBasedLineDetector::getDimFV() const { return dimFV;}

#endif	/* FEEDBACKBASEDLINEDETECTOR_H */

