/* 
 * \file        FingerprintImg.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * The implementation of FingerprintImg (Fingerprint Image) class
 *
 * 
 */

#include "ImgUtil.h"
#include "MemUtil.hpp"
#include "FingerprintImg.h"

#include "FingerCode.h"
#include "GaborFilter.h"
#include "AdaptingImageSize.h"
#include "GuardarRealce.h"
#include "ImageEnhancementHong.h"


#include "ParametersLeung.h"


/** Default constructor */
FingerprintImg::FingerprintImg() {
    w = 0;
    h = 0;
    wOr = 0;
    hOr = 0;
    wTf = 0;
    wTf = 0;
    wOr2 = 0;
    hOr2 = 0;
    sfgw = 0;
    sfgh = 0;
    has_conup = false;
    
    segmentedPCASYS = 0;
    segmentedFgPCASYS = 0;
    enhancedPCASYS = 0;
    pixelOrPCASYS = 0;
    avOrX = 0;
    avOrY = 0;
    avOrRegX = 0;
    avOrRegY = 0;
    avOrX2 = 0;
    avOrY2 = 0;
    
    featVect = 0;
    
    image = 0;
    enhancedImage = 0;
    
    coreX = -1;
    coreY = -1;
    
    pcasys = 0;
    
    fingercode = 0;
    numSectors = -1;
    nFilters = -1;
    segmentedImage = 0;
    
    enhanceHong = false;
}

/** Default destructor */
FingerprintImg::~FingerprintImg() {
    if (ImgUtil::DEBUG)
        cout << "FingerprintImg: Destructing the fingerprint..." << endl;
    MemUtil::freeMatrix(&image);    
    MemUtil::freeMatrix(&enhancedImage);
    
    
    MemUtil::freeMatrix(&segmentedPCASYS);
    
    if (segmentedFgPCASYS != 0) {
        for (int i = 0; i < sfgh; i++) {
                delete[] (segmentedFgPCASYS[i]);
        }
        delete[] (segmentedFgPCASYS);
    }
    
    if (enhancedPCASYS != 0) {
        for (int i = 0; i < h; i++) {
                delete[] (enhancedPCASYS[i]);
        }
        delete[] (enhancedPCASYS);
    }
    
    if (pixelOrPCASYS != 0) {
        for (int i = 0; i < h; i++) {
                delete[] (pixelOrPCASYS[i]);
        }
        delete[] (pixelOrPCASYS);
    }
    
    if (avOrX != 0) {
        for (int i = 0; i < hOr; i++) {
                delete[] (avOrX[i]);
                delete[] (avOrY[i]);
        }
        delete[] (avOrX);
        delete[] (avOrY);
    }
    
    if (avOrRegX != 0) {
        for (int i = 0; i < hOr; i++) {
                delete[] (avOrRegX[i]);
                delete[] (avOrRegY[i]);
        }
        delete[] (avOrRegX);
        delete[] (avOrRegY);
    }
    if (featVect != 0)
        delete[] (featVect);
    
    if (avOrX2 != 0) {
        for (int i = 0; i < hOr2; i++) {
                delete[] (avOrX2[i]);
                delete[] (avOrY2[i]);
        }
        delete[] (avOrX2);
        delete[] (avOrY2);
    }

    MemUtil::freeMatrix(&segmentedImage);
    
    pcasys = 0;
    if (ImgUtil::DEBUG)
        cout << "FingerprintImg: Destructed" << endl;
}

float *FingerprintImg::run(string file, PCASYSManager *pcasys){
    float *featVect;

    setPCASYSManager(pcasys);
    readFingerprintImage(file);
    segment();

    enhance();
    computeOrientations();
    computeCore_r92a();
    registerOrientations();

    featVect = extractOrientationFeatureVector();

    dimFV = hTf;
    
    return featVect;
}

int FingerprintImg::readFingerprintImage(string name) {
    
    int state;
    int img_type;
    unsigned char *idata;
    int id, ippi, ilen;
    unsigned char *img;
    float **imgEnhanced;
    unsigned char **imgReSized;
    
    // Si hay que utilizar huellas realzadas
    if (enhanceHong){
        if (ImgUtil::DEBUG)
            cout << "Hay que realzar la huella" << endl;        

        string fileName;
        GuardarRealce *gr = new GuardarRealce();
        
        fileName = name;
        
        if (ImgUtil::doAdjust){
            if (ImgUtil::fAdjust == 0.5)
                fileName.append("_adjust05_realce.dat");
            else
                fileName.append("_adjust_realce.dat");
        }
        else
            fileName.append("_redim_realce.dat");

        //Leemos la matriz
        image = gr->leer(fileName, &h, &w);
        
        //Si no se ha podido leer, la creamos para guardarla
        if (image == 0){
            
            //Leer la imagen
            state = read_and_decode_grayscale_image((char *)name.c_str(), &img_type,
                   &idata, &ilen, &w, &h, &id, &ippi);
            img = MemUtil::createVector<unsigned char>(h * w); //new unsigned char[h * w];
            memcpy(img, idata, sizeof(unsigned char) * h * w);
            //
            
            //Redimensionarla si es necesario
            if ((!ImgUtil::doAdjust) && ((h_std != h) || (w_std != w))) {              
                if (state == 0) {
                    AdaptingImageSize *ais = new AdaptingImageSize(h_std, w_std, img, h, w);
                    imgReSized = ais->reSize(ais->getImage(), h, w);
                    image = ais->imageToVector(imgReSized, ais->gethStd(), ais->getwStd());
                    h = ais->gethStd();
                    w = ais->getwStd();

                    delete ais;
                    MemUtil::freeMatrix(&imgReSized);
                }
                
                MemUtil::freeMatrix(&img);
                segment();
            }
            else{              
                image = img;
                setSegmentedPCASYS(image, h, w);
            }
            //

            //Realzamos la huella 
            ImageEnhancementHong *ieh = new ImageEnhancementHong(segmentedPCASYS, h, w);
            ieh->enhance(/*MO*/100, /*VO*/100, true);
            int aux = gaborSupLeungHong * ImgUtil::fAdjust;
            if (aux % 2 == 0)
                aux--;
            imgEnhanced = ieh->filterImage(aux, deltaGaborLeungHong * ImgUtil::fAdjust, deltaGaborLeungHong * ImgUtil::fAdjust, true);
            //
            
            //Guardamos la huella realzada en el directorio de la huella .wsq
            gr->escribir(fileName, imgEnhanced, h, w);
            //
            
            MemUtil::freeMatrix(&image);        
            image = MemUtil::createVector<unsigned char>(h * w);
            /////////////// CAMBIAR
            for (int i = 0; i < h; i++)
                for (int j = 0; j < w; j++)
                    image[i * w + j] = imgEnhanced[i][j];
            //image = imgEnhanced;
            ///////////////
            MemUtil::freeMatrix(&imgEnhanced);  
            
            delete ieh;         
            return state;
        }
        delete gr;
        
        return 0;
    }
    //
    
    //Si no hay que realzar se lee la huella normalmente.
    else{
        state = read_and_decode_grayscale_image((char *)name.c_str(), &img_type,
                    &idata, &ilen, &w, &h, &id, &ippi);

        if (ImgUtil::DEBUG) {
            cout << "FingerprintImg: \tImage read: " << name << endl 
                << "\tImgSize: " << w << "x" << h << endl
                << "\tid: " << id  << endl
                << "\tDPI " << ippi << endl;
        }

        img = MemUtil::createVector<unsigned char>(h * w); //new unsigned char[h * w];

        if (ImgUtil::DEBUG)
            cout << "FingerprintImg: Image matrix created" << endl;
        memcpy(img, idata, sizeof(unsigned char) * h * w);


        if ((!ImgUtil::doAdjust) && ((h_std != h) || (w_std != w))){
            if (state == 0) {
                AdaptingImageSize *ais = new AdaptingImageSize(h_std, w_std, img, h, w);
                imgReSized = ais->reSize(ais->getImage(), h, w);
                image = ais->imageToVector(imgReSized, ais->gethStd(), ais->getwStd());
                h = ais->gethStd();
                w = ais->getwStd();


                if (ImgUtil::DEBUG) {
                    cout << "\tNew ImgSize: " << w << "x" << h << endl;
                    string fileName2 = "AdaptingImageSize_reSized";
                    ImgUtil::writeGrayScaleImage(imgReSized, h, w, fileName2);
                }  
                delete ais;
                MemUtil::freeMatrix(&imgReSized);
            }
            MemUtil::freeMatrix(&img);
        }
        else
            image = img;
       

        delete[] (idata);
        if (ImgUtil::DEBUG)
            cout << "FingerprintImg: All read OK " << endl;

        return state;    
    }
    //
}

int FingerprintImg::setPCASYSManager(PCASYSManager* pcasys) { this->pcasys = pcasys; }

// segmenta, enhance, orientaciones, centro, registro orientaciones
int FingerprintImg::segment() {
    
    if (ImgUtil::DEBUG)
        cout << "FingerprintImg: Segmenting image" << endl;
    
    
    segmentedPCASYS = pcasys->segment(image, w, h, &segmentedFgPCASYS);
    
    if (ImgUtil::DEBUG)
        cout << "FingerprintImg: Image segmented!" << endl;
    
    /* segmentedPCASYS  Dimensiones h x w!! es 2-D */
    
    w = pcasys->getSW();
    h = pcasys->getSH();
 
    sfgh = pcasys->getSfgH();
    sfgw = pcasys->getSfgW();
    
    unsigned char *segImg = MemUtil::createVector<unsigned char>(w * h);
    unsigned char **segPCASYS = MemUtil::create2DMatrix<unsigned char>(h, w);
    
    for (int i = 0; i < h; i++) {
        memcpy(segImg + i * w, segmentedPCASYS[i], w);        
        memcpy(segPCASYS[i], segmentedPCASYS[i], w);
    }
       
    // reserva viene del C
    delete [] image;
    image = segImg;
    // Así luego podemos utilizar freeMatrix en segmentedPCASYS
    for (int i = 0; i < h; i++) {
            delete[] (segmentedPCASYS[i]);
    }
    delete[] (segmentedPCASYS);
    segmentedPCASYS = segPCASYS;
    
    segmentedImage = MemUtil::create2DMatrix<float>(h,w);
    // segmentedImage initialization 
    // Necesario así porque pasamos de unsigned char a float
    for (int i = 0; i < h; i++)
        for (int j = 0; j < w; j++)
                segmentedImage[i][j] = (float) segmentedPCASYS[i][j]; 
    
    if (ImgUtil::DEBUG) {
        string fileName2 = "Fingerprint";
        ImgUtil::writeGrayScaleImage(segmentedPCASYS, h, w, fileName2);
    }
    
    if (ImgUtil::DEBUG){
        cout << "FingerprintImg: Image segmented copied in original!" << endl;
    }
    return 0;
}

int FingerprintImg::enhance() {
    
    if (ImgUtil::DEBUG)
        cout << "FingerprintImg: Enhancing image" << endl;
    

    enhancedPCASYS = pcasys->enhance(segmentedPCASYS, w, h);
    
    enhancedImage = MemUtil::createVector<unsigned char>(w * h);
    
    for (int i = 0; i < h; i++)
        memcpy(enhancedImage + i * w, enhancedPCASYS[i], w);

    if (ImgUtil::DEBUG) {    
        cout << "FingerprintImg: Image enhanced!" << endl;

        string fileName2 = "FingerprintEnhancedPCASYS";
        ImgUtil::writeGrayScaleImage(enhancedPCASYS, h, w, fileName2);
    }
    return 0;
}

int FingerprintImg::computeOrientations() {
    if (ImgUtil::DEBUG)
        cout << "FingerprintImg: Computing orientation image" << endl;
      
    pixelOrPCASYS = pcasys->computeOrientations(enhancedPCASYS, w, h, &avOrX, &avOrY);
       
    wOr = pcasys->getOrW();
    hOr = pcasys->getOrH();
    
    if (ImgUtil::DEBUG)
        cout << "FingerprintImg: Orientation image computed!" << endl;
    return 0;
}


int FingerprintImg::computeCore_r92a() {
    if (ImgUtil::DEBUG)
        cout << "FingerprintImg: Computing the core with r92a" << endl;
      
    pcasys->computeCore_r92a(avOrX, avOrY, &coreX, &coreY);

    if (ImgUtil::DEBUG)
        cout << "FingerprintImg: Core found: " << coreX << ", " << coreY << endl;
    return 0;
}

int FingerprintImg::computeCore_r92a(float **avOrX, float **avOrY, int wOr, int hOr) {
    if (ImgUtil::DEBUG)
        cout << "FingerprintImg: Computing the core with r92a" << endl;
      
    pcasys->computeCore_r92a(avOrX, avOrY, wOr, hOr ,&coreX, &coreY);

    if (ImgUtil::DEBUG)
        cout << "FingerprintImg: Core found: " << coreX << ", " << coreY << endl;
    return 0;
}

int FingerprintImg::computeCore_r92a(float **avOrX, float **avOrY, int wOr, int hOr, int block) {
    if (ImgUtil::DEBUG)
        cout << "FingerprintImg: Computing the core with r92a" << endl;
      
    pcasys->computeCore_r92a(avOrX, avOrY, wOr, hOr ,&coreX, &coreY, block);

    if (ImgUtil::DEBUG)
        cout << "FingerprintImg: Core found: " << coreX << ", " << coreY << endl;
    return 0;
}

int FingerprintImg::registerOrientations() {
    if (ImgUtil::DEBUG)    
        cout << "FingerprintImg: Registering orientation image..." << endl;
      
    pcasys->registerOrientations(pixelOrPCASYS, coreX, coreY, &avOrRegX, &avOrRegY);

    if (ImgUtil::DEBUG)
        cout << "FingerprintImg: Orientations image registered" << endl;
    return 0;
}

float *FingerprintImg::extractOrientationFeatureVector() {
    
    
    // tfh has the size of the feature vector, tfw the original size of the image
    featVect = pcasys->extractOrientationFeatureVector(avOrRegX, avOrRegY);
    
    wTf = pcasys->getTfW();
    hTf = pcasys->getTfH();
    
    float *fvOut = MemUtil::createVector<float>(hTf);
    memcpy(fvOut, featVect, sizeof(float) * hTf);
    
    return fvOut;
}


int FingerprintImg::computeOrientationsForPseudoRidge() {
    
    pcasys->computeOrientationsForPseudoRidge(pixelOrPCASYS, &avOrX2, &avOrY2);

    wOr2 = pcasys->getOrW2();
    hOr2 = pcasys->getOrH2();
    
    return 0;
}

int FingerprintImg::smoothOrientationsForPseudoRidge() {
    
    has_conup = pcasys->smoothOrientationsForPseudoRidge(segmentedFgPCASYS, avOrX2, avOrY2);

    
    return 0;
}

int FingerprintImg::getPseudoRidges() {
    
    unsigned char **pseudoRidges;
    pseudoRidges = pcasys->getPseudoRidges(segmentedFgPCASYS, avOrX2, avOrY2);

    
    return 0;
}



float **FingerprintImg::getFingerCode(int centerDisplacement, int nSectors, int nBands, int bandWidth, 
        int nFilters, int V0, int M0, double gFreq, double gDX, double gDY, 
        int gSupX, int gSupY, int rotate, int type, bool **masc, int hMasc ) {
    
    
    if (ImgUtil::DEBUG)
        cout << "FingerprintImg: Creando FingerCode" << endl;
    
    this->nFilters = nFilters;
    int squareSize = (nBands * bandWidth * 2) + bandWidth + 1; 
        
    GaborFilter *gb = new GaborFilter(nFilters, gFreq, 1, gDX, gDY, gSupX, gSupY);
    FingerCode *fc = new FingerCode(squareSize, segmentedImage, h, w, coreX, coreY, centerDisplacement, gSupX, gSupY); 
    
    
        //cout << "FingerprintImg: go set mask" << endl;
    if (masc != 0)
        fc->setSegmentationMask(masc, hMasc);
        //cout << "FingerprintImg: mask puesta" << endl;
    
    float **fingercodeOut = 0;
    if (fc->setLabelsFingerCode(nSectors, nBands, bandWidth) != -1) { // Fingerprint not rejected
        numSectors = fc->getNumSectors();
        fc->normalize(V0, M0);
        fc->filterImage(gb);
        fingercode = fc->computeFingerCode(type);
        
        if (rotate != -1)
            fingercode = fc->rotateFingerCode(rotate);
        
        fingercodeOut = MemUtil::copy2DMatrix(fingercode, nFilters, nBands * nSectors); // solución temporal
        fingercode = fingercodeOut;
    }
    else{      
        fingercode = MemUtil::create2DMatrix<float>(nFilters,nBands * nSectors,-1.0);   
        fingercodeOut = fingercode; //MemUtil::copy2DMatrix(fingercode, nFilters, nBands * nSectors);  
        this->numSectors = nBands * nSectors;
    }
    
    delete gb;
    delete fc;
    
    
    if (ImgUtil::DEBUG) {
        cout << "FingerprintImg: FingerCode creado" << endl;
        
        string fileName2 = "FingerprintCore";
        ImgUtil::writeGrayScaleImageCore(segmentedImage, h, w, fileName2, coreX, coreY);
    }
    
    return fingercodeOut;
}



float **FingerprintImg::getFingerCodeLeung(float **imagen, int centerDisplacement, bool **masc, int hMasc) {

    nFilters = numGaborFiltersModLeung;
    int squareSize = (numBlocksLeung * 2 + 1) * (int)(pixelPerBLockLeung * ImgUtil::fAdjust);

    
    //Comprobamos los valores de los soportes de Gabor porque no pueden ser pares.
    int auxGSX = gaborSupportXLeung * ImgUtil::fAdjust;
    if (auxGSX % 2 == 0)
        auxGSX--;
    
    int auxGSY = gaborSupportYLeung * ImgUtil::fAdjust;
    if (auxGSY % 2 == 0)
        auxGSY--;

    GaborFilter *gb = new GaborFilter(nFilters, gaborFrequencyLeung / ImgUtil::fAdjust, factorLeung * ImgUtil::fAdjust, gaborDXLeung * ImgUtil::fAdjust, gaborDYLeung * ImgUtil::fAdjust, auxGSX, auxGSY);
    FingerCode *fc = new FingerCode(squareSize, imagen, h, w, coreX, coreY, centerDisplacement * ImgUtil::fAdjust,auxGSX, auxGSY); 
    
    fc->setSegmentationMask(masc, hMasc);
    
    fc->setLabelsSquareFingerCode();
    numSectors = fc->getNumSectors();
    fc->normalize(V0Leung, M0Leung);
    fc->circleAverage(floor(filterRadioLeung * ImgUtil::fAdjust));
    fc->filterImage(gb);
    fc->computeMascPerBlock();
    fingercode = fc->computeFingerCode(0); // std

    float **fingercodeOut = MemUtil::copy2DMatrix(fingercode, nFilters, fc->getNumSectors());
    fingercode = fingercodeOut;
    delete gb;
    delete fc;
   
    
    return fingercodeOut;
}


int FingerprintImg::writeOrientationImage(string fileName, int type) {
    
    float **X, **Y;
    int wIn, hIn;
    
    if (ImgUtil::DEBUG)
        cout << "FingerprintImg: Writing orientation image, construction of image..." << endl;
    
    if (type == ORIENTATION) {
        X = avOrX;
        Y = avOrY;
        wIn = wOr;
        hIn = hOr;
    }
    else if (type == OR_REGISTERED) {
        X = avOrRegX;
        Y = avOrRegY;
        wIn = wOr;
        hIn = hOr;
    }
    else if (type == OR_PSEUDO) {
        X = avOrX2;
        Y = avOrY2;
        wIn = wOr2;
        hIn = hOr2;
    }
    else 
        return -1;
        
    ImgUtil::writeOrImg(X, Y, wIn, hIn, fileName);
    
    return 0;
}

void FingerprintImg::setSegmentedPCASYS(unsigned char *img, int h, int w){
        
    segmentedPCASYS = MemUtil::create2DMatrix<unsigned char>(h, w);
        
    for (int i = 0; i < h; i++)
        memcpy(segmentedPCASYS[i], img + i * w, w);
    
    segmentedImage = MemUtil::create2DMatrix<float>(h,w);
    // segmentedImage initialization
    for (int i = 0; i < h; i++)
        for (int j = 0; j < w; j++)
                segmentedImage[i][j] = (float) segmentedPCASYS[i][j]; 

    
	if (ImgUtil::DEBUG){
    	string filePrueba = "ImagenRealzadaLeida";
    	ImgUtil::writeGrayScaleImage(segmentedImage, h, w, filePrueba,1);
	}
        
}



int FingerprintImg::setCore(int coreX, int coreY) { this->coreY = coreX; this->coreX = coreY; } // están cambiados en FingerprintImg

float **FingerprintImg::doAndGetSegmentedImageFloat(){
    
    segmentedImage = MemUtil::create2DMatrix<float>(h, w);
    int a, b;

    for (int i = 0; i < h * w; i++){     
        a = i/w;
        b = i%w;
        segmentedImage[a][b] = (float) image[i];
    }
    
    return segmentedImage;
}

void FingerprintImg::setSegmentationMask(bool **masc, int hMasc, int wMasc) {
   
    if (ImgUtil::DEBUG){
        for (int i = 0; i < hMasc; i++) {
            for (int j = 0; j < wMasc; j++) 
                cout << " " << masc[i][j];
            cout << endl;
        }
    }
    mascara = MemUtil::create2DMatrix<bool>(hMasc,wMasc);
    // fingerMasc initialization
    for (int i = 0; i < hMasc; i++){   
        for (int j = 0; j < wMasc; j++) {
                mascara[i][j] = masc[i][j];            
        }
    }
    if (ImgUtil::DEBUG) {
        string fileName3  = "FingerMasc";
        ImgUtil::writeGrayScaleImage(mascara, hMasc, wMasc, fileName3);
    }
}

void FingerprintImg::alterOrientations(){
    
    for (int i = 0; i < hOr; i++)
        for (int j = 0; j < wOr; j++)
            if (!mascara[i][j]){
                avOrX[i][j] = 0;
                avOrY[i][j] = 0;
            }
    
    if (ImgUtil::DEBUG){
        
        string fileName2;
        fileName2 = "Orientations00";
        ImgUtil::writeOrImg(avOrX, avOrY, wOr, hOr, fileName2);
    }
}



void FingerprintImg::setEnhanceHong(bool enhance) { 
    this->enhanceHong = enhance; 
}