/* 
 * \file        FingerprintImg.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * Header file for FingerprintImg (Fingerprint Image) class
 *
 * 
 */

#ifndef FINGERPRINTIMG_H
#define	FINGERPRINTIMG_H

#include "PCASYSManager.h"


// CONSTANTS to write orientation image

const int ORIENTATION = 0;
const int OR_REGISTERED = 1;
const int OR_PSEUDO = 2;
const int h_std = 480;
const int w_std = 512;

using namespace std;

/**
 * \class FingerprintImg
 *
 * The FingerprintImg class represents a Fingerprint Image which is then used to extract the features for classification
 */
class FingerprintImg {
    
public:
    
    /** Default constructor */
    FingerprintImg();

    /** Default destructor */
    ~FingerprintImg();
    
    /**
     * Executes the feature extractor PCASYS method from the fingerprint in file
     * \param file fingerprint
     * \param pcasys pcasys manager
     * \return fingerprint feature vector
     */
    float *run(string file, PCASYSManager *pcasys);
    
    /**
     * Read a fingerprint file in WSQ format.
     * \param name String with the file name
     * \return -1 if an error has ocurred. 0 otherwise
     */
    int readFingerprintImage(string name);
    
    /**
     * It returns the number of filters used when FilterCode done
     * \return number of filters
     */
    int getNFilters() const;
    
    /**
     * It returns the number of sectors used when FilterCode done
     * \return number of sectors
     */
    int getNumSectors() const;
    
    /**
     * It returns the width of the image
     * \return the width of the image
     */
    int getW() const ;
    
    /**
     * It returns the height of the image
     * \return the height of the image
     */
    int getH()const;
    
    /**
     * It returns the X coordinate of the core
     * \return the width of the image
     */
    int getCoreX() const ;
    
    /**
     * It returns the Y coordinate of the core
     * \return the width of the image
     */
    int getCoreY()const;
    
    /**
     * It returns the length of the feature vector
     * \return the length of the feature vector
     */
    int getFeatureVectorSize() const;
    
    /**
     * It returns whether the fingerprint has a conup or not (upwards concave lobe)
     * \return whether the fingerprint has a conup or not (upwards concave lobe)
     */
    bool hasConup() const;
    
    /**
     * It returns the pointer to the image
     * \return the image pointer
     */
    unsigned char *getImage() const;
    
    /**
     * Set the pcasys manager to be used to process the image
     * @param pcasys PCASYSManager to be used
     * @return -1 if an error has ocurred. 0 otherwise
     */
    int setPCASYSManager(PCASYSManager *pcasys);
    
    /**
     * Segment the fingerprint image and replaces the original image
     * \return -1 if an error has ocurred. 0 otherwise
     */
    int segment();
    
    /**
     * Enhances the fingerprint image
     * \return -1 if an error has ocurred. 0 otherwise
     */
    int enhance();
    
    /**
     * Computes the orientation image of a fingerprint image
     * \return -1 if an error has ocurred. 0 otherwise
     */
    int computeOrientations();
    
    /**
     * Computes the core of the fingerprint using r92a algorithm
     * \return -1 if an error has ocurred. 0 otherwise
     */
    int computeCore_r92a();
    
    /**
     * Computes the core of the fingerprint using r92a algorithm
     * \param avOrX (sin(2 theta)), averaged orientation map
     * \param avOrY (cos(2 theta)), averaged orientation map
     * \param wOr
     * \param hOr
     * \return -1 if an error has ocurred. 0 otherwise
     */
    int computeCore_r92a(float **avOrX, float **avOrY, int wOr, int hOr);

    /**
     * Computes the core of the fingerprint using r92a algorithm
     * \param avOrX (sin(2 theta)), averaged orientation map
     * \param avOrY (cos(2 theta)), averaged orientation map
     * \param wOr
     * \param hOr
     * \param block orientation block size
     * \return -1 if an error has ocurred. 0 otherwise
     */
    int computeCore_r92a(float **avOrX, float **avOrY, int wOr, int hOr, int block);
    
    /**
     * Computes the registered orientation image using the core previously computed
     * that is, it computes, from the pixelwise-orientations made earlier, a new,
     * "registered" 28x30-vector grid of average orientations, by using 
     *  averaging squares that are shifted according to the r92 resultcore
     * \return -1 if an error has ocurred. 0 otherwise
     */
    int registerOrientations();

    /**
     * It stracts the feature vector using the registered orientation image
     * The feature vector has 64 o 128 components depending on the pcasys 
     * configuration parameters (RNN or PNN)
     * \return the feature vector
     */
    float *extractOrientationFeatureVector();
    
    /**
     * It computes the new orientation image components for the pseudo ridge analyzer
     * \return -1 if an error has ocurred. 0 otherwise
     */
    int  computeOrientationsForPseudoRidge();
    
    /**
     * It computes the smoothed orientation image (x and y) from the orientations image for pseudo ridges
     * It uses the function for finding upwardsconcave lobes
     * \return -1 if an error has ocurred. 0 otherwise
     */
    int  smoothOrientationsForPseudoRidge();
    
    /**
     * It computes the pseudo-ridges from the smoothed orientation image (x and y) 
     * \return -1 if an error has ocurred. 0 otherwise
     */
    int getPseudoRidges();
    
    
////////    int computeSingularPointsPoincare();
    
    /**
     * It writes an orientation image to a file
     * \param fileName file name of the orientation image (without extension)
     * \param type of orientation map (see constant definition)
     * \return -1 if an error has ocurred. 0 otherwise
     */
    int writeOrientationImage(string fileName, int type);
    
    /**
     * It returns the fingercode of a fingerprint which core has been previously computed
     * see JAIN99 for details. 
     * \param centerDisplacement the displacement of the center downwards
     * \param nSectors number of sectors for FC
     * \param nBands number of bands for FC
     * \param bandWidth width of each band for FC
     * \param nFilters number of filters to be used
     * \param V0 desired variance for the normalization
     * \param M0 desired mean for the normalization
     * \param gFreq frequency of the gabor filter
     * \param gDX sigma_x for the gaussian of the gabor filter
     * \param gDY sigma_y for the gaussian of the gabor filter
     * \param gSupX size of the gabor support X 
     * \param gSupY size of the gabor support Y
     * \return the finger code. 0 if the fingerprint is rejected
     */
    float **getFingerCode(int centerDisplacement, int nSectors, int nBands, int bandWidth, 
        int nFilters, int V0, int M0, double gFreq, double gDX, double gDY, int gSupX, int gSupY, int rotate, int type, bool **masc, int hMasc);
    
    
    /**
     * It returns the modified fingercode of a fingerprint which core has been 
     * previously computed (see LEUNG2011 for details)
     * THE CURRENT IMPLEMENTATION AS THE ORIGINAL FINGERCODE SHIFTS DOWN THE CENTER
     * \param imagen image to obtain the finger code from
     * \param centerDisplacement the displacement of the center downwards
     * \param masc mask of the fingerprint segmentation
     * \param hMasc height of the mask
     * \return the finger code.
     */
    float **getFingerCodeLeung(float **imagen, int centerDisplacement, bool **masc, int hMasc);
        
    /**
     * Access operator
     * \param i Column index
     * \param j Row index
     * \return The pixel intensity value in the positions (i,j) of the image
     */
    unsigned char get(int i, int j)const;
    
    /**
     * It returs the segmented image
     * \return the segmented image
     */
    unsigned char **getSegmentedImage()const;
    
    
    float **getSegmentedImageFloat()const;
    
    /**
     *  Returns the Y (vertical) components of the orientation map
     * \return The SinY components of the orientation map
     */
    float **getYSin() const;
    
    /**
     *  Returns the X (horizontal) components of the orientation map
     * \return The CosX components of the orientation map
     */    
    float **getXCos() const;
    
    /**
     *  Returns the Y (vertical) components of the registered orientation map
     * \return The SinY components of the registered  orientation map
     */
    float **getRegisteredYSin() const;
    
    /**
     *  Returns the X (horizontal) components of the registered  orientation map
     * \return The CosX components of the registered  orientation map
     */    
    float **getRegisteredXCos() const;
    
    /**
     * It sets the core of the fingerprint
     * \param coreX position of the core in X (row, is changed to column)
     * \param coreY position of the core in Y (column, is changed to row)
     * \return 
     */
    int setCore(int coreX, int coreY);
    
    /**
     * It returns the width of the orientation map
     * \return the width of the orientation map
     */
    int getOrW() const;

    /**
     * It returns the height of the orientation map
     * \return the height of the orientation map
     */
    int getOrH() const;
    
    /**
     * It returns the feature vector's size
     * \return feature vector's size
     */
    int getDimFV() const;
    
    /**
     * It initializates the matrix segmentedPCASYS when no segment is done before in pcasys
     * \param img image to copy as segmented image
     * \param h image's height
     * \param w image's width
     */
    void setSegmentedPCASYS(unsigned char *img, int h, int w);
    
    /**
     * It initializates the matrix segmentedImage when no segment is done before in pacasys
     * \return the segmented image
     */
    float **doAndGetSegmentedImageFloat();
    
    /**
     * It sets the mask of the segmentation for the fingerprint
     * \param masc mask of the fingerprint image (0 background, 1 foreground)
     * \param hMasc height of the masc 
     * \param wMasc width of the masc 
     */
    void setSegmentationMask(bool **masc, int hMasc, int wMasc);
    
    /**
     * Eliminates the orientations out of the segmentation image (masc)
     */
    void alterOrientations();
    
    /**
     * It specifies whether to enhance the fingerprint or not
     * @param enhance Whether to enhance or not
     */
    void setEnhanceHong(bool enhance);
    
private:
    
    PCASYSManager *pcasys;              //!< reference to the PCASYSmanager of the fingerprint
    
    unsigned char *image;               //!< image of the fingerprint
    unsigned char *enhancedImage;       //!< enhanced image of the fingerprint
    bool **mascara;                  //!< Poincare masc to segment orientations
    
    
    int w;              //!< Width of the image
    int h;              //!< Height of the image
    
    int coreX;          //!< X position of the core (X in w)
    int coreY;          //!< Y position of the core (Y in h)
    
    int sfgw;    //!< Width of the segmented foreground image (segmentedFgPCASYS)
    int sfgh;    //!< Height of the segmented foreground image (segmentedFgPCASYS)
    
    int wOr;     //!< Width of the orientation map (avOr-)
    int hOr;     //!< Height of the orientation map (avOr-)
    
    int wTf;     //!< Original size of the image before appliying the PCA
    int hTf;    //!< Size of the new feature vector after applying the PCA
    
    int wOr2;    //!< Width of the overlapped orientation map (avOr-2)
    int hOr2;    //!< Height of the overlapped orientation map (avOr-2) 
    
    unsigned char **segmentedPCASYS;    //!< Segmented fingerprint image
    unsigned char **segmentedFgPCASYS;  //!< Foreground mask of the segmented fingerprint image (each pixel corresponds to a block of the segmentedPCASYS image
    unsigned char **enhancedPCASYS;     //!< Enhanced fingerprint image
    char **pixelOrPCASYS;               //!< Pixel orientation map (an orientation for each pixel)
    float **avOrX;                      //!< (Sin(2 theta), averaged orientation maps, blocks of 16x16
    float **avOrY;                      //!< (Cos(2 theta), averaged orientation maps, blocks of 16x16
    float **avOrRegX;                   //!< avOrX registered to the core point
    float **avOrRegY;                   //!< avOrY registered to the core point
    float **avOrX2;                     //!< (Sin(2 theta), averaged orientation maps, overlapped blocks of 16x16
    float **avOrY2;                     //!< (Cos(2 theta), averaged orientation maps, overlapped of 16x16
    
    bool has_conup;                     //!< Whether the fingerprint has a conup or not (is a whorl)
    float *featVect;                    //!< Feature vector extracted by PCASYS, reduced by PCA
    float **fingercode;                 //!< FingerCode obtained from the Gabor Filters
    
    int nFilters;                       //!< Number of filters used in the fingercode
    int numSectors;                     //!< Number of sectors in the fingercode
    float **segmentedImage;             //!< The float copy of the segmentedPCASYS image
    
    int dimFV;                          //!< Dimension of feature vector
    
    
    bool enhanceHong;                       //!< Whether to enhance the fingerprint with Hong98
};

inline int FingerprintImg::getDimFV() const { return dimFV;}

inline bool FingerprintImg::hasConup() const{ return has_conup;}

inline int FingerprintImg::getW() const{ return w;}

inline int FingerprintImg::getH() const{ return h;}


inline int FingerprintImg::getOrW() const{ return wOr;}

inline int FingerprintImg::getOrH() const{ return hOr;}


inline int FingerprintImg::getCoreX() const{ return coreX;}

inline int FingerprintImg::getCoreY() const{ return coreY;}


inline unsigned char FingerprintImg::get(int i, int j) const{ return segmentedPCASYS[i][j];}

inline int FingerprintImg::getFeatureVectorSize() const{ return hTf;}

inline unsigned char *FingerprintImg::getImage() const{ return image;}

inline unsigned char **FingerprintImg::getSegmentedImage() const{ return segmentedPCASYS;}


inline float **FingerprintImg::getSegmentedImageFloat() const{ return segmentedImage;}


inline float **FingerprintImg::getXCos() const { return avOrY; }
inline float **FingerprintImg::getYSin() const { return avOrX; }
inline float **FingerprintImg::getRegisteredXCos() const { return avOrRegY; }
inline float **FingerprintImg::getRegisteredYSin() const { return avOrRegX; }


inline int FingerprintImg::getNFilters() const { return nFilters; }
inline int FingerprintImg::getNumSectors() const { return numSectors; }


#endif	/* FINGERPRINTIMG_H */

