/* 
 * \file        GaborFilter.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * The implementation of GaborFilter class
 *
 * 
 */


#include "GaborFilter.h"
#include "ImgUtil.h"
#include "MemUtil.hpp"


GaborFilter::GaborFilter() {
    gaborFilters = 0;
    orientations = 0;
    nFilters = -1;
}

GaborFilter::GaborFilter(int nFilters, float gabor_frequency, double leung_factor, double gabor_dx, double gabor_dy,int gaborSupX, int gaborSupY) {
        
    int idxFilter;
    int idxRow2, idxCol2;
    int x, y;
    float xprima, yprima;
    float theta;
    
    gaborSupportX=gaborSupX;
    gaborSupportY=gaborSupY;
    this->nFilters = nFilters;
    
    int gaborHalfSupportX = (gaborSupportX - 1) / 2;
    int gaborHalfSupportY = (gaborSupportY - 1) / 2;      
    
    
    gaborFilters = MemUtil::create3DMatrix<float>(nFilters,gaborSupportX,gaborSupportY);
    gaborMask = MemUtil::create3DMatrix<bool>(nFilters,gaborSupportX,gaborSupportY);
      
    orientations = MemUtil::createVector<float>(nFilters);


    for (int i = 0; i < nFilters; i++)
        orientations[i] = PI / (float)nFilters * i;

    for (idxFilter = 0; idxFilter < nFilters; idxFilter++){
        for (idxRow2 = 0; idxRow2 < gaborSupportX; idxRow2++){
            x = idxRow2 - gaborHalfSupportX;

            for (idxCol2 = 0; idxCol2 < gaborSupportY; idxCol2++){
                y = idxCol2 - gaborHalfSupportY;

                theta = orientations[idxFilter];
                xprima = x * sin(theta) + y * cos(theta);
                yprima = x * cos(theta) - y * sin(theta);

                gaborFilters[idxFilter][idxRow2][idxCol2] =
                        exp(-0.5 * ((xprima * xprima) / (gabor_dx * gabor_dx) + 
                        ((leung_factor * leung_factor) * (yprima * yprima) / (gabor_dy * gabor_dy)))) *  
                        cos( 2.0 * PI * gabor_frequency * xprima);
                
                if (abs(gaborFilters[idxFilter][idxRow2][idxCol2]) < 0.05)
                    gaborMask[idxFilter][idxRow2][idxCol2] = false;
                else
                    gaborMask[idxFilter][idxRow2][idxCol2] = true;
                
            }
        }
    }
    
    max = 0;
    min = 0;
    for (idxRow2 = 0; idxRow2 < gaborSupportX; idxRow2++){
        //x = idxRow2 - gaborHalfSupportX;
        for (idxCol2 = 0; idxCol2 < gaborSupportY; idxCol2++){
            //y = idxCol2 - gaborHalfSupportY;
                
            if (gaborFilters[0][idxRow2][idxCol2] > 0)
                max += gaborFilters[0][idxRow2][idxCol2];
            else
                min += gaborFilters[0][idxRow2][idxCol2];
    
        }
    }
    max /= (float)(gaborSupportX * gaborSupportY);
    min /= (float)(gaborSupportX * gaborSupportY);
    
    if (ImgUtil::DEBUG)
        cout << "max gabor = " << max << " min gabor = " << min << endl;
}


GaborFilter::~GaborFilter() {
    MemUtil::freeMatrix(&orientations);
    
    MemUtil::freeMatrix(&gaborFilters);
    
    MemUtil::freeMatrix(&gaborMask);
}


float ***GaborFilter::filterImage(float **image, int h, int w) { // h and w out = h and w in - GABOR_SUPPORT -1
    int idxFilter;
    int posCol, posRow, posCol2, posRow2;
    float min, max, min2, max2;
    float sum;
    int gSup2X = (gaborSupportX - 1) / 2;
    int gSup2Y = (gaborSupportY - 1) / 2;
    
    //The filtered image is created and initialized .   
    float ***filteredImages = MemUtil::create3DMatrix<float>(nFilters,h - (gaborSupportX - 1),w - (gaborSupportY - 1));            

    for (idxFilter = 0; idxFilter < nFilters; idxFilter++){
        min = 25500000; max = -2505005;
        for (posRow = gSup2X + 1; posRow < h - gSup2X + 1; posRow++){
            for (posCol = gSup2Y + 1; posCol< w - gSup2Y + 1; posCol++){
                sum = 0;

                for (posRow2 = 0; posRow2 < gaborSupportX; posRow2++){
                    for (posCol2 = 0; posCol2 < gaborSupportY; posCol2++){

                        if (gaborMask[idxFilter][posRow2][posCol2])
                            sum += (image[posRow + posRow2 - gSup2X - 1]
                                        [posCol + posCol2 - gSup2Y - 1] 
                                    * gaborFilters[idxFilter][posRow2][posCol2]);
                    }
                }
                sum = (sum / (float)(gaborSupportX * gaborSupportY));
//cout << "Suma =" << sum << endl;                
                filteredImages[idxFilter][posRow - gSup2X - 1][posCol - gSup2Y - 1] = sum;

                if (sum < min)
                    min = sum;
                if (sum > max)
                    max = sum; 

            }
        }

        for (posRow = gSup2X + 1; posRow < h - gSup2X + 1; posRow++){
            for (posCol = gSup2Y + 1; posCol < w - gSup2Y + 1; posCol++){
                max2 = max * 0.4 * ImgUtil::fAdjust ; //0.4; // max * 0.55;// * 0.4;
                if (filteredImages[idxFilter][posRow - gSup2X - 1][posCol - gSup2Y - 1] > max2)
                    filteredImages[idxFilter][posRow - gSup2X - 1][posCol - gSup2Y - 1] = max2;
                min2 = min * 0.4 * ImgUtil::fAdjust;// max * 0.2; // min * 0.2;
                if (filteredImages[idxFilter][posRow - gSup2X - 1][posCol - gSup2Y - 1] < min2)
                    filteredImages[idxFilter][posRow - gSup2X - 1][posCol - gSup2Y - 1] = min2;           
                filteredImages[idxFilter][posRow - gSup2X - 1][posCol - gSup2Y - 1] = 
                        (filteredImages[idxFilter][posRow - gSup2X - 1][posCol - gSup2Y - 1] - min2) / (max2 - min2) * 255.0;
            }
        }
    }

    
    return filteredImages;
}



