/* 
 * \file        ImageEnhancementHong.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * Header file for ImageEnhancementHong (required for Leung2010) class
 *
 * 
 */

#ifndef IMAGEENHANCEMENTHONG_H
#define	IMAGEENHANCEMENTHONG_H

#include "PCASYSManager.h"


/* Constants to be moved to ParametersEnhancementHong.h as const! */
const int TEST_2 = 0;                 // 1 for TESTING frequency, amplitude and variance in ImageEnhancedHong

/**
 * \class ImageEnhancementHong
 *
 * The ImageEnhancementHong class implements 
 * 
 * Lin Hong, Yifei Wan, Anil Jain, 
 * "Fingerprint Image Enhancement: Algorithm and Performance Evaluation," 
 * IEEE Transactions on Pattern Analysis and Machine Intelligence, 
 * pp. 777-789, August, 1998 
 * 
 * It preprocesses the finger image for further feature extraction
 */
class ImageEnhancementHong {
public:
    /** Default constructor */
    ImageEnhancementHong();
    
    /**
     * Constructs an instance of the class to preprocess the finger image
     * \param image Finger  image to be preprocessed
     * \param h height of the image
     * \param w width of the image
     */
    ImageEnhancementHong(unsigned char **image, int h, int w);
    
    /** Default destructor */
    ~ImageEnhancementHong();
    
    /**
     * It initializes the images, width and height.
     * \param image original image
     * \param h image's height 
     * \param w image's width
     */
    void initialize(float **image, int h, int w);
    
    /**
     * It initializes the block size for image orientation by using ImgUtil::fAdjust 
     */
    void initializeBlocks();
    
    /**
     * Executes the image enhancement Hong method from the fingerprint in file
     * \param file fingerprint
     * \param pcasys pcasys manager
     * \param case2 true if we want the final image binarized 
     * \param segmentar indicates if Poincare's segmentation must be done
     * \return fingerprint fingerprint feature vector
     */
    float *run(string file, PCASYSManager *pcasys, bool case2, bool segmentar = false);
    
    /**
     * Eliminates the orientations out of the segmentation image (masc)
     */
    void alterOrientations();
    
    /**
     * Enhances (preprocesses) the fingerprint image performing all the phases
     * of the algorithm
     * \param Mo Mean to normalize the finger image
     * \param Vo Variance to normalize the finger image
     * \param binarize indicates if binarization must be done when calculating enhanced image
     * \return 
     */
    float **enhance(int Mo, int Vo, bool binarize = false);    //TODO: Needs more parameterization
    
    /**
     * Normalizes the image to a given mean and variance
     * \param desiredMean desired mean for the image
     * \param desiredVariance desired variance for the image
     */
    void normalize(int desiredMean, int desiredVariance);
    
    /**
     * Computes the orientation map for the fingerprint image in blocks of size
     * blockW x blockW.It uses sobel's method for gradient computation and then
     * it smooths the orientation map
     * The orientation map is stored in O and xCos, ySin
     * \param blockW the size of each orientation block
     */
    void createOrientationImage(int blockW);    //TODO: Parameterize with the smooth parameters!
    
    /**
     * Computes the ridge frequency map in each blockW x blockW block (same size
     * of the orientation blocks)
     * \param blockW block size used for the orientation computation
     * \param blockL block size to be used for the computation of the frequency in each orientation block
     */
    void createRidgeFrecuencyImage(int blockW, int blockL);
    
    /**
     * Computes the ridge frequency from a signature vector
     * \param X signature vector
     * \param l size of the signature vector
     * \param tabla block to safe the average between peaks and valleys
     * \param i tablas's row to be trated
     * \param j tablas's column to be trated
     * \return the ridge frequency
     */
    float computeRidgeFrequencyFromSignature(float *X, int l, float **tabla, int i, int j);
    
    /**
     * Filters (enhances) the image with the gabor filter which is adapted to each
     * block of the image given the frequency and the orientation
     * \param gaborSupport size of the gabor filter
     * \param sigmaX standard deviation in X for the gabor filter
     * \param sigmaY standard deviation in Y for the gabor filter
     * \param binarize indicates if binarization must be done when calculating enhanced image
     * \return the filtered image
     */
    float **filterImage(int gaborSupport, float  sigmaX, float sigmaY, bool binarize);
    
    /**
     * Computes the gabor filter coefficient for the given parameters
     * \param u x-position of the factor
     * \param v y-position of the factor
     * \param theta orientation information 
     * \param f frequency
     * \param sigmaX standard deviation in X for the gabor filter
     * \param sigmaY standard deviation in Y for the gabor filter
     * \return the factor for the gabor filter
     */
    float computeGaborFactor(int u, int v, float theta, float f, float sigmaX, float sigmaY);
    
    /**
     * Computes the segmentation image from the variance of the signature of each
     * orientation block
     * \param threshold the threshold to consider a block as foreground or background
     * \return the masc of the segmentated image in blocks
     */
    void varianceSegment(float threshold);
    
    
    void varianceSegment(float threshold, float **filteredImage);
    
    /**
     * It computes the segmentation image from the variance, amplitude and frequency
     * of the signature of each orientation block.
     * \return the masc of the segmentated image in blocks
     */
    bool **segmentImg();
    
    /**
     * Calculates the distance of the variables variance, frequency and amplitude to every cluster
     * There are 6 cluster. Two represents background of fingerprint and four the fingerprint.
     * \param i block's row index
     * \param j block's column index
     * \param clase cluster in wich calculate the distance
     * \return distance between cluster and block
     */
    float calculateDistance(int i, int j, int clase);
    
    //TODO: To be removed from here, located in UTIL.H
    //TODO: REMOVE!!!
    
    /**
     * It gives the pointer to the orientation map in X-coordinate
     * \return the pointer to xCos (cos(2 O))
     */
    float **getxCos() const;
    /**
     * It gives the pointer to the orientation map in Y-coordinate
     * \return the pointer to ySin (sin(2 O))
     */
    float **getySin() const;
    
    /**
     * It gives the width of the orientation map
     * \return width of the orientation map
     */
    int getwOr() const;
    
    /**
     * It gives the height of the orientation map
     * \return height of the orientation map
     */
    int gethOr() const;
    
    /**
     * It gives the masc matrix of the segmented image
     * \return 
     */
    bool **getMasc() const;
    
    /**
     * It returns the feature vector's size
     * \return feature vector's size
     */
    int getDimFV() const;
    
    /**
     * It gives the width of the image
     * \return width of the image
     */
    int getW() const;
    
    /**
     * It gives the height of the image
     * \return height of the image
     */
    int getH() const;
    
    
private:
    float **img;                //!< Reference to the fingerprint image data
    float **imgCopy;            //!< Initial image without any modification
    float **filteredImg;        //!< Reference to the filtered image data
    int h;                      //!< Height of the image
    int w;                      //!< Width of the image
    
    int orH;            //!< Height of the orientation map
    int orW;            //!< Width of the orientation map
    int blockW;         //!< Block size used to compute the orientation map
    
    
    float **O;          //!< Orientation of each block (in radians)
    float **xCos;       //!< cos(2 O) value of the cosine of the orientation (x-direction)
    float **ySin;       //!< sin(2 O) value of the sine of the orientation (y-direction)
    
    bool **segmentation;        //!< Image's segmentation
    
    float **blockVariance;              //!< Variance of the signature in each block
    float **ridgeFrequency;             //!< Ridge frequency in each block
    
    float **averagePeaksValleysSeg;        //!< Average height of peaks - average depth of valleys for segmentation               //!< Frecuency of pixels in every block
    float **ridgeFrequencySeg;             //!< Ridge frequency in each block
    float **blockVarianceSeg;              //!< Variance of the signature in each block
    
    bool **masc;                //!< Masc of the segmented image
    int **mascPrueba;
    
    float maxVar;         //!< Max value of variance
    float minVar;         //!< Min value of variance
    
    float maxAvg;         //!< Max value of ridges and valleys
    float minAvg;         //!< Min value of ridges and valleys
    
    float maxFreq;         //!< Max value of frequencies
    float minFreq;         //!< Min value of frequencies
    
    int dimFV;                          //!< Dimension of feature vector
    
    //Adjust parameters
    int block;             //!< Adjusted Leung's block size.
    int Wblock;            //!< Adjusted orientation map block width.
    int Lblock;            //!< Adjusted orientation map block lenght.
    
};

inline int ImageEnhancementHong::getDimFV() const { return dimFV;}
inline float **ImageEnhancementHong::getxCos() const { return xCos;}
inline float **ImageEnhancementHong::getySin() const { return ySin;}
inline int ImageEnhancementHong::getwOr() const {return orW;}
inline int ImageEnhancementHong::gethOr() const {return orH;}
inline bool **ImageEnhancementHong::getMasc() const {return masc;}
inline int ImageEnhancementHong::getW() const {return w;}
inline int ImageEnhancementHong::getH() const {return h;}

#endif	/* IMAGEENHANCEMENTHONG_H */

