/* 
 * \file        LinWangSegmentation.cpp
 * \author      Carlos Lopez Molina (carlos.lopez@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * Implementation of the LinWangSegmentation class
 *
 * 
 */


#include "LinWangSegmentation.h"
#include "MemUtil.hpp"
#include "MathUtil.h"
#include "ImgUtil.h"

LinWangSegmentation::LinWangSegmentation(int h, int w) {
    this->h = h;
    this->w = w;
}

LinWangSegmentation::LinWangSegmentation(const LinWangSegmentation& orig) {
}

LinWangSegmentation::~LinWangSegmentation() {
    
    if (this->energy>0){MemUtil::freeMatrix(&energy);}
    if (this->coherence>0){MemUtil::freeMatrix(&coherence);}
    if (this->segmentation>0){MemUtil::freeMatrix(&segmentation);}
    if (this->smoothSegmentation>0){MemUtil::freeMatrix(&smoothSegmentation);}
    if (this->tiledImage>0){MemUtil::freeMatrix(&tiledImage);}
    
}


void LinWangSegmentation::GaussianFiltering(float ***image, int h, int w, int gausFilterH, int gausFilterW, float sigma) {

    int x, y;
    double total = 0;
    float **gaussianFilter = MemUtil::create2DMatrix<float>(gausFilterH, gausFilterW);
    
    //gaussianFilter initialization
    for (int i = 0; i < gausFilterH; i++) {
        for (int j = 0; j < gausFilterW; j++) {
            x = i - (gausFilterH / 2);
            y = j - (gausFilterW / 2);
            gaussianFilter[i][j] = exp(-(x * x + y * y) / (2 * sigma * sigma));
            total += gaussianFilter[i][j];
        }
    }
    
    //Normalization of gaussian filter
    for (int i = 0; i < gausFilterH; i++) {
        for (int j = 0; j < gausFilterW; j++) {
            x = i - (gausFilterH / 2);
            y = j - (gausFilterW / 2);
            gaussianFilter[i][j] = gaussianFilter[i][j] / total;
        }
    }

    float **imgFiltered = ImgUtil::getFilteredImg<float>(*image, h, w, gaussianFilter, gausFilterH, gausFilterW);
    
    MemUtil::freeMatrix(&(*image));
    *image = imgFiltered;
    MemUtil::freeMatrix(&gaussianFilter);

    if (ImgUtil::DEBUG) {
        string fileName = "EnergyFilteredImage";
        ImgUtil::writeGrayScaleImage(*image, h, w, fileName, 1);
    }

    return;
}


void LinWangSegmentation::setEnergy(float **E){
    
    this->energy=MemUtil::copy2DMatrix<float>(E, h, w);
      
}

void LinWangSegmentation::setCoherence(float **C){
    
    this->coherence=MemUtil::copy2DMatrix<float>(C, h, w);
        
}

void LinWangSegmentation::createSegmentation(){
    
    if (ImgUtil::DEBUG){
    ImgUtil::writeGrayScaleImage(this->coherence,h,w,"LWSCoherence",255);    
        ImgUtil::writeGrayScaleImage(this->energy,h,w,"LWSEnergy",255);
        ImgUtil::writeGrayScaleImage(this->coherence,h,w,"LWSCoherence",255);
    }
    
    int *nHistogram=MemUtil::createVector<int>(100,0);//[0,0.01,...,1]
    float *nValues=MemUtil::createVector<float>(100,0);
    
    float maxEnergy=-1;
    float maxCoherence=-1;
    for (int i=0; i<this->h; i++){
        for (int j=0; j<this->w; j++){
            if (this->coherence[i][j]>maxCoherence){maxCoherence=this->coherence[i][j];}
            if (this->energy[i][j]>maxEnergy){maxEnergy=this->energy[i][j];}
        }
    }
    
    GaussianFiltering(&energy,this->h,this->w, 15,15,4);
    
    for (int i=0; i<this->h; i++){
        for (int j=0; j<this->w; j++){
            this->coherence[i][j]=this->coherence[i][j]/maxCoherence;
            this->energy[i][j]=this->energy[i][j]/maxEnergy;
        }
    }
    
    

    int pos=-1;
    for (int i=0; i<this->h; i++){
        //cout <<i<<endl;
        for (int j=0; j<this->w; j++){
            pos = floor(100*this->coherence[i][j]*this->energy[i][j]);
            if (pos == 100)
                pos--;
            nHistogram[pos]++;
        }
    }
    
    
    for (int i=0; i<100; i++){
        nValues[i]=i*0.01;
    }
    
    
    float thr=ImgUtil::otsuThresholding(nHistogram,nValues,100)*0.75;
    
    MemUtil::freeMatrix(&nHistogram);
    MemUtil::freeMatrix(&nValues);

    
    if (ImgUtil::DEBUG){cout << "Selected Threshold is : "<<thr<<endl;}
    
    
    segmentation = MemUtil::create2DMatrix<bool>(h, w);
    for (int i=0; i<this->h; i++){
        for (int j=0; j<this->w; j++){
            if (floor(100*this->coherence[i][j]*this->energy[i][j])>thr){
                this->segmentation[i][j]=true;
            }else{
                this->segmentation[i][j]=false;
            }
        }
    }
    
    if (ImgUtil::DEBUG){ImgUtil::writeGrayScaleImage(segmentation,h,w,"segmentation_initial");}
    
    if (ImgUtil::DEBUG){
        string fileName="recentSegmentation1";
        ImgUtil::writeGrayScaleImage(segmentation,h,w,fileName);
    }
    
}

bool **LinWangSegmentation::getSegmentation(){return this->segmentation;}
bool **LinWangSegmentation::getSmoothSegmentation(){return this->smoothSegmentation;}

bool **LinWangSegmentation::getBlockwiseSegmentation(int blockSize,bool isOverlapping){ 
    if (~isOverlapping){
        tileImage(this->segmentation,blockSize); 
        if (ImgUtil::DEBUG){ImgUtil::writeGrayScaleImage(tiledImage,floor(h/blockSize),floor(w/blockSize),"segmentation_tiled2");}
    }
    else{
        tileImageOverlapping(this->segmentation,blockSize); 
        if (ImgUtil::DEBUG){ImgUtil::writeGrayScaleImage(tiledImage,floor(h/blockSize)-1,floor(w/blockSize)-1,"segmentation_tiled2");}
    }
    return tiledImage;
}


bool **LinWangSegmentation::getBlockwiseSmothSegmentation(int blockSize,bool isOverlapping){ 
    
    if (ImgUtil::DEBUG){cout <<"Value is "<<isOverlapping<<endl;}
    if (isOverlapping){
        if (ImgUtil::DEBUG){cout <<"Doing the overlapping one "<<endl;}
        tileImageOverlapping(this->smoothSegmentation,blockSize); 
        if (ImgUtil::DEBUG){ImgUtil::writeGrayScaleImage(tiledImage,floor(h/blockSize)-1,floor(w/blockSize)-1,"segmentation_tiled2");}
    }
    else{
        tileImage(this->smoothSegmentation,blockSize); 
        if (ImgUtil::DEBUG){ImgUtil::writeGrayScaleImage(tiledImage,floor(h/blockSize),floor(w/blockSize),"segmentation_tiled2");}
    }
    return tiledImage;
}
 
void LinWangSegmentation::createSmoothSegmentation(int halfWindowSize){
    
    
    this->smoothSegmentation=MemUtil::create2DMatrix<bool>(this->h,this->w);
    
    int minNumTrues=(int)round((2*halfWindowSize+1)*(2*halfWindowSize+1)/2);
    
    //Smoothing segmentation
    for (int i=0; i<this->h; i++){
        
        //cout << i<<endl;
        int fromI=max(0,i-halfWindowSize);
        int toI=min(h-1,i+halfWindowSize);
        
        for (int j=0; j<this->w; j++){
            
            int fromJ=max(0,j-halfWindowSize);
            int toJ=min(w-1,j+halfWindowSize);
            
            int acc=0;
            
            for (int u=fromI; u<=toI;u++){
                for (int v=fromJ;v<=toJ; v++){
                    if (segmentation[u][v])
                        acc++;
                }
            }
            if (acc>=minNumTrues){
                this->smoothSegmentation[i][j]=true;
            }
            else{
                this->smoothSegmentation[i][j]=false;
            }
        }
    }
  
    
    if (ImgUtil::DEBUG){
        string fileName="recentSegmentation2";
        ImgUtil::writeGrayScaleImage(this->smoothSegmentation,h,w,fileName);
    }
}

void LinWangSegmentation::tileImageOverlapping(bool **originalImage,int blockSize){
    
    int orH=(h/blockSize)-1;
    int orW=(w/blockSize)-1;
    
    if (ImgUtil::DEBUG){cout<<"Creating a segmentation with orH,orW="<<orH<<orW<<endl;}
    
    this->tiledImage=MemUtil::create2DMatrix<bool>(orH,orW);
    
    int offset=(int)blockSize*0.5;
    
    for (int i=0; i<orH; i++){
        
        for (int j=0; j<orW; j++){
            
            int acc=0;
            
            for (int u=i*blockSize+offset; u<(i+1)*blockSize+offset; u++){
                for (int v=(j)*blockSize+offset; v<(j+1)*blockSize+offset; v++){
                    //cout << "[u][v]=["<<u<<"]["<<v<<"]"<<endl;
                    if (originalImage[u][v]){
                        acc++;
                    }
                }
            }
            
            if ((float)acc/(float)(blockSize*blockSize)>=0.5)
                this->tiledImage[i][j]=true;
            else
                this->tiledImage[i][j]=false;            
        }
        
    }
    
    if (ImgUtil::DEBUG){ImgUtil::writeGrayScaleImage(tiledImage,orH,orW,"segmentation_tiled");}
        
}

void LinWangSegmentation::tileImage(bool **originalImage,int blockSize){
    
    int orH=(h/blockSize);
    int orW=(w/blockSize);
    
    this->tiledImage=MemUtil::create2DMatrix<bool>(orH,orW);
    
    for (int i=0; i<orH; i++){
        
        for (int j=0; j<orW; j++){
            
            int acc=0;
            
            for (int u=(i)*blockSize; u<(i+1)*blockSize; u++){
                for (int v=(j)*blockSize; v<(j+1)*blockSize; v++){
                    //cout << "[u][v]=["<<u<<"]["<<v<<"]"<<endl;
                    if (originalImage[u][v]){
                        acc++;
                    }
                }
            }
            
            if ((float)acc/(float)(blockSize*blockSize)>=0.5)
                this->tiledImage[i][j]=true;
            else
                this->tiledImage[i][j]=false;            
        }
        
    }
    
    ImgUtil::writeGrayScaleImage(tiledImage,orH,orW,"segmentation_tiled");
        
}

