/* 
 * \file        MaskVertex.cpp
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * The implementation of DynamicMask (Capelli99) class
 *
 * 
 */

#include <iostream>
#include <cmath>

#include "MaskVertex.h"

MaskVertex::MaskVertex() {
    this->x = 0;
    this->y = 0;
    this->type = fix;
    
    this->maxDeltaX = 0;
    this->maxDeltaY = 0;
    
    deltaX = 0; 
    deltaY = 0;
    
    
    globalDeltaX = 0;
    globalDeltaY = 0;
    globalAng = 0;
    
    isBorder = true;
    isCorner = true;
    
    actualX = x;
    actualY = y;
    
    myReferencePoint = 0;
    globalActualX = actualX;
    globalActualY = actualY;
}

MaskVertex::MaskVertex(int x, int y, vertexType type, int maxDeltaX, int maxDeltaY) {
    this->x = x;
    this->y = y;
    this->type = type;
    
    this->maxDeltaX = maxDeltaX;
    this->maxDeltaY = maxDeltaY;

    deltaX = 0; 
    deltaY = 0;
    
    globalDeltaX = 0;
    globalDeltaY = 0;
    globalAng = 0;
    
    if ((x <= 0) || (y <= 0) || (x >= orW - 1) || (y >= orH -1)) {
        isBorder = true;
       // cout << "isBorder! " << x << ", " << y << endl;
    }
    else
        isBorder = false;
    
    if (( (x <= 0) && (y <= 0)  ) || ( (x <= 0) && (y >= orH - 1)  ) ||
            ( (x >= orW - 1) && (y <= 0)  ) || ( (x >= orW - 1) && (y >= orH - 1)  )) {
        isCorner = true;
        
        /* These displacements are needed to better perform the isInRegion operation, but they do not
         * affect anything else (neither the form of the mask, nor their movements
         */
        if (x <= 0)
            x -= 100;
        else 
            x += 100;
        if (y <= 0)
            y -= 100;
        else
            y += 100;
       // cout << "isCorner! " << x << ", " << y << " isCorner = " << isCorner << endl;
    }
    else
        isCorner = false;
        
    
    actualX = x;
    actualY = y;
    myReferencePoint = 0;
    globalActualX = actualX;
    globalActualY = actualY;
  //  cout << "Actuales al crear: " << actualX << ", " << actualY << endl;
}

MaskVertex::~MaskVertex() {
    
    myDependents.clear();
    myReferencePoint = 0;
    
}

int MaskVertex::addDependent(MaskVertex *p) {
    
    if (type != mobile) {
        cerr << "ERROR: This is not a mobile vertex!" << endl;
        return 0;
    }
    
    myDependents.push_back(p);
    p->maxDeltaX = maxDeltaX;
    p->maxDeltaY = maxDeltaY;
    
    
    return 1;    
}


int MaskVertex::orH = 28;
int MaskVertex::orW = 30;

void MaskVertex::setOrientationMapSize(int h, int w) {
    orH = h;
    orW = w;
}

int MaskVertex::getOrH() {
    return orH;
}
int MaskVertex::getOrW() {
    return orW;
}


void MaskVertex::alterGlobalRotationDisplacement(int deltaX, int deltaY, float ang) {
    
    if (!isCorner) {
        this->globalDeltaX = deltaX;
        this->globalDeltaY = deltaY; 

        actualX = x + globalDeltaX;
        actualY = y + globalDeltaY;


        setGlobalAng(ang);
        
        globalActualX = actualX;
        globalActualY = actualY;
    }
}


void MaskVertex::alterLocalDisplacement(int deltaX, int deltaY) {
    
        
    actualX = globalActualX;
    actualY = globalActualY;
    
    if (type == fix)    // Esquinas y fijos (esquinas son fijas!) QUIETOS PARAOS
        return;
    else if (type == mobile) {    // este es el caso primero, desplazamiento según el ángulo y llamamos a sus dependientes!!!
        this->deltaX = deltaX;
        this->deltaY = deltaY;
        // deltaX and deltaY are already centred on 0, 0 
        int deltaXwithAng = round(deltaX * cos(globalAng) - deltaY * sin(globalAng));
        int deltaYwithAng = round(deltaX * sin(globalAng) + deltaY * cos(globalAng));
        
        actualX += deltaXwithAng;
        actualY += deltaYwithAng;

        for (int i = 0; i < myDependents.size(); i++)
            myDependents[i]->alterLocalDisplacement(deltaX, deltaY);
    }
    else {      // Soy un dependiente!
        
        if (isBorder) { // desplazamiento en su eje, según el ángulo!!!!!!
            
            // solo puede ser dependentX o dependentY pero no las dos porque no sería borde!
            if (type == dependentX) 
                this->deltaX = deltaX;
            else 
                this->deltaY = deltaY;
            
           
            
        }
        else { // desplazamiento completo IGUAL que el padre!!
            this->deltaX = deltaX;
            this->deltaY = deltaY;
        }
        int deltaXwithAng = round(this->deltaX * cos(globalAng) - this->deltaY * sin(globalAng));
        int deltaYwithAng = round(this->deltaX * sin(globalAng) + this->deltaY * cos(globalAng));

        actualX += deltaXwithAng;
        actualY += deltaYwithAng;
        
            
    }
    
    // en ambos casos es necesario corregirlo respecto a su punto de referencia
    correctVertexPosition();
}

void MaskVertex::setGlobalAng(float ang) { 
    
    if (isCorner)       // Por si acaso, con el uso normal no se entra aquí
        return; 
    //if (!isBorder) {
        this->globalAng = ang; 
    
      //  cout << "Actuales ang: " << actualX << ", " << actualY << endl;
        
        int halfOrH = (orH - 1) / 2;
        int halfOrW = (orW - 1) / 2;

        int xMod = actualX - halfOrW;
        int yMod = actualY - halfOrH;

        int xprima = round(xMod * cos(globalAng) - yMod * sin(globalAng));
        int yprima = round(xMod * sin(globalAng) + yMod * cos(globalAng));

        xprima += halfOrW;
        yprima += halfOrH;

        actualX = xprima;
        actualY = yprima;
        
     //   cout << "Rotated ang: " << actualX << ", " << actualY << endl;
        
}

void MaskVertex::correctVertexPosition(bool global) {
    
        // Si no tengo punto de referencia he terminado, me quedo donde estoy SOY ESQUINA O INTERIOR, SI SOY ESQUINA ME QUEDO FIJO Y SINO DONDE ESTÉ :D
        // Sino quiere decir que debo ajustarme al borde con la recta que formo con dicho punto!
        
        if (myReferencePoint != 0) {
            
       //     cout << "Inicial: " << x << ", " << y << endl;
      //      cout << "Rotated: " << actualX << ", " << actualY << endl;
      //      cout << "My ReferencePoint: " << myReferencePoint->actualX << ", " << myReferencePoint->actualY << endl;
            
            int originalX = actualX;
            int originalY = actualY;
            
            float refX = myReferencePoint->actualX;
            float refY = myReferencePoint->actualY;
            
            // obtener vector del punto de referencia al propio punto
            float u = actualX - refX;
            float v = actualY - refY;
            //    cout << "vector ("<<u << ", " << v << ") "<< endl;

            float a = v / u; // a = vNormalized
            // float uNormalized = u / u; u = 1
            //    cout << "vector ("<< 1 << ", " << a << ") "<< endl;

            // cout << "recta: y = "<< a << "x + b" <<  endl;
            float b = a * (0 - refX) + refY;
            //    cout << "recta: y = "<< a << "x + " << b <<  endl;

            if ((u == 0) && (v == 0)) {
                // pasar        NUNCA DEBERÍA OCURRIR
            }
            else if (u == 0) {
                // solo buscar la intersección con la y = 0 e y = hOr - 1 (con la más cercana de las dos)
                if (actualY < abs( (orH - 1) - actualY)) 
                    actualY = 0;
                else
                    actualY = orH - 1;
                
                // actualX = actualX; se mantiene!
            }
            else if (v == 0) {
                // solo buscar la intersección con la x = 0 e x = wOr - 1 (con la más cercana de las dos)
                if (actualX < abs( (orW - 1) - actualX))
                    actualX = 0;
                else
                    actualX = orW - 1;
                // actualY = actualY; se mantiene!
            }
            else {
                float xW, yW, xH, yH, dW, dH;
                if (u < 0)      // intersección con x = 0
                    xW = 0;
                else            // intersección con x = wOr - 1
                    xW = orW - 1;
                
                yW = a * xW + b;

                if (v < 0)    // intersección con y = 0
                    yH = 0;
                else         // intersección con y = 0
                    yH = orH - 1;
                
                xH = (yH - b) / a;

                dW = sqrt( (xW - refX) * (xW - refX) + (yW - refY) * (yW - refY));
                dH = sqrt( (xH - refX) * (xH - refX) + (yH - refY) * (yH - refY));

                if (dW <= dH) {
                    actualX = round(xW);
                    actualY = round(yW);
                }
                else {
                    actualX = round(xH);
                    actualY = round(yH);
                }

            }
            if ((actualX == refX) && (actualY == refY)) {
                actualX = originalX;
                actualY = originalY;
            }
            if (((actualX - refX) * (originalX - refX) < 0 ) && ((actualY - refY) * (originalY - refY) < 0)) {
                //cout << "actual = " << actualX << ", " << actualY << " original = " << originalX << ", " << originalY << " Referencia = " << refX << ", " << refY << endl;
                actualX = originalX;
                actualY = originalY;
            }
            
           // cout << "Veamos : " << x << ", " << y << " rotated end = " << actualX << ", " << actualY << endl;
        }
        else { // si no tiene referencia pero se va fuera, meterlo como sea
            
        }
        
  
        if (global) {
            
            globalActualX = actualX;
            globalActualY = actualY;
        }
        
}


void MaskVertex::print() {
    
    cout << "(" << actualX << ", " << actualY << ")";
    
}