/* 
 * \file        NewPoincareSingularPoints.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * Header file for NewPoincareSingularPoints class
 *
 * 
 */

#ifndef NEWPOINCARESINGULARPOINTS_H
#define	NEWPOINCARESINGULARPOINTS_H

#include <vector>
#include "PCASYSManager.h"

using namespace std;

enum operationModePoincare {increasingSmoothPoincare = 0, decreasingCurvePoincare = 1};

/**
 * \class NewPoincareSingularPoints
 *
 * The NewPoincareSingularPoints class implements the poincare method
 */
class NewPoincareSingularPoints {
public:
    
    /** Default constructor */
    NewPoincareSingularPoints();
    
    /**
     * Basic constructor
     * \param image     Grayscale fingerprint image
     * \param w         width of the image (in pixels)
     * \param h         height of the image (in pixels)
     */
    NewPoincareSingularPoints(float **image, int w, int h);
    
    /**
     * Advanced constructor using an orientation image
     * \param O         Orientation image
     * \param xCos      x (cosine) components of the orientations
     * \param ySin      y (sine) components of the orientations
     * \param orH       Height of the orientataion image
     * \param orW       Width of the orientation image
     */
    NewPoincareSingularPoints(float **O, float **xCos, float **ySin, int orH, int orW);
    
    
    
    /** Default destructor */
    ~NewPoincareSingularPoints();
   
    /**
     * It initializes the image, width and height.
     * \param image original image
     * \param h image's height 
     * \param w image's width
     */
    void initialize(float **image, int h, int w);
    
    /**
     * Executes the method Poincare Singular Points from the fingerprint in file
     * \param file fingerprint
     * \param pcasys pcasys manager
     */
    void run(string file, PCASYSManager *pcasys);
    
    
    int runReject(string file, PCASYSManager *pcasys);

    /**
     * 
     * \param auxVector
     * \param covMargins
     * \param w
     * \param h
     * \param bSize
     * \return 
     */
    int *selectCenterInBlock(int *auxVector, int *covMargins, int w, int h, int bSize);
    
    /**
     * Compute all the cores and deltas using the Poincare Method with a curve of 
     *      curveW x curveW positions
     * \param curveW        Side of the curve region
     */
    void computeAllCoresAndDeltas(int curveW);

    /**
     * Computes all the cores and deltas. At each iteration of the Poincare method
     *      the smoothing of the orientations is increased. It starts from the extraction of the orientation image
     * \param curveW        Window size of the curve used for the computation of the Poincare index.
     * \param smoothW       Window size used in the orientations smoothing.
     * \param mode          The operation mode increasingSmooth/decreasingCurve
     * \param appliedRules  Special rules that might apply. (0=none, 1=karu, 2=hong,default 0)
     * \param normalize     Whether the orientation vectors must be normalized or not (default true).
     */
    int computeCoresAndDeltasFromImg(int curveW, int smoothW = 0, operationModePoincare mode = increasingSmoothPoincare, int appliedRules=0, int minNBlock = 0, bool normalize= true);
    
    /**
     * Computes all the cores and deltas. At each iteration of the Poincare method
     *      the smoothing of the orientations is increased (i.e. they are re-smoothed).
     * \param curveW        Window size of the curve used for the computation of the Poincare index.
     * \param smoothW       Window size used in the orientations smoothing.
     * \param mode          The operation mode increasingSmooth/decreasingCurve
     * \param specialRules  Special rules that might apply. (0=none, 1=karu, 2=hong,default 0)
     * \param normalize     Whether the orientation vectors must be normalized or not (default true).
     * \return              The numer of times the orientation image had to be smoothed.
     */
    int computeCoresAndDeltas(int curveW, int smoothW, operationModePoincare mode, int appliedRules=0, int minNBlock = 0, bool normalize= true);
    
    /**
     * Constructs the orientation image.
     * \param blockW        Size of the orientations block used to construct each tile of the
     *      orientation image.
     */
    void createOrientationImage(int blockW);
    
    /**
     * Segments the image based upon the block mean and variance.
     * \param image     Grayscale image
     * \param h         Height of the image
     * \param w         Width of the image
     */
    void generateSegmentation(float** image, int h, int w);
    
    /**
     * It segments the image (it uses the image from the current object)
     */
    float generateSegmentation();
    
    /**
     * Computes the poincare index of the current orientation image with a curve
     *      of size curveW x curveW.
     * \param curveW        Size of the window used for the local computation of the poincare index.
     */
    void computePoincareIndex(int curveW);
    
    /**
     * Combines the segmentation block created from Poincare with the segmentation block parameterized
     * \param segmentation
     */
    void combineWithSegmentation(bool **segmentation);
    
    
    /**
     * Removes all the candidate blocks outside of the segmented (valid) region of the image.
     */
    void deleteLabelsOutsideSegmentedRegion();
    
    /**
     * Labels the cores after the Poincare index are computed
     * \return      The number of core regions (note that each region might contain more than 1 block).
     */
    int labelCores();
    
    /**
     * Labels the deltas after the Poincare index are computed
     * \return      The number of core regions (note that each region might contain more than 1 block).
     */
    int labelDeltas();
    
    /*
     * 
     */
    
    
    /**
     * Constructs the core region connected to a given position of the poincare index image.
     * \param posI  Vertical position from which the region starts expanding
     * \param posJ  Horizontal position from which the region starts expanding
     * \param label Label to be assigned to the region
     * \param coreI Accumulator to compute the vertical position of the centroid of the region.
     * \param coreJ Accumulator to compute the horizontal position of the centroid of the region.
     * \return The number of elements found in the region
     */
    int fillCoreRegion(int posI, int posJ, int label, int *coreI, int *coreJ);
    
    /**
     * Constructs the delta region connected to a given position of the poincare index image.
     * \param posI  Vertical position from which the region starts expanding
     * \param posJ  Horizontal position from which the region starts expanding
     * \param label Label to be assigned to the region
     * \param coreI Accumulator to compute the vertical position of the centroid of the region.
     * \param coreJ Accumulator to compute the horizontal position of the centroid of the region.
     * \return The number of elements found in the region
     */
    int fillDeltaRegion(int posI,int posJ,int label, int *deltaI, int *deltaJ);
    
    /**
     * Set a segmentation created outsise the NewPoincareSingularPoints class
     * \param blockSegmentation     Mask where the active (valid) positions are marked with 'true'
     */
    void setBlockSegmentation(bool **blockSegmentation);
    
    /**
     * Remove from the cores (and deltas) list the invalid singular points,
     *      according to the rules given by Karu and Jain in 
     * 
     * Kalle Karu, Anil K. Jain 
     * Fingerprint classification. 
     * Pattern Recognition 29(3), 389-404 (1996)
     * 
     */
    void cleanSPsWithKaruRules();
    
        
    /**
     * Remove from the cores (and deltas) list the invalid singular points,
     *      according to the rules given by Hong et al. in
     * 
     * Jin-Hyuk Hong, Jun-Ki Min, Ung-Keun Cho, Sung-Bae Cho 
     * Fingerprint classification using one-vs-all support vector machines dynamically ordered with naive Bayes classifiers. 
     * Pattern Recognition 41(2), 662-671 (2008)
     * 
     */
    void cleanSPsWithHongRules(int minNBlock);
    
    /**
     * Sets which are the SP-removal rules to be applied depending on a numerical code
     *  0 = None
     *  1 = Karu rules
     *  2 = Hong rules
     *  3 = First Karu rules and then Hong rules
     * \param ruleCode  Numerical code of the rule
     */
    void processSpecialRules(int ruleCode);
    
    /**
     *      Sets the orientation map 
     * \param O         Orientation image
     * \param xCos      x (cosine) components of the orientations
     * \param ySin      y (sine) components of the orientations
     * \param mapH       Height of the orientataion image
     * \param mapW       Width of the orientation image
     */
    void setOrientationMap(float **O, float **xCos, float **ySin, int mapH, int mapW);
    
    void cleanSPsMinBlock(int minNBlock);
    
    /**
     *  ImgUtil::DEBUGGING FUNCTION ONLY! 
     */
    void fakeCoresAndDeltas();
    
    /**
     * Deletes the cores or deltas joined together in a same cluster
     * \param minDistance minimum distance between two singular points to belong a same cluster
     */
    void cleanSPsWithClusterRules(int minDistance);
    
    /**
     * Gets the number of blocks each core region is composed of.
     * \return the vector containing all the sizes
     */
    vector<int> getBlocksPerCore();
    
    /**
     * Gets the number of blocks each delta region is composed of.
     * \return the vector containing all the sizes
     */
    vector<int> getBlocksPerDelta();
    
    /**
     * Gets the centroids of the cores of the image. Each even position is a vertical component, while
     *      each odd one is a horizontal one. I.e. the coordinates of the first core are
     *      the positions 0 (vertical) and 1 (horizontal), the coordinates of the second one
     *      are in the positions 2 and 3, and so on.
     * The coordinates are given IN TERMS OF BLOCKS, NOT PIXELS
     * \return the vector containing all the sizes
     */
    vector<float> getCoreCentroids();

    /**
     * Gets the centroids of the deltas of the image. Each even position is a vertical component, while
     *      each odd one is a horizontal one. I.e. the coordinates of the first delta are
     *      the positions 0 (vertical) and 1 (horizontal), the coordinates of the second one
     *      are in the positions 2 and 3, and so on.
     * The coordinates are given IN TERMS OF BLOCKS, NOT PIXELS
     * \return the vector containing all the sizes
     */
    vector<float> getDeltaCentroids();
   
    
    /**
     * Gets the orientation map
     * \return  Orientation map
     */
    float **getOrientationMap()const;
    
    /**
     * Gets the horizontal (cosine) components of the orientation map
     * \return The horizontal (cosine) components of the orientation map
     */
    float **getXCos()const;
    
    /**
     * Gets the vertical (sine) components of the orientation map
     * \return The vertical (sine) components of the orientation map
     */
    float **getYSin()const;
    
    /**
     * Gets the size of the orientation blocks (in pixels).
     * CAN BE CALLED IF AND ONLY IF THE CONSTRUCTOR (float **, int, int) was used
     * \return the size of such blocks
     */
    int getBlockSize();
    
    /**
     * Gets the number of cores detected in the image
     * \return The number of cores detected in the image
     */
    int getnCores();
    
    /**
     * Gets the number of deltas detected in the image
     * \return The number of deltas detected in the image
     */
    int getnDeltas();
    
    /**
     * Reasigns the image's width
     * \param width The width of the image
     */
    void setWidth(int width);
    
    /**
     * Reasigns the image's height
     * \param height The height of the image
     */
    void setHeight(int height);
    
    /**
     * Retrieves the current segmentation
     * @return the current segmentation
     */
    bool **getSegmentation() const;
    
    /**
     * It set the width of the orientation map
     * \param orW The width of the orientation map
     */
    void setOrW(int orW);
    
    /**
     * It set the height of the orientation map
     * \param orH The height of the orientation map
     */
    void setOrH(int orH);
    
    /**
     * It returns the height of the orientatation map
     * \return The height of the orientatation map
     */
    int getOrH() const;
    
    /**
     * It returns the width of the orientatation map
     * \return The width of the orientatation map
     */
    int getOrW() const;    
private:
    float **img;                    //!< Gray scale image of the fingerprint
    float **imgNormalized;          //!< Normalized image using Hong's method, it is needed o segment the images
    int w;                          //!< Width of the image
    int h;                          //!< Height of the image

    float **gradX;                  //!< Horizontal gradients of the image
    float **gradY;                  //!< Vertical gradients of the image
    float **gradMagnitude;          //!< Magnitude of the gradients

    float **O;                      //!< Orientation map extracted from the image (or directly introduced in the constructor)
    float **xCos;                   //!< x-component of the orientation map
    float **ySin;                   //!< y-component of the orientation map

    int orW;                        //!< Width of the orientation map
    int orH;                        //!< Height of the orientation map

    int **labelImg;                 //!< Labeling of each block in the orientaion map used in the processing of the connected components

    bool **blockSegmentation;       //!< Segmentation of the image in blocks (size of the orientation map)

    float **blockMean;              //!< Mean intensity of the pixels in each block
    float **blockVariance;          //!< Variance of the intensity of the pixels in each block

    float **poincareIndex;          //!< The poincaré index in each block of the orientation map

    bool karuRules;                 //!< Whether to use Karu rules to delete cores/deltas
    bool hongRules;                 //!< Whether to use Hong rules to delete core/deltas
    bool clusterRules;              //!< Whether to use nearest neighbor clustering to delete core/deltas

    vector<int> blocksPerCore;      //!< Number of blocks/components in each core found
    vector<float> coreCentroids;    //!< The centroids of the core found (x1,y1,x2,y2...)

    vector<int> blocksPerDelta;     //!< The centroids of the delta found (x1,y1,x2,y2...)
    vector<float> deltaCentroids;   //!< Number of blocks/components in each delta found

    operationModePoincare mode;     //!< The operation procedure of the poincare method increasingSmooothing/decreasingCurve    
};

inline vector<int> NewPoincareSingularPoints::getBlocksPerCore() { return  blocksPerCore; }
inline vector<int> NewPoincareSingularPoints::getBlocksPerDelta() { return blocksPerDelta; }
inline vector<float> NewPoincareSingularPoints::getCoreCentroids() { return coreCentroids; }
inline vector<float> NewPoincareSingularPoints::getDeltaCentroids() { return deltaCentroids; }

inline float **NewPoincareSingularPoints::getOrientationMap()const {return O;} 
inline float **NewPoincareSingularPoints::getXCos()const {return xCos;}
inline float **NewPoincareSingularPoints::getYSin()const {return ySin;}


inline int NewPoincareSingularPoints::getOrH()const {return orH;}
inline int NewPoincareSingularPoints::getOrW()const {return orW;}

inline void NewPoincareSingularPoints::setOrW(int orW) { this->orW = orW; }
inline void NewPoincareSingularPoints::setOrH(int orH) { this->orH = orH; }

inline int NewPoincareSingularPoints::getBlockSize(){return (h / orH);}

inline int NewPoincareSingularPoints::getnCores() { return blocksPerCore.size(); }
inline int NewPoincareSingularPoints::getnDeltas() { return blocksPerDelta.size(); }
inline void NewPoincareSingularPoints::setWidth(int width) { w = width; }
inline void NewPoincareSingularPoints::setHeight(int height) { h = height; }

inline bool **NewPoincareSingularPoints::getSegmentation() const{return this->blockSegmentation;}

#endif	/* POINCARESINGULARPOINTS_H */

