/* 
 * \file        ImageEnhancementCappelli.cpp
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \author      Carlos Lopez Molina (carlos.lopez@unavarra.es)
 * \version     1.0
 * 
 * \section     Orientation enhancement performed by Cappelli et al. on the
 *                  orientations map extracted by PCASYS
 * 
 * Implementation of the ImageEnhancementCappelli class
 *
 * 
 */

#include "OrientationEnhancementCappelli.h"
#include "ImgUtil.h"

OrientationEnhancementCappelli::OrientationEnhancementCappelli() {
}

OrientationEnhancementCappelli::~OrientationEnhancementCappelli() {

    MemUtil::freeMatrix(&this->atr);
    MemUtil::freeMatrix(&this->enhancedXCos);
    MemUtil::freeMatrix(&this->enhancedYSin);
    MemUtil::freeMatrix(&this->smoothSTR);
    MemUtil::freeMatrix(&this->xCos);
    MemUtil::freeMatrix(&this->ySin);

}

OrientationEnhancementCappelli::OrientationEnhancementCappelli(float **xCosImage, float **ySinImage, int orImageH, int orImageW) {

    //Copying the dimensions
    this->orH = orImageH;
    this->orW = orImageW;

    //Declaring the vars
    xCos = MemUtil::create2DMatrix<float>(orH, orW);
    ySin = MemUtil::create2DMatrix<float>(orH, orW);

    //Copying the data
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {
            xCos[i][j] = (float) xCosImage[i][j];
            ySin[i][j] = (float) ySinImage[i][j];
        }
    }


}

void OrientationEnhancementCappelli::enhance(float sigma, float gamma) {

    if (ImgUtil::DEBUG) {
        string fileName = "./pruebaOriginalOr";
        ImgUtil::writeOrImg(xCos, ySin, orW, orH, fileName);
    }

    //Smooth the orientations (overwrites the original data, which is locally copied in the constructor)
    ImgUtil::smoothOrientations(0, &xCos, &ySin, orH, orW, 3, false);

    //Calculate attenuation (atr) and strengthening rates (str)
    calculateAttenuation(sigma);
    calculateStrenghthening(1, 1);
    this->smoothSTR = smooth(str, orH, orW, 1, 1);

    //Declare the enhanced cos/sin
    this->enhancedXCos = MemUtil::create2DMatrix<float>(orH, orW);
    this->enhancedYSin = MemUtil::create2DMatrix<float>(orH, orW);

    //Fill it in
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {
            this->enhancedXCos[i][j] = xCos[i][j] * (1 + gamma * atr[i][j] * smoothSTR[i][j]);
            this->enhancedYSin[i][j] = ySin[i][j] * (1 + gamma * atr[i][j] * smoothSTR[i][j]);
        }
    }

    if (ImgUtil::DEBUG) {

        string fileName2 = "./pruebaSmoothOr";
        ImgUtil::writeOrImg(xCos, ySin, orW, orH, fileName2);

        string fileName3 = "./pruebaEnhancedOr";
        ImgUtil::writeOrImg(enhancedXCos, enhancedYSin, orW, orH, fileName3);

        string fileName4 = "./pruebaSTR";
        ImgUtil::writeGrayScaleImage(smoothSTR, orH, orW, fileName4, 255);

        string fileName5 = "./pruebaATT";
        ImgUtil::writeGrayScaleImage(atr, orH, orW, fileName5, 5000);//112);
    }

}

float **OrientationEnhancementCappelli::smooth(float **data, int sizeH, int sizeW, int rowHalfSize, int colHalfSize) {

    //Local version (which is later returned)
    float **smoothData = MemUtil::create2DMatrix<float>(sizeH, sizeW);

    for (int i = 0; i < sizeH; i++) {

        int fromI = max(0, i - rowHalfSize);
        int toI = min(sizeH - 1, i + rowHalfSize);

        for (int j = 0; j < sizeW; j++) {

            int fromJ = max(0, j - colHalfSize);
            int toJ = min(sizeW - 1, j + colHalfSize);

            float acc = 0;

            for (int u = fromI; u <= toI; u++) {
                for (int v = fromJ; v <= toJ; v++) {

                    acc += data[u][v];
                }
            }


            smoothData[i][j] = acc / (float) ((1 + toJ - fromJ)*(1 + toI - fromI));

        }
    }

    return smoothData;
}

void OrientationEnhancementCappelli::calculateAttenuation(float sigma) {

    //Allocating memory for the attenuation rate map
    this->atr = MemUtil::create2DMatrix<float>(orH, orW);

    //Center of the image
    float centerH = orH / 2.;
    float centerW = orW / 2.;

    //Constants all over the image in the Gaussian function
    float outerFraction = 1.0 / (sqrt(2. * PI) * sigma);
    float twoSigmaSquare = 2. * (sigma * sigma);


    //Filling it in
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {

            float dist = sqrt((i - centerH) * (i - centerH) + (j - centerW)*(j - centerW));
            float distSquare = dist * dist;
            this->atr[i][j] = outerFraction * exp(-(distSquare / twoSigmaSquare));
        }
    }

    return;

}

void OrientationEnhancementCappelli::calculateStrenghthening(int rowHalfSize, int colHalfSize) {

    this->str = MemUtil::create2DMatrix<float>(orH, orW);
    float **compositeMagnitudes = MemUtil::create2DMatrix<float>(orH, orW); //Sum vectors -> magnitude
    float **sumOfMagnitudes = MemUtil::create2DMatrix<float>(orH, orW); //Sum magnitudes


    for (int i = 0; i < orH; i++) {

        int fromI = max(0, i - rowHalfSize);
        int toI = min(orH - 1, i + rowHalfSize);

        for (int j = 0; j < orW; j++) {

            int fromJ = max(0, j - colHalfSize);
            int toJ = min(orW - 1, j + colHalfSize);

            float accX = 0;
            float accY = 0;
            float accMagnitudes = 0;

            for (int u = fromI; u <= toI; u++) {
                for (int v = fromJ; v <= toJ; v++) {

                    accX += xCos[u][v];
                    accY += ySin[u][v];
                    accMagnitudes += sqrt((xCos[u][v] * xCos[u][v]) + (ySin[u][v] * ySin[u][v]));

                }
            }

            compositeMagnitudes[i][j] = sqrt((accX * accX) + (accY * accY));
            sumOfMagnitudes[i][j] = accMagnitudes;

            if (sumOfMagnitudes[i][j] == 0)
                this->str[i][j] = 0;
            else
                this->str[i][j] = 1 - (compositeMagnitudes[i][j] / sumOfMagnitudes[i][j]);

        }
    }

}

float **OrientationEnhancementCappelli::getSin2Theta() {
    return this->enhancedYSin;
}

float **OrientationEnhancementCappelli::getcos2Theta() {
    return this->enhancedXCos;

}


