/* 
 * \file        OrientationFieldMei.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * The implementation of OrientationFieldMei (Mei2009) class
 *
 * 
 */


#include "ImgUtil.h"
#include "MemUtil.hpp"

#include "OrientationFieldMei.h"
#include "ParametersLe.h"


OrientationFieldMei::OrientationFieldMei() {
    h = -1;
    w = -1;
    img = 0;
    Gx = 0;
    Gy = 0;
    Gmagnitude = 0;
    gx = 0;
    gy = 0;
    gmagnitude = 0;
    gsx = 0;
    gsy = 0;
    O = 0;
    orH = 0;
    orW = 0;
    blockW = 0;
    GBx = 0;
    GBy = 0;
    variance = 0;
    combinedO = 0;
    combinedGBx = 0;
    combinedGBy = 0;
}


OrientationFieldMei::OrientationFieldMei(unsigned char **image, int h, int w) {
    
    img = MemUtil::create2DMatrix<float>(h,w);
    // img initialition
    for (int i = 0; i < h; i++) {
        for (int j = 0; j < w; j++)
            img[i][j] = (float) image[i][j];
    }
    
    this->h = h;
    this->w = w;
    Gx = 0;
    Gy = 0;
    Gmagnitude = 0;
    gx = 0;
    gy = 0;
    gmagnitude = 0;
    gsx = 0;
    gsy = 0;
    O = 0;
    orH = 0;
    orW = 0;
    blockW = 0;
    GBx = 0;
    GBy = 0;
    variance = 0;
    combinedO = 0;
    combinedGBx = 0;
    combinedGBy = 0;
}


OrientationFieldMei::~OrientationFieldMei() {
   
    if (O != 0) {
        for (int i = 0; i < 3; i++) {
            MemUtil::freeMatrix(&O[i]);
        }
        delete[] O;
    }
    if (GBx != 0) {
        for (int i = 0; i < 3; i++) {
            MemUtil::freeMatrix(&GBx[i]);
        }
        delete[] GBx;
    }
    if (GBy != 0) {
        for (int i = 0; i < 3; i++) {
            MemUtil::freeMatrix(&GBy[i]);
        }
        delete[] GBy;
    }
    if (variance != 0) {
        for (int i = 0; i < 3; i++) {
            MemUtil::freeMatrix(&variance[i]);
        }
        delete[] variance;
    }

    MemUtil::freeMatrix(&img);
    
    MemUtil::freeMatrix(&Gx);
    
    MemUtil::freeMatrix(&Gy);
    
    MemUtil::freeMatrix(&Gmagnitude);

    MemUtil::freeMatrix(&gx);
    
    MemUtil::freeMatrix(&gy);
    
    MemUtil::freeMatrix(&gsx);
    
    MemUtil::freeMatrix(&gsy);
        
    MemUtil::freeMatrix(&combinedO);
    
    MemUtil::freeMatrix(&combinedGBx);
    
    MemUtil::freeMatrix(&combinedGBy);
    
    MemUtil::freeMatrix(&orH);
    
    MemUtil::freeMatrix(&orW);
    
    MemUtil::freeMatrix(&blockW); 
}

float **OrientationFieldMei::computeOrientationField(int *W, float *varThreshold) {
    
    computeGradients();
    computeFirstLevelOrientationFields(W, 0.5, varThreshold);
    combineFirstLevelOrientationFileds();
    predictNonEstimatedAreas();
        
}

void OrientationFieldMei::computeGradients() {
    
    Gx = MemUtil::create2DMatrix<float>(h,w);
    
    Gy = MemUtil::create2DMatrix<float>(h,w);
    
    Gmagnitude = MemUtil::create2DMatrix<float>(h,w);

    gx = MemUtil::create2DMatrix<float>(h,w);

    gy = MemUtil::create2DMatrix<float>(h,w);

    gsx = MemUtil::create2DMatrix<float>(h,w);
    
    gsy = MemUtil::create2DMatrix<float>(h,w);
   
    /* compute SOBEL */
    for (int i = 1; i < h - 1; i++) {
        for (int j = 1; j < w - 1; j++) {
            Gx[i][j] = (img[i][j + 1] - img[i][j - 1]) ;
            
            Gy[i][j] = -(img[i + 1][j] - img[i - 1][j]) ;
            
            Gmagnitude[i][j] = sqrt(Gx[i][j] * Gx[i][j] + Gy[i][j] * Gy[i][j]); // this is r in the paper
            
            /* If they are non-zero vectors, normalize them, otherwise set to 0 both components*/
            if (Gmagnitude[i][j] != 0) {
                gx[i][j] = Gx[i][j] / Gmagnitude[i][j];
                gy[i][j] = Gy[i][j] / Gmagnitude[i][j];
            }
            else {
                gx[i][j] = 0;
                gy[i][j] = 0;
            }
            
            /* Square the normalized gradients */
            
            gsx[i][j] = (gx[i][j] * gx[i][j] - gy[i][j] * gy[i][j]);
            gsy[i][j] = 2.0 * gx[i][j] * gy[i][j];    // cuidado que cambian los nombres...
            
        }
    }
    
    
    if (ImgUtil::DEBUG) {
        string fileName2 = "GradX";
        ImgUtil::writeGrayScaleImage(Gx, h, w, fileName2, 1);
        fileName2 = "GradY";
        ImgUtil::writeGrayScaleImage(Gy, h, w, fileName2, 1);
        fileName2 = "Grad";
        ImgUtil::writeGrayScaleImage(Gmagnitude, h, w, fileName2, 1);
    }
}



void OrientationFieldMei::computeFirstLevelOrientationFields(int *W, float coherenceThreshold, float *varThreshold) {
    
    O = MemUtil::createVector<float**>(3);    
    GBx = MemUtil::createVector<float**>(3);
    GBy = MemUtil::createVector<float**>(3);
    
    variance = new float**[3];
    
    blockW = MemUtil::createVector<float>(3);
    orH = MemUtil::createVector<float>(3);
    orW = MemUtil::createVector<float>(3);

    for (int i = 0; i < 3; i++) {
        blockW[i] = W[i];
        orH[i] = h / W[i];
        orW[i] = w / W[i];
        O[i] = computeFirstLevelOrientationField(W[i], coherenceThreshold, &GBx[i], &GBy[i]);
        
        variance[i] = computeVarianceAndSmooth_NonSmoothAreas(W[i], O[i], varThreshold[i]);
    }
}

float **OrientationFieldMei::computeFirstLevelOrientationField(int W, float coherenceThreshold, float ***GBxNew, float ***GByNew) {
    
    int orH = h / W;
    int orW = w / W;
    float **O = MemUtil::create2DMatrix<float>(orH, orW);
    float **GBx = MemUtil::create2DMatrix<float>(orH, orW);
    float **GBy = MemUtil::create2DMatrix<float>(orH, orW);
    *GBxNew = MemUtil::create2DMatrix<float>(orH, orW);
    *GByNew = MemUtil::create2DMatrix<float>(orH, orW);
    float **coherence = MemUtil::create2DMatrix<float>(orH, orW);
    
    float nNonZero = (float)(W * W) / 2.;
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {
            int N = 0;
            for (int u = i * W; u < (i + 1) * W; u++) {
                for (int v = j * W; v < (j + 1) * W; v++) {
                    
                    if (Gmagnitude[u][v] > minMagnitudeLeMei) {        // se usa como segmentación... OJO!!!
                        GBx[i][j] += gsx[u][v];
                        GBy[i][j] += gsy[u][v];
                        N++;
                        coherence[i][j] += sqrt(gsx[u][v] * gsx[u][v] + gsy[u][v] * gsy[u][v]);
                    }
                    
                }
            }
            if (N > nNonZero) {
                coherence[i][j] = sqrt(GBx[i][j] * GBx[i][j] + GBy[i][j] * GBy[i][j]) / coherence[i][j];
                
                GBx[i][j] /= (float) N;
                GBy[i][j] /= (float) N;
            }
            else {
                GBx[i][j] = 0;
                GBy[i][j] = 0;
                coherence[i][j] = 0;
            }
            
           /* if (coherence[i][j] < coherenceThreshold) {
                GBx[i][j] = 0;
                GBy[i][j] = 0;
                coherence[i][j] = 0;
                
            }*/
            
        }
    }
    
    float **xCos = MemUtil::create2DMatrix<float>(orH, orW);
    float **ySin = MemUtil::create2DMatrix<float>(orH, orW);
    
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {
            float GBxAux = 0;
            float GByAux = 0;
            if (coherence[i][j] < coherenceThreshold) {
                int N = 0;
                for (int u = max(0, i - 1); u < min(orH, i + 2); u++) {
                    for (int v = max(0, j - 1); v < min(orW, j + 2); v++) {
                //        if ((coherence[u][v] >= coherenceThreshold) && !((u == i) && (v == j)) ) {
                        if ((!((u == i) && (v == j))) && (coherence[i][j] != 0) && !((GBx[u][v] == 0) && (GBy[u][v] == 0))) {
                            GBxAux += GBx[u][v];
                            GByAux += GBy[u][v];
                            N++;
                        }
                    }
                }
                if (N != 0) {
                    GBxAux /= (float) N;
                    GByAux /= (float) N;
                }
                else {
                    GBxAux = 0;
                    GByAux = 0;
                }
            }
            else 
            {
                GBxAux = GBx[i][j];
                GByAux = GBy[i][j];
            }
            
            (*GBxNew)[i][j] = GBxAux;
            (*GByNew)[i][j] = GByAux;
            if (!((GBxAux == 0) && (GByAux == 0))) {
                float ang =  0;
                if (GBxAux == 0) {
                    if (GByAux > 0) 
                        ang = PI / 2.;
                    else
                        ang = - PI / 2.;
              //      cout << "GBx es cero! y atan2 = " << atan2(GBy[i][j], GBx[i][j]) << endl;
                }
                else if (GBxAux > 0)
                    ang = atan(GByAux / GBxAux);
                else if (GByAux >= 0) 
                    ang = atan(GByAux / GBxAux) + PI;
                else if (GByAux < 0)
                    ang = atan(GByAux / GBxAux) - PI;

                O[i][j] = 0.5 * PI + 0.5 * ang;
                
                xCos[i][j] = cos(2.0 * O[i][j]);
                ySin[i][j] = sin(2.0 * O[i][j]);
            }
            else {
                xCos[i][j] = 0;
                ySin[i][j] = 0;
                O[i][j] = -1;
            }
        }
        
    }
            
        if (ImgUtil::DEBUG) {
            string fileName2 = "OrientationSinCosMei";
            char str[20];
            sprintf(str, "%d", W);
            fileName2.append(str);
            ImgUtil::writeOrImg(xCos, ySin, orW, orH, fileName2);
        }
    
    MemUtil::freeMatrix(&xCos);
    MemUtil::freeMatrix(&ySin);
    MemUtil::freeMatrix(&GBx);
    MemUtil::freeMatrix(&GBy);
    MemUtil::freeMatrix(&coherence);
    
    return O;
}

float **OrientationFieldMei::computeVarianceAndSmooth_NonSmoothAreas(int W, float **O, float varThreshold) {
    
    int orH = h / W;
    int orW = w / W;
    
    float **variance = MemUtil::create2DMatrix<float>(orH, orW);
    float **xCos = MemUtil::create2DMatrix<float>(orH, orW);
    float **ySin = MemUtil::create2DMatrix<float>(orH, orW);
    
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {
            if (O[i][j] != -1) { // non-zero vector
                int n = 0;
                for (int u = max(0, i - 1); u < min(orH, i + 2); u++) {
                    for (int v = max(0, j - 1); v < min(orW, j + 2); v++) {
                        if ((!((u == i) && (v == j))) && O[u][v] != -1) {
                            variance[i][j] += min(abs(O[u][v] - O[i][j]), (float)(PI - abs(O[u][v] - O[i][j])));
                            n++;
                        }
                        else if ((O[u][v] == -1) && (!((u == i) && (v == j)))) {
                            variance[i][j] += PI ;
                            n++;
                        }
                    }
                }

                if (n != 0)
                    variance[i][j] = variance[i][j] / (float) n;
                else
                    variance[i][j] = 100;
            }
            else
                variance[i][j] = 100;
            

        }
    }
    
    
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {
            if (variance[i][j] >= varThreshold) { // NON-SMOOTH!
                O[i][j] = -1;
                xCos[i][j] = 0;
                ySin[i][j] = 0;
            }
            else {
                xCos[i][j] = cos(2.0 * O[i][j]);
                ySin[i][j] = sin(2.0 * O[i][j]);
            }
        }
    }
    if (ImgUtil::DEBUG) {
            string fileName2 = "OrientationSinCosMeiVariance";
            char str[20];
            sprintf(str, "%d", W);
            fileName2.append(str);
            ImgUtil::writeOrImg(xCos, ySin, orW, orH, fileName2);
        }
    
    MemUtil::freeMatrix(&xCos);
    MemUtil::freeMatrix(&ySin);
    
    return variance;
}


void OrientationFieldMei::combineFirstLevelOrientationFileds() {
    
    int orH0 = orH[0];
    int orW0 = orW[0];
    
    combinedO = MemUtil::create2DMatrix<float>(orH0,orW0);
    combinedGBx = MemUtil::create2DMatrix<float>(orH0,orW0);
    combinedGBy = MemUtil::create2DMatrix<float>(orH0,orW0);
    float **xCos = MemUtil::create2DMatrix<float>(orH0,orW0);
    float **ySin = MemUtil::create2DMatrix<float>(orH0,orW0);
    
    int proportion1 = (blockW[2] / blockW[0]);
    int proportion2 = (blockW[1] / blockW[0]);
    for (int i = 0; i < orH[0]; i++) {
        for (int j = 0; j < orW[0]; j++) {
            // mirar el correspondiente de O[2], si es smooth, cogerlo
            int u = i / proportion1;
            int v = j / proportion1;
            if ( ((u < orH[2]) && (v < orW[2]) ) && (O[2][u][v] != -1)) {
                combinedO[i][j] = O[2][u][v];
                combinedGBx[i][j] = GBx[2][u][v];
                combinedGBy[i][j] = GBy[2][u][v];
            }
            else {
                // sino mirar el correspondiente de O[1], si es smooth cogerlo
                u = i / proportion2;
                v = j / proportion2;
                
                if (((u < orH[1]) && (v < orW[1]) ) &&  (O[1][u][v] != -1)) {
                    combinedO[i][j] = O[1][u][v];
                    combinedGBx[i][j] = GBx[1][u][v];
                    combinedGBy[i][j] = GBy[1][u][v];
                }
                else {
                    // sino mirar el de O[0], si es smooth cogerlo
                    u = i;
                    v = j;
                    if (O[0][u][v] != -1) {
                        combinedO[i][j] = O[0][u][v];
                        combinedGBx[i][j] = GBx[0][u][v];
                        combinedGBy[i][j] = GBy[0][u][v];
                    }
                    else {// sino marcarlo como non-estimated
                        combinedO[i][j] = -1;
                        combinedGBx[i][j] = 0;
                        combinedGBy[i][j] = 0;
                    }
                }
            }        
            if (combinedO[i][j] != -1 ) {
                xCos[i][j] = cos(2.0 * combinedO[i][j]);
                ySin[i][j] = sin(2.0 * combinedO[i][j]);
            }
            else {
                xCos[i][j] = 0;
                ySin[i][j] = 0;
            }
        }
    }
    
     if (ImgUtil::DEBUG) {
        string fileName2 = "OrientationSinCosMeiCombined";
        ImgUtil::writeOrImg(xCos, ySin, orW[0], orH[0], fileName2);
    }

    MemUtil::freeMatrix(&xCos);
    MemUtil::freeMatrix(&ySin);
}


void OrientationFieldMei::predictNonEstimatedAreas() {
    
    int orH0 = orH[0];
    int orW0 = orW[0];
    // predecir las no estimadas de la parte exterior y marcarlas como estimadas.
    
    // basados en las nuevas estimaciones, repetir la operación hasta que no queden más por estimar
    bool **status = MemUtil::create2DMatrix<bool>(orH0, orW0);
    float **xCos = MemUtil::create2DMatrix<float>(orH0, orW0);
    float **ySin = MemUtil::create2DMatrix<float>(orH0, orW0);
    
    int nNonEstimated = 0;
    for (int i = 0; i < orH[0]; i++) {
        for (int j = 0; j < orW[0]; j++) {
        // dividir en bloques de 6 x 6 (W[0] que es combinedO!)
            if (combinedO[i][j] != -1) {// si combinedO[i][j] != -1 --> poner estatus a 1, sino a 0
                status[i][j] = true;
                xCos[i][j] = cos(2.0 * combinedO[i][j]);
                ySin[i][j] = sin(2.0 * combinedO[i][j]);
            }
            else {
                status[i][j] = false;
                nNonEstimated++;
                xCos[i][j] = 0;
                ySin[i][j] = 0;
            }
        }
        
    }

    int nReEstimated = 0;
    vector<int> reEstimated;
    do {
        nReEstimated = 0;
        for (int i = 0; i < orH[0]; i++) {
            for (int j = 0; j < orW[0]; j++) {
                if (!status[i][j]) {
                    int N = 0;
                    float GBxAux = 0;
                    float GByAux = 0;
                    for (int u = max(0, i - 1); u < min(orH0, i + 2); u++) {
                        for (int v = max(0, j - 1); v < min(orW0, j + 2); v++) {
                            if ((!((u == i) && (v == j))) && (status[u][v])) {
                                GBxAux += combinedGBx[u][v];
                                GByAux += combinedGBy[u][v];
                                N++;
                            }
                        }
                    }
                    if ( (N >= 4) && (GBxAux != 0) && (GByAux != 0) ) {
                        reEstimated.push_back(i);
                        reEstimated.push_back(j);
                        nReEstimated++;
                        float ang =  0;
                        if (GBxAux == 0) {
                            if (GBxAux > 0) 
                                ang = PI / 2.;
                            else
                                ang = - PI / 2.;
                        }
                        else if (GBxAux > 0)
                            ang = atan(GByAux / GBxAux);
                        else if (GByAux >= 0) 
                            ang = atan(GByAux / GBxAux) + PI;
                        else if (GByAux < 0)
                            ang = atan(GByAux / GBxAux) - PI;


                        combinedO[i][j] = 0.5 * PI + 0.5 * ang;
                        xCos[i][j] = cos(2.0 * combinedO[i][j]);
                        ySin[i][j] = sin(2.0 * combinedO[i][j]);

                        combinedGBx[i][j] = GBxAux / (float)N;
                        combinedGBy[i][j] = GByAux / (float)N;    
                    }
                }

            }
        }
        while (!reEstimated.empty()) {
            int v = reEstimated.back();
            reEstimated.pop_back();
            int u = reEstimated.back();
            reEstimated.pop_back();
            status[u][v] = true;
        }
        // cout << "nReEstimated = " << nReEstimated << endl;
        nNonEstimated -= nReEstimated;
    } while (nReEstimated > 0);

    
    // II ) para cada bloque
        // si su status es 0 y el número de bloques con 1s en su 8-neighborhood es >= 4
                // re-estimar el block-gradient con la suma de los 8-vecinos y poner status a 1
    
                // contar el número de bloques cambiados!
                // si es mayor que cero, repetir II
    // pasar todos los block gradients re-estimados a orientaciones
    
            
            
    if (ImgUtil::DEBUG) {
        string fileName2 = "OrientationSinCosMeiEstimated";
        ImgUtil::writeOrImg(xCos, ySin, orW[0], orH[0], fileName2);
    }
    
    MemUtil::freeMatrix(&status);
    MemUtil::freeMatrix(&xCos);
    MemUtil::freeMatrix(&ySin);
    
}
