/* 
 * \file        OrientationReconstructionLi.cpp
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * The implementation of OrientationReconstructionLi (Li2008) class
 *
 * 
 */


#include "OrientationReconstructionLi.h"
#include "MemUtil.hpp" 
#include "ComplexFiltersNilssonBigun.h"
#include "ParametersLi.h"



OrientationReconstructionLi::OrientationReconstructionLi() {
    O = 0;
    xCos = 0;
    ySin = 0;   
    coherence = 0;
    blockW = 0;
    orH = -1;
    orW = -1;
    dimFV = 0;
    blockW = blockWLi * ImgUtil::fAdjust;
}

OrientationReconstructionLi::~OrientationReconstructionLi() {
    MemUtil::freeMatrix(&O);
    MemUtil::freeMatrix(&xCos);
    MemUtil::freeMatrix(&ySin);
    MemUtil::freeMatrix(&coherence);
    cores.clear();
    deltas.clear();
}

float *OrientationReconstructionLi::run(string file, PCASYSManager *pcasys, bool segmentationPoincare) {
    FingerprintImg fi;

    if (ImgUtil::fAdjust != 1){   
        fi.setPCASYSManager(pcasys);   
        fi.readFingerprintImage(file);     
        fi.setSegmentedPCASYS(fi.getImage(), fi.getH(), fi.getW());
    }
    else{
        fi.setPCASYSManager(pcasys);
        fi.readFingerprintImage(file);
        fi.segment();
    }
    computeOrientations(fi.getSegmentedImageFloat(), fi.getH(), fi.getW());
    
    // Hay que llamar a nilssonbigun y obtener cores/deltas
    // luego cogerlos por pares e ir llamando a reconstruir etc
    
    ComplexFiltersNilssonBigun *cfnb = new ComplexFiltersNilssonBigun();
    cfnb->run(file, pcasys, segmentationPoincare);
    
    cores = cfnb->getCores();
    deltas = cfnb->getDeltas();
   
    
    MatrixXd allCoefficientsCores, allPositionsCores;
    verifySingularPoints(cores, allCoefficientsCores, allPositionsCores, true);
        
    MatrixXd allCoefficientsDeltas, allPositionsDeltas;
    verifySingularPoints(deltas, allCoefficientsDeltas, allPositionsDeltas);
               
    int nSingularPoints = (cores.size() + deltas.size()) / 2;
            
    MatrixXd allCoefficients(6, nSingularPoints), allPositions(2, nSingularPoints);
    
    int n = 0;
    for (int i = 0; i < allCoefficientsCores.cols(); i++) {
        allCoefficients.col(n) = allCoefficientsCores.col(i);
        allPositions.col(n) = allPositionsCores.col(i);
        n++;
    }
    for (int i = 0; i < allCoefficientsDeltas.cols(); i++) {
        allCoefficients.col(n) = allCoefficientsDeltas.col(i);
        allPositions.col(n) = allPositionsDeltas.col(i);
        n++;
    }
    
    vector<float> coreAngles = cfnb->getCoreAngles();
    vector<float> coreAnglesNew;
    vector<int> cores2 = cfnb->getCores();
    for (int i = 0; i < cores.size() / 2; i++) {  
        int a = allPositions(0, i);
        int b = allPositions(1, i);
        for (int j = 0; j < cores2.size() / 2; j++){            
            int c = cores2[j * 2] / blockW;
            int d = cores2[j * 2 + 1] / blockW;

            if ((a == c) && (b == d)) {
                coreAnglesNew.push_back(coreAngles[i]);
                break;
            }
        }
    }
    
    float **imgCrop = imageAlignement(fi.getSegmentedImageFloat(), fi.getH(), fi.getW(), coreAnglesNew);
     
    MemUtil::freeMatrix(&O);
    MemUtil::freeMatrix(&xCos);
    MemUtil::freeMatrix(&ySin);
    MemUtil::freeMatrix(&coherence);
    computeOrientations(imgCrop, croppingDimHLi, croppingDimWLi);   
        
    float **Opred = 0, **xCosPred = 0, **ySinPred = 0;
    float **coherenceTrain = MemUtil::copy2DMatrix<float>(coherence, orH, orW);
    
    if ((predictLi) && (reconstructPieceWiseOrientations(&xCosPred, &ySinPred, &Opred) == 0)) {    
        formTrainingData(xCosPred, ySinPred, Opred, coherenceTrain); 
    }
    else {
        Opred = MemUtil::copy2DMatrix<float>(O, orH, orW);
        xCosPred = MemUtil::copy2DMatrix<float>(xCos, orH, orW);
        ySinPred = MemUtil::copy2DMatrix<float>(ySin, orH, orW);    
    }
    
    int order = orderLi;

    //  computeWeightedLeastSquares(xCosPred, ySinPred, coherenceTrain, order);
    
     n = 0;
    // reconstruir first order de cada singular point en modo no piecewise y obtener los coeficientes
    for (int i = 0; i < cores.size() / 2; i++) {
        float D = 0;
        VectorXd coefficients = computeFirstOrderCoefficients(cores.at(2 * i), cores.at(2 * i + 1), Opred, &D);
        allCoefficients.col(n) = coefficients;
        allPositions(0, n) = cores.at(2 * i) / blockW;
        allPositions(1, n) = cores.at(2 * i + 1) / blockW;
        n++;
    }
    for (int i = 0; i < deltas.size() / 2; i++) {
        float D = 0;
        VectorXd coefficients = computeFirstOrderCoefficients(deltas.at(2 * i), deltas.at(2 * i + 1), Opred, &D);
        allCoefficients.col(n) = coefficients;
        allPositions(0, n) = deltas.at(2 * i) / blockW;
        allPositions(1, n) = deltas.at(2 * i + 1) / blockW;
        n++;
    }

    VectorXd coefficients = computeConstrainedWeightedLeastSquares(xCosPred, ySinPred, coherenceTrain, order, allPositions, allCoefficients);
        
    float *fv = createFeatureVector(coefficients, order, cfnb->getCoreMaxMag(), cfnb->getDeltaMaxMag());
    
    delete cfnb;
    
    MemUtil::freeMatrix(&Opred);
    MemUtil::freeMatrix(&xCosPred);
    MemUtil::freeMatrix(&ySinPred);
    MemUtil::freeMatrix(&imgCrop);
    
    return fv;
}


void OrientationReconstructionLi::verifySingularPoints(vector<int> &SPs, MatrixXd &allCoefficients, MatrixXd &allPositions, bool atLeast1) {
    
    int nSingularPoints = (SPs.size()) / 2;
    MatrixXd allC(6, nSingularPoints);
    MatrixXd allP(2, nSingularPoints);

    int n = 0;
    // reconstruir first order de cada singular point en modo no piecewise y obtener los coeficientes
    int nCores = SPs.size() / 2;
    vector<int> SPsNew;
    vector<int> SPsD;
    float Dmin = 10000;
    int DminI = -1;
    for (int i = 0; i < nCores; i++) {
        float D = 0;
        VectorXd coefficients = computeFirstOrderCoefficients(SPs.at(2 * i), SPs.at(2 * i + 1), O, &D);
        
        if (D <= DLi) {
            allC.col(n) = coefficients;
            allP(0, n) = SPs.at(2 * i) / blockW;
            allP(1, n) = SPs.at(2 * i + 1) / blockW;
            n++;
            SPsNew.push_back(SPs.at(2 * i));
            SPsNew.push_back(SPs.at(2 * i + 1));
            SPsD.push_back(D);
        }
        if ((atLeast1) && (D < Dmin)) {
            Dmin = D;
            DminI = i;
        }
    }
    
    if ((SPsD.size() == 0) && (atLeast1)) {  
        float D = 0;
        VectorXd coefficients = computeFirstOrderCoefficients(SPs.at(2 * DminI), SPs.at(2 * DminI + 1), O, &D);
        allC.col(n) = coefficients;
        allP(0, n) = SPs.at(2 * DminI) / blockW;
        allP(1, n) = SPs.at(2 * DminI + 1) / blockW;
        n++;
        SPsNew.push_back(SPs.at(2 * DminI));
        SPsNew.push_back(SPs.at(2 * DminI + 1));
        SPsD.push_back(D);
    }
    
    SPs.clear();
    // si hay más de 2 cores, elegir los de mayor respuesta, lo mismo con los deltas
    if (SPsNew.size() / 2 > 2) {
        // buscar el máximo y 2º máximo
        float max = -1, max2 = -1;
        float maxI = -1, max2I = -1;
        for  (int i = 0; i < SPsD.size(); i++) {
            if (SPsD[i] > max) {
                max2 = max;
                max2I = maxI;
                max = SPsD[i];
                maxI = i;
            }
            else if (SPsD[i] > max2) {
                max2 = SPsD[i];
                max2I = i;
            }
        }
        allCoefficients.resize(6, 2);
        allPositions.resize(2, 2);
        SPs.push_back(SPsNew[maxI * 2]);
        SPs.push_back(SPsNew[maxI * 2 + 1]);
        allCoefficients.col(0) = allC.col(maxI);
        allPositions.col(0) = allP.col(maxI);
        SPs.push_back(SPsNew[max2I * 2]);
        SPs.push_back(SPsNew[max2I * 2 + 1]);
        allCoefficients.col(1) = allC.col(max2I);
        allPositions.col(1) = allP.col(max2I);
    }
    else {
        nSingularPoints = SPsNew.size() / 2;
        allCoefficients.resize(6, nSingularPoints);
        allPositions.resize(2, nSingularPoints);
        for (int i = 0; i < nSingularPoints; i++) {
            allCoefficients.col(i) = allC.col(i);
            allPositions.col(i) = allP.col(i);
            SPs.push_back(SPsNew[i * 2]);
            SPs.push_back(SPsNew[i * 2 + 1]);
        }     
    }

    SPsD.clear(); 
    SPsNew.clear();
}

void OrientationReconstructionLi::verifySingularPoints(float **xCosPred, float **ySinPred, float **coherenceTrain, int order, MatrixXd allPositions, MatrixXd allCoefficients) {

//    cout << "allPositions = " << endl << allPositions << endl;
//    cout << "allCoefficients = " << endl << allCoefficients << endl;
    // para cada posible core
    for (int i = 0; i < cores.size() / 2; i++) {
        float **xCosNew = MemUtil::copy2DMatrix<float>(xCosPred, orH, orW);
        float **ySinNew = MemUtil::copy2DMatrix<float>(xCosPred, orH, orW);    
        float **Onew = MemUtil::create2DMatrix<float>(orH, orW);   
        
        MatrixXd allPos(2, 1), allCoeff(6, 1);
        allPos = allPositions.col(i);
        allCoeff = allCoefficients.col(i);
        
        VectorXd coefficients = computeConstrainedWeightedLeastSquares(xCosPred, ySinPred, coherenceTrain, order, allPos, allCoeff, &Onew);
        
        int iArea = influenceAreaLi * 2 + 1;
        float D = orientationDifference(O, Onew, coherence, orH, orW, iArea, iArea, allPos(0, 0), allPos(1, 0));
        //, int smallH, int smallW, int centralH, int centralW
        
        cout << "D core " << i << " = " << D << endl;
        
        MemUtil::freeMatrix(&xCosNew);
        MemUtil::freeMatrix(&ySinNew);
        MemUtil::freeMatrix(&Onew);
    }
    
    for (int i = 0; i < cores.size() / 2; i++) {
        for (int j = 0; j < deltas.size() / 2; j++) {
            float **xCosNew = MemUtil::copy2DMatrix<float>(xCosPred, orH, orW);
            float **ySinNew = MemUtil::copy2DMatrix<float>(xCosPred, orH, orW);    
            float **Onew = MemUtil::create2DMatrix<float>(orH, orW);   

            MatrixXd allPos(2, 2), allCoeff(6, 2);
            allPos.col(0) = allPositions.col(i);
            allPos.col(1) = allPositions.col(cores.size() / 2 + j);
            allCoeff.col(0) = allCoefficients.col(i);
            allCoeff.col(1) = allCoefficients.col(cores.size() / 2 + j);

            VectorXd coefficients = computeConstrainedWeightedLeastSquares(xCosPred, ySinPred, coherenceTrain, order, allPos, allCoeff, &Onew);

            int iArea = influenceAreaLi * 2 + 1;
            float D = orientationDifference(O, Onew, coherence, orH, orW, iArea, iArea, allPos(0, 0), allPos(1, 0));
            float D2 = orientationDifference(O, Onew, coherence, orH, orW,  iArea, iArea, allPos(0, 1), allPos(1, 1));

            //, int smallH, int smallW, int centralH, int centralW
            cout << "D core " << i << " = " << D << " delta " << j << " = " << D2 << endl;

            MemUtil::freeMatrix(&xCosNew);
            MemUtil::freeMatrix(&ySinNew);
            MemUtil::freeMatrix(&Onew);
        }
    }
    
    
    
//    VectorXd coefficients = computeConstrainedWeightedLeastSquares(xCosPred, ySinPred, coherenceTrain, order, allPositions, allCoefficients);

}



VectorXd OrientationReconstructionLi::computeConstrainedWeightedLeastSquares(float **xCosPred, float **ySinPred, float **coherenceTrain, int order, MatrixXd allPositions, MatrixXd allCoefficients, float ***Onew) { // Non-constrained
    
    // Formar la matriz A, luego la bdx y bdy, hacer leastsquares A * a = bdx y A * b = bdy, con pesos, multiplicar todo por \srqt{w_i}
    int nCoefficients = (order + 1) * (order + 2) / 2;
    MatrixXd A = MatrixXd::Zero(orH * orW, nCoefficients);
    VectorXd bdx = VectorXd::Zero(orH * orW);
    VectorXd bdy = VectorXd::Zero(orH * orW);
    
    int corePosI = allPositions(0, 0);
    int corePosJ = allPositions(1, 0);
    
    int n = 0;
    for (int i = max(0, 0); i < min(orH, orH); i++) {
        for (int j = max(0, 0); j <  min(orW, orW); j++) {

            int i3 = orH - i;
            int j3 = j + 1;
            
            float w = sqrt(coherenceTrain[i][j]);
            
            int n2 = 0;
            for (int i2 = 0; i2 <= order; i2++) {
                for (int j2 = 0; j2 <= i2; j2++) {
                    A(n, n2) = w * pow(j3, i2 - j2) * pow(i3, j2);   
                    n2++;
                }
            }
            bdx(n) = xCosPred[i][j] * w;
            bdy(n) = ySinPred[i][j] * w;
            
            n++;
        }        
    }  
    // Construir constrainst para cada SP
    // Por cada SP hay 3 para a's y 3 par b's (dx y dy)
    // hay allPositions.rows() SPs
    // B es la de las restricciones. Solo hacemos 3 por SP en vez de 6 porque al igual que la A, coincide B para dx  y dy
    // sin embargo, no coincide para el d de Bx = d (ddx y ddy)!!
    int nSingularPoints = allPositions.cols();
    
    int nConstraintsPerSP = 3; 
    
    MatrixXd B(nConstraintsPerSP * nSingularPoints, nCoefficients);
    VectorXd ddx(nConstraintsPerSP * nSingularPoints);
    VectorXd ddy(nConstraintsPerSP * nSingularPoints);
    
    for (int i = 0; i < nSingularPoints; i++) {    
        
        int i3 = orH - allPositions(0, i);
        int j3 = allPositions(1, i) + 1;
               
        int n2 = 0;
        for (int i2 = 0; i2 <= order; i2++) {
            for (int j2 = 0; j2 <= i2; j2++) {
                B(i * nConstraintsPerSP, n2) = pow(-j3, i2 - j2) * pow(-i3, j2);   
                B(i * nConstraintsPerSP + 1, n2) = (i2 - j2) * pow(-j3, i2 - j2) * pow(-i3, j2); 
                B(i * nConstraintsPerSP + 2, n2) = j2 * pow(-i3, i2 - j2) * pow(-j3, j2); 
                n2++;
            }
        }
        
        ddx(i * nConstraintsPerSP) = 0;
        ddy(i * nConstraintsPerSP) = 0;
        ddx(i * nConstraintsPerSP + 1) = allCoefficients(2, i);
        ddx(i * nConstraintsPerSP + 2) = allCoefficients(3, i);
        ddy(i * nConstraintsPerSP + 1) = allCoefficients(0, i);
        ddy(i * nConstraintsPerSP + 2) = allCoefficients(1, i);
             
    }
      
    HouseholderQR<MatrixXd> qr(B.transpose());
    MatrixXd Q = qr.householderQ();
    MatrixXd R = qr.matrixQR().triangularView<Upper>();
    
    
    int m = A.rows(); 
    n = A.cols(); // y B.cols()!!
    int p = B.rows();
    
    MatrixXd R_upper = R.topRows(p).transpose();
    
    
    MatrixXd y_dx = R_upper.colPivHouseholderQr().solve(ddx);
    MatrixXd y_dy = R_upper.colPivHouseholderQr().solve(ddy);
    
//    if (ImgUtil::DEBUG) {
//
//        cout << "Q = " << endl << Q << endl;
//        cout << "R = " << endl << R << endl;
//        cout << "Q * R - Bt = " << (Q * R - B.transpose()) << endl;
//
//        cout << "Q * R - Bt = " << (Q * R - B.transpose()).norm() << endl;
//
//
//        cout << "Q * Q = " << endl << Q * Q << endl;
//
//        cout << "R_upper transpose = " << endl << R_upper << endl;
//
//        cout << "y_dx = " << endl << y_dx << endl;
//        cout << "y_dy = " << endl << y_dy << endl;
//
//
//        cout << "Error R_upper transpose * y = d  with y_dx = " << endl << (R_upper * y_dx - ddx).norm() << endl;
//        cout << "Error R_upper transpose * y = d  with y_dy = " << endl << (R_upper * y_dy - ddy).norm() << endl;
//    }
//    
    MatrixXd AQ = A * Q;    
    MatrixXd A1 = AQ.leftCols(p); // p
    MatrixXd A2 = AQ.rightCols(n - p); // n-p
    
     
    VectorXd bdx2 = (bdx - A1 * y_dx);
    VectorXd bdy2 = (bdy - A1 * y_dy);
    
    MatrixXd yz_dx(n, 1);
    yz_dx.topRows(p) = y_dx;
    MatrixXd yz_dy(n, 1);
    yz_dy.topRows(p) = y_dy;
    
    // ahora dependiendo de A2, se añaden las p restantes, si n - p > 0
    if (n - p > 0) {    
        JacobiSVD<MatrixXd> svd(A2, ComputeThinU | ComputeThinV);
        MatrixXd zdx = svd.solve(bdx2);   
        MatrixXd zdy = svd.solve(bdy2);         
        
//        if (ImgUtil::DEBUG) {
//            cout << "zdx = " << endl << zdx << endl;
//            cout << "zdy = " << endl << zdy << endl;
//
//            cout << "Error A2z = (b - A1y) with dx ->" << (A2 * zdx - bdx2).norm() << endl;    
//            cout << "Error A2z = (b - A1y) with dy ->" << (A2 * zdy - bdy2).norm() << endl;  
//        }

        yz_dx.bottomRows(n - p) = zdx;
        yz_dy.bottomRows(n - p) = zdy;    
    }
    
    MatrixXd x_dx = Q * yz_dx;
    MatrixXd x_dy = Q * yz_dy;
    
//    if (ImgUtil::DEBUG) {
//        cout << "x_dx = " << x_dx << endl;    
//        cout << "Error = " << (A * x_dx - bdx).norm() << endl;    
//        cout << "x_dy = " << x_dy << endl;
//        cout << "Error = " << (A * x_dy - bdy).norm() << endl;
//
//
//        cout << "Error Bx = d with dx ->" << (B * x_dx - ddx).norm() << endl;    
//        cout << "Error Bx = d with dy ->" << (B * x_dy - ddy).norm() << endl;  
//
//
//        cout << "Error Bx = d with dx ->" << (B * x_dx - ddx) << endl;    
//        cout << "Error Bx = d with dy ->" << (B * x_dy - ddy) << endl;    
//
//
//        cout << " FINAL ERROR " << endl;
//        cout << "Error Ax = b with dx ->" << (A * x_dx - bdx).norm() << endl;    
//        cout << "Error Ax = b with dy ->" << (A * x_dy - bdy).norm() << endl;  
//    }
    
    n = 0;
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {

            int i3 = orH - i;
            int j3 = j + 1;
            
            float dx = 0, dy = 0;
                                    
            int n2 = 0;
            for (int i2 = 0; i2 <= order; i2++) {
                for (int j2 = 0; j2 <= i2; j2++) {
                    dx += pow(j3, i2 - j2) * pow(i3, j2) * x_dx(n2, 0);
                    dy += pow(j3, i2 - j2) * pow(i3, j2) * x_dy(n2, 0);
                    n2++;
                }
            }
            
            xCosPred[i][j] = dx;
            ySinPred[i][j] = dy;
            
            float Oaux = 0.5 * atan2(dy, dx);
            
            xCosPred[i][j] = cos(2. * Oaux);
            ySinPred[i][j] = sin(2. * Oaux);
            
            if (Onew != 0) {
                (*Onew)[i][j] = Oaux;
            }
            
            n++;
        }
    }
    
    if (ImgUtil::DEBUG) {
        string fileName2 = "ReconstructedConstrainedWeightedLeastSquares";
        ImgUtil::writeOrImg(xCosPred, ySinPred, orW, orH, fileName2);
    }
    
    VectorXd coefficients(nCoefficients * 2);
    coefficients.head(nCoefficients) = x_dx;
    coefficients.tail(nCoefficients) = x_dy;
    
    return coefficients;
}


VectorXd OrientationReconstructionLi::computeWeightedLeastSquares(float **xCosPred, float **ySinPred, float **coherenceTrain, int order) { // Non-constrained
    
    // Formar la matriz A, luego la bdx y bdy, hacer leastsquares A * a = bdx y A * b = bdy, con pesos, multiplicar todo por \srqt{w_i}
    int nCoefficients = (order + 1) * (order + 2) / 2;
    MatrixXd A(orH * orW, nCoefficients);
    VectorXd bdx(orH * orW);
    VectorXd bdy(orH * orW);
    
    int n = 0;
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {

            int i3 = orH - i;
            int j3 = j + 1;
            
            float w = sqrt(coherenceTrain[i][j]);
                                    
            int n2 = 0;
            for (int i2 = 0; i2 <= order; i2++) {
                for (int j2 = 0; j2 <= i2; j2++) {
                    A(n, n2) = w * pow(j3, i2 - j2) * pow(i3, j2);   
                    n2++;
                }
            }

            bdx(n) = xCosPred[i][j] * w;
            bdy(n) = ySinPred[i][j] * w;
            
            n++;
            //cout << "w = " << w << " nu = " << nu << endl;
        }
    }
//    
    MatrixXd x_dx = A.jacobiSvd(ComputeThinU | ComputeThinV).solve(bdx);    
    MatrixXd x_dy = A.jacobiSvd(ComputeThinU | ComputeThinV).solve(bdy);       
    
    
    n = 0;
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {

            int i3 = orH - i;
            int j3 = j + 1;
            
            float dx = 0, dy = 0;
                                    
            int n2 = 0;
            for (int i2 = 0; i2 <= order; i2++) {
                for (int j2 = 0; j2 <= i2; j2++) {
                    dx += pow(j3, i2 - j2) * pow(i3, j2) * x_dx(n2, 0);
                    dy += pow(j3, i2 - j2) * pow(i3, j2) * x_dy(n2, 0);
                    n2++;
                }
            }

            xCosPred[i][j] = dx;
            ySinPred[i][j] = dy;
            
            float Oaux = 0.5 * atan2(dy, dx);
            
            xCosPred[i][j] = cos(2. * Oaux);
            ySinPred[i][j] = sin(2. * Oaux);
            
            n++;
        }
    }
    
    if (ImgUtil::DEBUG) {
        string fileName2 = "ReconstructedWeightedLeastSquares";
        ImgUtil::writeOrImg(xCosPred, ySinPred, orW, orH, fileName2);
    }
}

VectorXd OrientationReconstructionLi::computeFirstOrderCoefficients(int posI, int posJ, float **Oold, float *D) {    
    MatrixXd omega2, omega4;    
    float squared = 2;
    int order = 1;
    
    computeFirstOrderOmegas(posI, posJ, omega2, omega4, squared, Oold);
    VectorXd coefficients = computeCoefficientsFromOmegas(omega2, omega4, order); // orden 1 para la prueba
    
    float **xCosNew = 0, **ySinNew = 0; 
    float **Onew = reconstructSPZoneOrientationsFromFirstOrderCoefficients(posI, posJ, coefficients, &xCosNew, &ySinNew, squared);
    
    int iArea = influenceAreaLi * 2 + 1;
    float D2 = orientationDifference(O, Onew, coherence, orH, orW, iArea, iArea, posI / blockW, posJ / blockW);
    
    if (D2 > DLi) {
            
        if (ImgUtil::DEBUG) {
            string fileName2 = "OrientationRecontructedPred1_";
            char str[20];
            sprintf(str, "%d_%d", posI, posJ);
            fileName2.append(str);                    
            ImgUtil::writeOrImg(xCosNew, ySinNew, orW, orH, fileName2);
        }
        
        
        MemUtil::freeMatrix(&Onew);
        MemUtil::freeMatrix(&xCosNew);
        MemUtil::freeMatrix(&ySinNew);
        squared += 1;
        
        Onew = reconstructSPZoneOrientationsFromFirstOrderCoefficients(posI, posJ, coefficients, &xCosNew, &ySinNew, squared);
        D2 = orientationDifference(O, Onew, coherence, orH, orW, iArea, iArea, posI / blockW, posJ / blockW);
    }
    
    if (D != 0)
        *D = D2;
    
    if (ImgUtil::DEBUG) {
        cout << "D first order 2 = " << *D << endl;
        string fileName2 = "OrientationRecontructedPred_";
        char str[20];
        sprintf(str, "%d_%d", posI, posJ);
        fileName2.append(str);                    
        ImgUtil::writeOrImg(xCosNew, ySinNew, orW, orH, fileName2);
    }

    MemUtil::freeMatrix(&Onew);
    MemUtil::freeMatrix(&xCosNew);
    MemUtil::freeMatrix(&ySinNew);
    
    return coefficients;    
}

void OrientationReconstructionLi::computeOrientations(float **img, int h, int w) {
    float **gradX, **gradY, **grad;
    ImgUtil::computeSobelGradients(img, h, w, &gradX, &gradY, &grad);
    
    
    if (ImgUtil::DEBUG) {
        string fileName2 = "GradX";
        ImgUtil::writeGrayScaleImage(gradX, h, w, fileName2, 1);
        fileName2 = "GradY";
        ImgUtil::writeGrayScaleImage(gradY, h, w, fileName2, 1);
        fileName2 = "Grad";
        ImgUtil::writeGrayScaleImage(grad, h, w, fileName2, 1);
    }
    
    coherence = MemUtil::create2DMatrix<float>(h / blockW, w / blockW);
    O = ImgUtil::computeOrientationsFromGradients(gradX, gradY, h, w, blockW, &xCos, &ySin, &orH, &orW, false, coherence);
    if (ImgUtil::DEBUG) {
        string fileName2 = "OrientationSinCos";
                
        ImgUtil::writeOrImg(xCos, ySin, orW, orH, fileName2);
    }
    
    ImgUtil::smoothOrientations(O, &xCos, &ySin, orH, orW, smoothOrientatioWLi);
    
    
    
    if (ImgUtil::DEBUG) {
        string fileName2 = "OrientationSinCosSmooth";
                
        ImgUtil::writeOrImg(xCos, ySin, orW, orH, fileName2);
        
        fileName2 = "OrientationCoherence";
        ImgUtil::writeGrayScaleImage(coherence, orH, orW, fileName2, 255);
    }
    
        
    MemUtil::freeMatrix(&gradX);
    MemUtil::freeMatrix(&gradY);
    MemUtil::freeMatrix(&grad);
    
}

int OrientationReconstructionLi::reconstructPieceWiseOrientations(float ***xCosPred, float ***ySinPred, float ***Opred) {
    
    // float **xCosNew = 0, **ySinNew = 0, **Onew = 0;
    // Caso de 1 solo core, con línea completamente horizontal que divide la huella, se asume la horizontalidad!!!
    if ((cores.size() == 2) && (deltas.empty())) { // Solo hay 1 core :D        
        reconstructPieceWiseUniqueCore(cores[0], cores[1], xCosPred, ySinPred, Opred);               
    }
    else if ((cores.size() == 2) && (deltas.size() == 2)) {       
        // Se asume 1 core y un delta, el delta siempre estríctamente debajo del core!!!!!!
        int out = reconstructPieceWiseCoreDelta(cores[0], cores[1], deltas[0], deltas[1], xCosPred, ySinPred, Opred); 
        
//        if (out == -1)
//            cout << "ABORTADA RECONSTRUCCIÓN NO PUEDE SER!!!!" << endl;
        return out;
    }
    else if ((cores.size() == 4) && (deltas.size() == 4)) {
        
        int out = reconstructPieceWise2Core2Delta(xCosPred, ySinPred, Opred); 
        
//        if (out == -1)
//            cout << "ABORTADA RECONSTRUCCIÓN NO PUEDE SER!!!!" << endl;
        return out;
        
    }
    else {
//        cout << "Cannot reconstruct with these number of cores/deltas" << endl;
        return -1;
    }
    
    return 0;
    
}

void OrientationReconstructionLi::reconstructOrientations() {
    
    MatrixXd omega2, omega4;
    
    float **xCosNew = 0, **ySinNew = 0; 
    float m;
    float **Onew = reconstructPieceWiseCore(cores[0], cores[1], &xCosNew, &ySinNew, &m);
    
    
    MemUtil::freeMatrix(&Onew);
    MemUtil::freeMatrix(&xCosNew);
    MemUtil::freeMatrix(&ySinNew);
    
    
    int **retMask;
    Onew = reconstructPieceWiseDelta(deltas[0], deltas[1], &xCosNew, &ySinNew, &retMask);
    
    
    MemUtil::freeMatrix(&Onew);
    MemUtil::freeMatrix(&xCosNew);
    MemUtil::freeMatrix(&ySinNew);
    MemUtil::freeMatrix(&retMask);
    
    
}

void OrientationReconstructionLi::computeFirstOrderOmegas(int posI, int posJ, MatrixXd &omega2, MatrixXd &omega4, float squared, float **Oold, bool **mask) {//float ***omega2, float ***omega4) {
    
    bool useMask = false;
    if (mask != 0)
        useMask = true;
    
    // usando influenceAreaLi se calculan las matrices en la posición posI, posJ
    
    int nElements = influenceAreaLi * 2 + 1;
    nElements *= nElements;
    
    
    if (useMask) {
        nElements = 0;
        int lMask = (influenceAreaLi * 2 + 1);
        for (int i = 0; i < lMask; i++)
            for (int j = 0; j < lMask; j++)
                if (mask[i][j])
                    nElements++;
    }
        
    omega2 = MatrixXd::Zero(nElements, 2);
    omega4 = MatrixXd::Zero(nElements, 4);
    
    
    // pasar posI y posJ a bloques de Li blockW
    posI = (posI / blockW);
    posJ = (posJ / blockW); 
    
    
    int n = 0;
    for (int i = max(0, posI - influenceAreaLi); i < min(orH, posI + influenceAreaLi + 1); i++) {
        for (int j = max(0, posJ - influenceAreaLi); j < min(orW, posJ + influenceAreaLi + 1); j++) {
            // int y1 = i;
            // int x1 = j;
            
            int i3 = (i - (posI - influenceAreaLi));
            int j3 = (j - (posJ - influenceAreaLi));
            
            if (!useMask || (mask[i3][j3])) {
            
                int i2 = orH - i; //(influenceAreaLi * 2 + 1) - (i - (posI - influenceAreaLi)); // (orH - i)
                int j2 = j + 1; //(j - (posJ - influenceAreaLi) + 1); // (j + 1)

                float nu = tan(squared * Oold[i][j]); // tan(O[y1][x1]);
                float w = cos(squared * Oold[i][j]); //coherence[i][j]; // coherence[y1][x1]

                omega2(n, 0) = w;
                omega2(n, 1) = -w * nu;

                omega4(n, 0) = j2 * w; // x1 * w;
                omega4(n, 1) = i2 * w; // y1 * w;
                omega4(n, 2) = -nu * j2 * w; // -nu * x1 * w;
                omega4(n, 3) = -nu * i2 * w; // -nu * y1 * w;

                n++;
                //cout << "w = " << w << " nu = " << nu << endl;
            }
        }
    }
    
}

VectorXd OrientationReconstructionLi::computeCoefficientsFromOmegas(MatrixXd omega2, MatrixXd omega4, int order) {
    
    int nParameters = (order + 1) * (order + 2); // - 2; // dos son siempre los independientes    
    int nElements = influenceAreaLi * 2 + 1;
    nElements *= nElements;

    
    MatrixXd Psi = -omega4.transpose() * omega4 + omega4.transpose() * omega2 * (omega2.transpose() * omega2).colPivHouseholderQr().solve(omega2.transpose() * omega4);    
    EigenSolver<MatrixXd> es(Psi);
    VectorXcd eigenvalues = es.eigenvalues();
    
    int i;
    eigenvalues.real().cwiseAbs().minCoeff(&i);
        
    MatrixXd L4 = es.eigenvectors().col(i).real();    
    
    MatrixXd L2 = -(omega2.transpose() * omega2).colPivHouseholderQr().solve(omega2.transpose() * omega4 * L4);
    VectorXd coefficients(nParameters);
    coefficients << L4.col(0).head(nParameters - 2), L2.col(0); 
    
    return coefficients;

}

float **OrientationReconstructionLi::reconstructSPZoneOrientationsFromFirstOrderCoefficients(int posI, int posJ, VectorXd coefficients, float ***xCosNew, float ***ySinNew, float squared, bool **mask, int *limits) {
    //float **Onew = MemUtil::create2DMatrix<float>(orH, orW);
    float **Onew = MemUtil::copy2DMatrix<float>(O, orH, orW);
    
    float alReves = 1;
    if (squared == 3) {
        squared = 2;
        alReves = -1;
    }
    
    if (((*xCosNew) == 0)) {
        *xCosNew = MemUtil::copy2DMatrix<float>(xCos, orH, orW); // MemUtil::create2DMatrix<float>(orH, orW);
        *ySinNew = MemUtil::copy2DMatrix<float>(ySin, orH, orW); //MemUtil::create2DMatrix<float>(orH, orW);
    }
    // pasar posI y posJ a bloques de Li blockW
    posI = (posI / blockW);
    posJ = (posJ / blockW);
    
    bool useMask = false;
    if (mask != 0)
        useMask = true;
    
    bool noLimits = false;
    if (limits == 0) {
        noLimits = true;
        limits = MemUtil::createVector<int>(4);
        limits[0] = max(0, posI - influenceAreaLi);
        limits[1] = min(orH, posI + influenceAreaLi + 1);
        limits[2] = max(0, posJ - influenceAreaLi);
        limits[3] = min(orW, posJ + influenceAreaLi + 1);
    }
    
    for (int i = limits[0]; i < limits[1]; i++) {
        for (int j = limits[2]; j < limits[3]; j++) {
            int i3 = (i - limits[0]);
            int j3 = (j - limits[2]);
            
            if (!useMask || (mask[i3][j3]) ) {
                
                int i2 = (orH - i);  //(influenceAreaLi * 2 + 1) - (i - limits[0] ); // (orH - i)
                int j2 = j + 1; //(j - limits[2] + 1); // (j + 1)
                // cout << "pasando la i = " << i << " a " << i2 << endl;
                (*ySinNew)[i][j] = coefficients(0) * j2 + coefficients(1) * i2 + coefficients(4);
                (*xCosNew)[i][j] = coefficients(2) * j2 + coefficients(3) * i2 + coefficients(5);
                Onew[i][j] = 1. / squared * atan2((*ySinNew)[i][j], (*xCosNew)[i][j]);

    //            if (delta) {
                if (((coefficients(0) * alReves > 0)  && (squared == 2))) { // && (squared == 2)) || ((coefficients(0) < 0) && (squared == 1)) ) {
                    if (Onew[i][j] <= 0)
                        Onew[i][j] = Onew[i][j] + 0.5 * PI;
                    else
                        Onew[i][j] = Onew[i][j] - 0.5 * PI;
                }

                if (Onew[i][j] <= (-PI / 2.))
                    Onew[i][j] += PI;
                else if (Onew[i][j] > (PI / 2.))
                    Onew[i][j] -= PI;

                (*ySinNew)[i][j] = sin(2. * Onew[i][j]); // squared
                (*xCosNew)[i][j] = cos(2. * Onew[i][j]);
            }
            
        }
    }
    
    if (noLimits) {
        MemUtil::freeMatrix(&limits);
    }
    
    return Onew;
}

float **OrientationReconstructionLi::reconstructSPZoneOrientationsFromFirstOrderCoefficientsABCD(int posI, int posJ, VectorXd coefficients, float ***xCosNew, float ***ySinNew, bool **mask, int *limits) {
    //float **Onew = MemUtil::create2DMatrix<float>(orH, orW);
    float **Onew = MemUtil::copy2DMatrix<float>(O, orH, orW);
    
   if (((*xCosNew) == 0)) {
        *xCosNew = MemUtil::copy2DMatrix<float>(xCos, orH, orW); // MemUtil::create2DMatrix<float>(orH, orW);
        *ySinNew = MemUtil::copy2DMatrix<float>(ySin, orH, orW); //MemUtil::create2DMatrix<float>(orH, orW);
    }
    // pasar posI y posJ a bloques de Li blockW
    posI = (posI / blockW);
    posJ = (posJ / blockW);
    
    bool useMask = false;
    if (mask != 0)
        useMask = true;
    
    bool noLimits = false;
    if (limits == 0) {
        noLimits = true;
        limits = MemUtil::createVector<int>(4);
        limits[0] = max(0, posI - influenceAreaLi);
        limits[1] = min(orH, posI + influenceAreaLi + 1);
        limits[2] = max(0, posJ - influenceAreaLi);
        limits[3] = min(orW, posJ + influenceAreaLi + 1);
    }
    //cout << "coef" << coefficients << endl;
    for (int i = limits[0]; i < limits[1]; i++) {
        for (int j = limits[2]; j < limits[3]; j++) {
            int i3 = (i - limits[0]);
            int j3 = (j - limits[2]);
            
            if (!useMask || (mask[i3][j3]) ) {
                
                /// ESTA EL ORDEN CAMBIADO A PROPOSITO!!!
                (*xCosNew)[i][j] = coefficients(0) * (j - posJ) + coefficients(1) * (posI - i);
                (*ySinNew)[i][j] = coefficients(2) * (j - posJ) + coefficients(3) * (posI - i); 

                Onew[i][j] =  atan2((*ySinNew)[i][j], (*xCosNew)[i][j]);

                if (Onew[i][j] <= (-PI / 2.))
                    Onew[i][j] += PI;
                else if (Onew[i][j] > (PI / 2.))
                    Onew[i][j] -= PI;

                (*ySinNew)[i][j] = sin(2. * Onew[i][j]);
                (*xCosNew)[i][j] = cos(2. * Onew[i][j]);
            
            }
            
        }
    }
    
    
    if (noLimits) {
        MemUtil::freeMatrix(&limits);
    }
    
    return Onew;
}

float **OrientationReconstructionLi::reconstructSPZoneOrientationsFromCoefficients(int posI, int posJ, VectorXd coefficients, int order, float ***xCosNew, float ***ySinNew, float squared) {
        //float **Onew = MemUtil::create2DMatrix<float>(orH, orW);
    float **Onew = MemUtil::copy2DMatrix<float>(O, orH, orW);
    
    *xCosNew = MemUtil::copy2DMatrix<float>(xCos, orH, orW); // MemUtil::create2DMatrix<float>(orH, orW);
    *ySinNew = MemUtil::copy2DMatrix<float>(ySin, orH, orW); //MemUtil::create2DMatrix<float>(orH, orW);
    
    // pasar posI y posJ a bloques de Li blockW
    posI = (posI / blockW);
    posJ = (posJ / blockW);
    
    for (int i = posI - influenceAreaLi; i < posI + influenceAreaLi + 1; i++) {
        for (int j = posJ - influenceAreaLi; j < posJ + influenceAreaLi + 1; j++) {
            
            int i3 = (influenceAreaLi * 2 + 1) - (i - (posI - influenceAreaLi) ); // (orH - i)
            int j3 = (j - (posJ - influenceAreaLi) + 1); // (j + 1)
            int n2 = 0;
            (*ySinNew)[i][j] = 0;
            (*xCosNew)[i][j] = 0;
            for (int i2 = 1; i2 <= order; i2++) {
                for (int j2 = 0; j2 <= i2; j2++) {
                    (*ySinNew)[i][j] += coefficients(n2) * pow(j3, i2 - j2) * pow(i3, j2);
                    (*xCosNew)[i][j] += coefficients(n2 + (coefficients.size() - 2) / 2) * pow(j3, i2 - j2) * pow(i3, j2); 
                    n2++;
                }
            }
            (*ySinNew)[i][j] += coefficients(coefficients.size() - 2);
            (*xCosNew)[i][j] += coefficients(coefficients.size() - 1);
            
            Onew[i][j] = 1. / squared * atan2((*ySinNew)[i][j], (*xCosNew)[i][j]);
            
            
            if (((coefficients(0) > 0) )) { // 
                if (Onew[i][j] <= 0)
                    Onew[i][j] = Onew[i][j] + 0.5 * PI;
                else
                    Onew[i][j] = Onew[i][j] - 0.5 * PI;
            }
            
            if (Onew[i][j] <= (-PI / 2.))
                Onew[i][j] += PI;
            else if (Onew[i][j] > (PI / 2.))
                Onew[i][j] -= PI;
                    
            (*ySinNew)[i][j] = sin(squared * Onew[i][j]);
            (*xCosNew)[i][j] = cos(squared * Onew[i][j]);
            
        }
    }    
    return Onew;
    
}


float **OrientationReconstructionLi::reconstructPieceWiseCore(int posI, int posJ, float ***xCosNew, float ***ySinNew, float *m_line, float core2) {
    
    // pasar posI y posJ a bloques de Li blockW
    posI = (posI / blockW);
    posJ = (posJ / blockW);
    
    
    int iArea = (influenceAreaLi + 1);
    // Partir en cuatro cuadrantes, y sacaar la varianza en cada uno
    
    // Sacar la media de la orientacion en cada cuadrante
    float **meanOcos = MemUtil::create2DMatrix<float>(2, 2);    
    float **meanOsin = MemUtil::create2DMatrix<float>(2, 2);    
    float **meanO = MemUtil::create2DMatrix<float>(2, 2);
    
    for (int i = max(0, posI - influenceAreaLi); i < min(orH, posI + influenceAreaLi + 1); i++) {
        for (int j = max(0, posJ - influenceAreaLi); j < min(orW, posJ + influenceAreaLi + 1); j++) {
            
            int i2 = (i - (posI - influenceAreaLi));
            int j2 = (j - (posJ - influenceAreaLi));             
                       
            meanOcos[i2 / iArea][j2 / iArea] += cos(2. * O[i][j]);
            meanOsin[i2 / iArea][j2 / iArea] += sin(2. * O[i][j]);
            
        }
    }
    for (int i = 0; i < 2; i++) {
        for (int j = 0; j < 2; j++) {
            meanO[i][j] = 0.5 * atan2(meanOsin[i][j], meanOcos[i][j]);
//            if (ImgUtil::DEBUG)
//                cout << "mean del cuadrante " << i << ", " << j << " = " << meanO[i][j] * 180. / PI << endl;
        }
    }
       
    float **varO = MemUtil::create2DMatrix<float>(2, 2);
    int **count = MemUtil::create2DMatrix<int>(2, 2);
    
    for (int i = max(0, posI - influenceAreaLi); i < min(orH, posI + influenceAreaLi + 1); i++) {
        for (int j = max(0, posJ - influenceAreaLi); j < min(orW, posJ + influenceAreaLi + 1); j++) {
            
            int i2 = (i - (posI - influenceAreaLi));
            int j2 = (j - (posJ - influenceAreaLi));
            
            float diff = abs(meanO[i2 / iArea][j2 / iArea] - O[i][j]);
            if (diff > PI / 2.)
                diff -= PI;
            
            varO[i2 / iArea][j2 / iArea] += diff * diff;
            count[i2 / iArea][j2 / iArea]++;
        }
    }
    
    int minIndexI = 0, minIndexJ = 0;
    float min = 10000000;
    for (int i = 0; i < 2; i++) {
        for (int j = 0; j < 2; j++) {
            varO[i][j] = varO[i][j] / (float)count[i][j];
//            if  (ImgUtil::DEBUG) 
//                cout << "var del cuadrante " << i << ", " << j << " = " << varO[i][j] * 180. / PI << endl;
            
            if (varO[i][j] < min) {
                min = varO[i][j];
                minIndexI = i;
                minIndexJ = j;
            }
        }
    }
    
//    if (ImgUtil::DEBUG) {
//        cout << "varianza minima = " << min << endl;
//        cout << "en el cuadrante " << minIndexI << ", " << minIndexJ << endl;
//    }
    
    float mAng = meanO[minIndexI][minIndexJ] + PI / 2.;    
    float m = tan(mAng);
    
//    if (ImgUtil::DEBUG)
//        cout << "slope =  " << m << endl;
    
    
    float b = (orH - posI - 1) - m * posJ; // equivale a y0 - m * x0 = b!
    
//    if (ImgUtil::DEBUG)
//        cout << "recta y = mx + b = " << " y = " << m << "x + " << b << endl;
    
    int lMask = 2 * influenceAreaLi + 1;
    bool **mask = MemUtil::create2DMatrix<bool>(lMask, lMask);
    bool **mask2 = MemUtil::create2DMatrix<bool>(lMask, lMask);
    for (int i = posI - influenceAreaLi; i < posI + influenceAreaLi + 1; i++) {
        for (int j = posJ - influenceAreaLi; j < posJ + influenceAreaLi + 1; j++) {
            
            int i2 = (i - (posI - influenceAreaLi));
            int j2 = (j - (posJ - influenceAreaLi)); // (j + 1)
            
            float parte1 = (m * (float)j + b) * core2;
            float parte2 = (float)(orH - i - 1) * core2;
            if (parte1 <= parte2) { // si la parte 2 es más grande (la y) estoy por encima!!
                mask[i2][j2] = true;
            }
            else {
                mask[i2][j2] = false;
            }
            mask2[i2][j2] = ! mask[i2][j2]; 
        }
    }
    
    if (ImgUtil::DEBUG)
        MemUtil::showMatrix(mask, lMask, lMask);
    
    MatrixXd omega2, omega4;
    float nonSquared = 1.0;
    computeFirstOrderOmegas(posI * blockW, posJ * blockW, omega2, omega4, nonSquared, O, mask);    
    VectorXd coefficients = computeCoefficientsFromOmegas(omega2, omega4, 1); // orden 1 para la prueba
     
     
        // Crear máscaras para reconstruir toda la imagen y para realizar la media
    bool **maskUpperZoneReconstruct = MemUtil::create2DMatrix<bool>(orH, orW);
    bool **maskLowerZoneReconstruct = MemUtil::create2DMatrix<bool>(orH, orW);
    
    
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {                                    
            float parte1 = (m * (float)j + b) * core2;
            float parte2 = (float)(orH - i - 1) * core2;
            if (parte1 <= parte2) { // si la parte 2 es más grande (la y) estoy por encima!!!
                maskUpperZoneReconstruct[i][j] = true;
                maskLowerZoneReconstruct[i][j] = false;
            }
            else {
                maskUpperZoneReconstruct[i][j] = false;
                maskLowerZoneReconstruct[i][j] = true;
            }
        }
    }
    
    
    
    
    int limits[4] = {0, orH, 0, orW}; 
    float **Onew = reconstructSPZoneOrientationsFromFirstOrderCoefficients(posI * blockW, posJ * blockW, coefficients, xCosNew, ySinNew, nonSquared, maskUpperZoneReconstruct, limits);
    MemUtil::freeMatrix(&Onew);
    
    nonSquared = 1.0;
    computeFirstOrderOmegas(posI * blockW, posJ * blockW, omega2, omega4, nonSquared, O, mask2);        
    coefficients = computeCoefficientsFromOmegas(omega2, omega4, 1); // orden 1 para la prueba
    Onew = reconstructSPZoneOrientationsFromFirstOrderCoefficients(posI * blockW, posJ * blockW, coefficients, xCosNew, ySinNew, nonSquared, maskLowerZoneReconstruct, limits);
    if (ImgUtil::DEBUG) {        
        string fileName2 = "OrientationRecontructed_piecewiseCore";
        if (core2 == -1)
            fileName2.append("_2");
        ImgUtil::writeOrImg(*xCosNew, *ySinNew, orW, orH, fileName2);
    }
    
    
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {
            Onew[i][j] = 0.5 * atan2((*ySinNew)[i][j], (*xCosNew)[i][j]);
            
            if (Onew[i][j] <= (-PI / 2.))
                Onew[i][j] += PI;
            else if (Onew[i][j] > (PI / 2.))
                Onew[i][j] -= PI;
        }
    }
    
    MemUtil::freeMatrix(&meanOcos);
    MemUtil::freeMatrix(&meanOsin);
    MemUtil::freeMatrix(&meanO);
    MemUtil::freeMatrix(&mask);
    MemUtil::freeMatrix(&mask2);
    
    MemUtil::freeMatrix(&maskUpperZoneReconstruct);
    MemUtil::freeMatrix(&maskLowerZoneReconstruct);
    
    // Obtener el cuadrante con menor varianza y tomar la media de sus orientaciones como orientacion predominante
    
    *m_line = m;
    
    return Onew;
}


void OrientationReconstructionLi::reconstructPieceWiseUniqueCore(int posI, int posJ, float ***xCosNew, float ***ySinNew, float ***Onew) {
    
    // pasar posI y posJ a bloques de Li blockW
    posI = (posI / blockW);
    posJ = (posJ / blockW);
    
    // Hay solo un core, línea horizontal que lo parte, se saca una máscara para cada zona y se reconstruyen 
    // de manera independiente, luego se media la zona de la línea para unificar todo bien
    
    
    int lMask = 2 * influenceAreaLi + 1;
    bool **maskUpperZone = MemUtil::create2DMatrix<bool>(lMask, lMask);
    bool **maskLowerZone = MemUtil::create2DMatrix<bool>(lMask, lMask);
    
    for (int i = max(0, posI - influenceAreaLi); i < min(posI + influenceAreaLi + 1, orH); i++) {
        for (int j = max(0, posJ - influenceAreaLi); j < min(posJ + influenceAreaLi + 1, orW); j++) {
            
            int i2 = (i - (posI - influenceAreaLi));
            int j2 = (j - (posJ - influenceAreaLi)); 
            
            if (i <= posI) { // está por encima!
                maskUpperZone[i2][j2] = true;
                maskLowerZone[i2][j2] = false;
            }
            else { // zona de abajo del core
                maskUpperZone[i2][j2] = false;
                maskLowerZone[i2][j2] = true;
            }
        }
    }
    
    if (ImgUtil::DEBUG) {
        cout << "UpperZone" << endl;
        MemUtil::showMatrix(maskUpperZone, lMask, lMask);
        cout << "LowerZone" << endl;
        MemUtil::showMatrix(maskLowerZone, lMask, lMask);
    }
    
    MatrixXd omega2, omega4;
    float nonSquared = 1.0;
    computeFirstOrderOmegas(posI * blockW, posJ * blockW, omega2, omega4, nonSquared, O, maskUpperZone);    
    VectorXd coefficients = computeCoefficientsFromOmegas(omega2, omega4, 1); // orden 1 para la prueba
    
    // Crear máscaras para reconstruir toda la imagen y para realizar la media
    bool **maskUpperZoneReconstruct = MemUtil::create2DMatrix<bool>(orH, orW);
    bool **maskLowerZoneReconstruct = MemUtil::create2DMatrix<bool>(orH, orW);
    
    
    bool **maskAverageZone = MemUtil::create2DMatrix<bool>(orH, orW);
    
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {                        
            if (i <= posI) { // está por encima!
                maskUpperZoneReconstruct[i][j] = true;
                maskLowerZoneReconstruct[i][j] = false;
            }
            else { // zona de abajo del core
                maskUpperZoneReconstruct[i][j] = false;
                maskLowerZoneReconstruct[i][j] = true;
            }
            
            if ((i >= posI - averageWLi) && (i <= posI + averageWLi))
                maskAverageZone[i][j] = true;
            else
                maskAverageZone[i][j] = false;
        }
    }
    
    if (ImgUtil::DEBUG) {
        cout << "maskUpperZoneReconstruct" << endl;
        MemUtil::showMatrix(maskUpperZoneReconstruct, orH, orW);
        cout << "maskLowerZoneReconstruct" << endl;
        MemUtil::showMatrix(maskLowerZoneReconstruct, orH, orW);
        cout << "maskAverageZone" << endl;
        MemUtil::showMatrix(maskAverageZone, orH, orW);
    }
    
    int limits[4] = {0, orH, 0, orW};
    
    *xCosNew = 0;
    *ySinNew = 0;
    *Onew = reconstructSPZoneOrientationsFromFirstOrderCoefficients(posI * blockW, posJ * blockW, coefficients, xCosNew, ySinNew, nonSquared, maskUpperZoneReconstruct, limits);
    if (ImgUtil::DEBUG) {        
        string fileName2 = "OrientationRecontructed_piecewiseUpper";
        ImgUtil::writeOrImg(*xCosNew, *ySinNew, orW, orH, fileName2);
    }
    
    MemUtil::freeMatrix(Onew);
    
    nonSquared = 1.0;
    computeFirstOrderOmegas(posI * blockW, posJ * blockW, omega2, omega4, nonSquared, O, maskLowerZone);    
    coefficients = computeCoefficientsFromOmegas(omega2, omega4, 1); // orden 1 para la prueba
    *Onew = reconstructSPZoneOrientationsFromFirstOrderCoefficients(posI * blockW, posJ * blockW, coefficients, xCosNew, ySinNew, nonSquared, maskLowerZoneReconstruct, limits);
    if (ImgUtil::DEBUG) {        
        string fileName2 = "OrientationRecontructed_piecewiseUpperAndLower";
        ImgUtil::writeOrImg(*xCosNew, *ySinNew, orW, orH, fileName2);
    }
    
    MemUtil::freeMatrix(Onew);
    
    *Onew = averageOrientations(xCosNew, ySinNew, maskAverageZone);    
    if (ImgUtil::DEBUG) {        
        string fileName2 = "OrientationRecontructed_piecewiseUpperAndLowerAveraged";
        ImgUtil::writeOrImg(*xCosNew, *ySinNew, orW, orH, fileName2);
    }
    
    
    MemUtil::freeMatrix(&maskUpperZoneReconstruct);
    MemUtil::freeMatrix(&maskLowerZoneReconstruct);
    MemUtil::freeMatrix(&maskAverageZone);
    MemUtil::freeMatrix(&maskUpperZone);
    MemUtil::freeMatrix(&maskLowerZone);
    
}


// Se asume 1 core y un delta, el delta siempre estríctamente debajo del core!!!!!!
int OrientationReconstructionLi::reconstructPieceWiseCoreDelta(int posIcore, int posJcore, int posIdelta, int posJdelta, float ***xCosNew, float ***ySinNew, float ***Onew) {
    
    // pasar posI y posJ a bloques de Li blockW
    int posI = (posIcore / blockW);
    int posJ = (posJcore / blockW);
    
    // Partir en cuatro cuadrantes, y sacaar la varianza en cada uno
    
    float m, b;
    
    float **xCosCore = 0, **ySinCore = 0;
    float **Ocore = reconstructPieceWiseCore(posIcore, posJcore, &xCosCore, &ySinCore, &m);
    b = (float)(orH - posI - 1) - m * (float)posJ;
     
    bool **maskUpperZoneReconstruct = MemUtil::create2DMatrix<bool>(orH, orW);
    bool **maskAverageZone = MemUtil::create2DMatrix<bool>(orH, orW);
    
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {                 
            
            float parte1 = (m * (float)j + b);
            float parte2 = (float)(orH - i - 1);
            if (parte1 <= parte2) { // si la parte 2 es más grande (la y) estoy por encima!!!
                maskUpperZoneReconstruct[i][j] = true;
            }
            else {
                maskUpperZoneReconstruct[i][j] = false;
            }
            
            if (!maskUpperZoneReconstruct[i][j] && 
                (i >= posI - influenceAreaLi) && (i <= posI + influenceAreaLi) && 
                (j >= posJ - influenceAreaLi) && (j <= posJ + influenceAreaLi)) {
                
//                maskUpperZoneReconstruct[i][j] = true;                
//                if ((i == posI - influenceAreaLi) || (i == posI + influenceAreaLi) ||
//                    (j == posJ - influenceAreaLi) || (j == posJ + influenceAreaLi))
//                    maskAverageZone[i][j] = true;
            }
            
            // si el punto está a menor distancia de la recta que averageWLi (o + 1), entonces se apuntar para average!
            float dist = abs(parte1  - parte2) / sqrt(m * m + 1);
            if (dist <= averageWLi + 1)
                maskAverageZone[i][j] = true;
            else
                maskAverageZone[i][j] = false;
        }
    } 
     
    if (ImgUtil::DEBUG) {
        cout << "Upper zone: " << endl;
        MemUtil::showMatrix(maskUpperZoneReconstruct, orH, orW);
        cout << "Average zone: " << endl;
        MemUtil::showMatrix(maskAverageZone, orH, orW);
        
        string a = "Upper";
        ImgUtil::writeGrayScaleImage(maskUpperZoneReconstruct, orH, orW, a);
    }
     
    // ver si el delta está debajo de la recta del core, sino se aborta la reconstrucción!!!!
    
    posI = (posIdelta / blockW);
    posJ = (posJdelta / blockW);
    
    float deltaPos = m * (float)posJ + b - (float)(orH - posI - 1);
        
    if (deltaPos <= 0) {    
        MemUtil::freeMatrix(&maskUpperZoneReconstruct);
        MemUtil::freeMatrix(&maskAverageZone);
        MemUtil::freeMatrix(&Ocore);
        MemUtil::freeMatrix(&xCosCore);
        MemUtil::freeMatrix(&ySinCore);
        return -1;
    }
    
    
    float **xCosDelta = 0, **ySinDelta = 0;
    int **retMask;
    float **Odelta = reconstructPieceWiseDelta(posIdelta, posJdelta, &xCosDelta, &ySinDelta, &retMask);
    
    MemUtil::freeMatrix(&retMask);
    
    // Es suficiente con utilizar la máscara del core y copiar los unos a los del delta, luego suavizar
    
    
//    MemUtil::freeMatrix(&Odelta);
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {
            if (maskUpperZoneReconstruct[i][j]) {
                Odelta[i][j] = Ocore[i][j];
                xCosDelta[i][j] = xCosCore[i][j];
                ySinDelta[i][j] = ySinCore[i][j];
            }
            else if (maskAverageZone[i][j]) {                
                xCosDelta[i][j] = (xCosDelta[i][j] + xCosCore[i][j]) / 2.0;
                ySinDelta[i][j] = (ySinDelta[i][j] + ySinCore[i][j]) / 2.0;
                Odelta[i][j] = 0.5 * atan2(ySinDelta[i][j], xCosDelta[i][j]);
                if (Odelta[i][j] <= (-PI / 2.))
                    Odelta[i][j] += PI;
                else if (Odelta[i][j] > (PI / 2.))
                    Odelta[i][j] -= PI;
            }
        }
    }    
    if (ImgUtil::DEBUG) {        
        string fileName2 = "OrientationRecontructed_piecewiseCoreDelta";
        ImgUtil::writeOrImg(xCosDelta, ySinDelta, orW, orH, fileName2);
    }
    MemUtil::freeMatrix(&Odelta);
    Odelta = averageOrientations(&xCosDelta, &ySinDelta, maskAverageZone);    
    if (ImgUtil::DEBUG) {        
        string fileName2 = "OrientationRecontructed_piecewiseCoreDeltaAveraged";
//        xCosDelta[posI][posJ] = 0;
//        ySinDelta[posI][posJ] = 0;
//        xCosDelta[posIcore/blockW][posJcore/blockW] = 0;
//        ySinDelta[posIcore/blockW][posJcore/blockW] = 0;
        
        ImgUtil::writeOrImg(xCosDelta, ySinDelta, orW, orH, fileName2);
    }
    
    
    MemUtil::freeMatrix(&Ocore);
    MemUtil::freeMatrix(&xCosCore);
    MemUtil::freeMatrix(&ySinCore);
    
    MemUtil::freeMatrix(&maskUpperZoneReconstruct);
    MemUtil::freeMatrix(&maskAverageZone);
        
    *Onew = Odelta;
    *xCosNew = xCosDelta;
    *ySinNew = ySinDelta;
    
    return 0;    
}



// Se asumen 2 cores y 2 deltas, los deltas siempre estríctamente debajo del core principal, que es el superior, y el cruce de sus asintotas por debajo de secundario!!!!!!
int OrientationReconstructionLi::reconstructPieceWise2Core2Delta(float ***xCosNew, float ***ySinNew, float ***Onew) {
    
    int posIcore1, posJcore1, posIcore2, posJcore2, posIdelta1, posIdelta2, posJdelta1, posJdelta2;
    
    /* Elegir core primario */
    posIcore1 = cores[0];
    posIcore2 = cores[2];
    if (posIcore1 < posIcore2) {
        posJcore1 = cores[1];
        posJcore2 = cores[3];
    }
    else {
        posIcore2 = cores[0];
        posJcore2 = cores[1];
        posIcore1 = cores[2];
        posJcore1 = cores[3];
    }
    // delta 1 a la izq
    posJdelta1 = deltas[1];
    posJdelta2 = deltas[3];
    if (posJdelta1 < posJdelta2) {
        posIdelta1 = deltas[0];
        posIdelta2 = deltas[2];
    }
    else {
        posIdelta2 = deltas[0];
        posJdelta2 = deltas[1];
        
        posIdelta1 = deltas[2];
        posJdelta1 = deltas[3];
    }
    
    // pasar posI y posJ a bloques de Li blockW
    int posI = (posIcore1 / blockW);
    int posJ = (posJcore1 / blockW);
    
    // Partir en cuatro cuadrantes, y sacaar la varianza en cada uno
    
    float m, b;
    
    float **xCosCore1 = 0, **ySinCore1 = 0;
    float **Ocore1 = reconstructPieceWiseCore(posIcore1, posJcore1, &xCosCore1, &ySinCore1, &m);
    b = (float)(orH - posI - 1) - m * (float)posJ;
     
    bool **maskUpperZoneReconstruct = MemUtil::create2DMatrix<bool>(orH, orW);
    bool **maskAverageZone = MemUtil::create2DMatrix<bool>(orH, orW);
    
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {                 
            
            float parte1 = (m * (float)j + b);
            float parte2 = (float)(orH - i - 1);
            if (parte1 <= parte2) { // si la parte 2 es más grande (la y) estoy por encima!!!
                maskUpperZoneReconstruct[i][j] = true;
            }
            else {
                maskUpperZoneReconstruct[i][j] = false;
            }
            
            // si el punto está a menor distancia de la recta que averageWLi (o + 1), entonces se apuntar para average!
            float dist = abs(parte1  - parte2) / sqrt(m * m + 1);
            if (dist <= averageWLi + 1)
                maskAverageZone[i][j] = true;
            else
                maskAverageZone[i][j] = false;
        }
    } 
     
    if (ImgUtil::DEBUG) {
        cout << "Upper zone: " << endl;
        MemUtil::showMatrix(maskUpperZoneReconstruct, orH, orW);
        cout << "Average zone: " << endl;
        MemUtil::showMatrix(maskAverageZone, orH, orW);
        
        string a = "Upper";
        ImgUtil::writeGrayScaleImage(maskUpperZoneReconstruct, orH, orW, a);
    }
    
    
    posI = (posIdelta1 / blockW);
    posJ = (posJdelta1 / blockW);    
    float deltaPos1 = m * (float)posJ + b - (float)(orH - posI - 1);
    
    posI = (posIdelta2 / blockW);
    posJ = (posJdelta2 / blockW);    
    float deltaPos2 = m * (float)posJ + b - (float)(orH - posI - 1);
        
    if ((deltaPos1 <= 0) || (deltaPos2 <= 0)) {    
        MemUtil::freeMatrix(&maskUpperZoneReconstruct);
        MemUtil::freeMatrix(&maskAverageZone);
        
        MemUtil::freeMatrix(&Ocore1);
        MemUtil::freeMatrix(&xCosCore1);
        MemUtil::freeMatrix(&ySinCore1);
        return -1;
    }
    
     
    posI = (posIcore2 / blockW);
    posJ = (posJcore2 / blockW);
    
    float **xCosCore2 = 0, **ySinCore2 = 0;
    float m2, b2;
    float **Ocore2 = reconstructPieceWiseCore(posIcore2, posJcore2, &xCosCore2, &ySinCore2, &m2, -1); //-1 por secundario core
    b2 = (float)(orH - posI - 1) - m2 * (float)posJ;
    
    
    
    float **xCosDelta1 = 0, **ySinDelta1 = 0;
    int **retMask1 = 0; 
    float **Odelta1 = reconstructPieceWiseDelta(posIdelta1, posJdelta1, &xCosDelta1, &ySinDelta1, &retMask1);
    float **xCosDelta2 = 0, **ySinDelta2 = 0;
    int **retMask2 = 0; 
    float **Odelta2 = reconstructPieceWiseDelta(posIdelta2, posJdelta2, &xCosDelta2, &ySinDelta2, &retMask2);
    
    // la base es la del core 2. Se mete el delta1 donde la mascara vale 1. luego el delta2 donde vale 0
    // Posteriormente 
    float **OdeltaComb = MemUtil::create2DMatrix<float>(orH, orW);
    float **xCosDeltaComb = MemUtil::create2DMatrix<float>(orH, orW);
    float **ySinDeltaComb = MemUtil::create2DMatrix<float>(orH, orW);
    
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {
            if (retMask1[i][j] == 1) {// izquierda del delta izquierdo 
                xCosDeltaComb[i][j] = xCosDelta1[i][j];
                ySinDeltaComb[i][j] = ySinDelta1[i][j];
                
                if ((j +  1 < orW) && (retMask1[i][j + 1] != 0)) {
                    maskAverageZone[i][j] = true;
                    maskAverageZone[i][j + 1] = true;
                    if (j + 2 < orW)
                        maskAverageZone[i][j + 2] = true;
                }
            }
            else if (retMask2[i][j] == 0) { // derecha del derecho
                xCosDeltaComb[i][j] = xCosDelta2[i][j];
                ySinDeltaComb[i][j] = ySinDelta2[i][j];
                
                if ((j - 1 >= 0) && (retMask2[i][j - 1] != 0)) {
                    maskAverageZone[i][j] = true;
                    maskAverageZone[i][j - 1] = true;
                    if (j - 2 >= 0)
                        maskAverageZone[i][j - 2] = true;
                }
            }
            else if ((retMask1[i][j] == 2) && (retMask2[i][j] == 2)) { // los dos abajo, la media                
                xCosDeltaComb[i][j] = (xCosDelta1[i][j] + xCosDelta2[i][j]) / 2.0;
                ySinDeltaComb[i][j] = (ySinDelta1[i][j] + ySinDelta2[i][j]) / 2.0;
                
                maskAverageZone[i][j] = true;
            }
            else if (retMask1[i][j] == 2) { // resto zona baja
                xCosDeltaComb[i][j] = xCosDelta1[i][j];
                ySinDeltaComb[i][j] = ySinDelta1[i][j];     
                
                if ((i - 1 >= 0) && (retMask1[i - 1][j] != 2)) {     
                    maskAverageZone[i - 1][j] = true;                    
                    if (i - 2 >= 0)
                        maskAverageZone[i - 2][j] = true;
                    maskAverageZone[i][j] = true;
                }
            }
            else if (retMask2[i][j] == 2) { // resto zona baja
                xCosDeltaComb[i][j] = xCosDelta2[i][j];
                ySinDeltaComb[i][j] = ySinDelta2[i][j];  
                
                if ((i - 1 >= 0) && (retMask2[i - 1][j] != 2)) {     
                    maskAverageZone[i - 1][j] = true;                    
                    if (i - 2 >= 0)
                        maskAverageZone[i - 2][j] = true;
                    maskAverageZone[i][j] = true;
                }
            }
            else { // centro para el core 2!
                xCosDeltaComb[i][j] = xCosCore2[i][j];
                ySinDeltaComb[i][j] = ySinCore2[i][j];            
                // si la distancia es cercana, average!!!
                
                    maskAverageZone[i][j] = true;
            }
            
            // si es zona del core 1, para el!
            if (maskUpperZoneReconstruct[i][j]) {
                xCosDeltaComb[i][j] = xCosCore1[i][j];
                ySinDeltaComb[i][j] = ySinCore1[i][j];                  
            }
            
            OdeltaComb[i][j] = 0.5 * atan2(ySinDeltaComb[i][j], xCosDeltaComb[i][j]);
            
            if (OdeltaComb[i][j] <= (-PI / 2.))
                OdeltaComb[i][j] += PI;
            else if (OdeltaComb[i][j] > (PI / 2.))
                OdeltaComb[i][j] -= PI;
        }
    }
    
    

    MemUtil::freeMatrix(&Ocore1);
    MemUtil::freeMatrix(&xCosCore1);
    MemUtil::freeMatrix(&ySinCore1);
    MemUtil::freeMatrix(&Ocore2);
    MemUtil::freeMatrix(&xCosCore2);
    MemUtil::freeMatrix(&ySinCore2);
    
    
    MemUtil::freeMatrix(&Odelta1);
    MemUtil::freeMatrix(&xCosDelta1);
    MemUtil::freeMatrix(&ySinDelta1);
    MemUtil::freeMatrix(&Odelta2);
    MemUtil::freeMatrix(&xCosDelta2);
    MemUtil::freeMatrix(&ySinDelta2);
    
    MemUtil::freeMatrix(&retMask1);
    MemUtil::freeMatrix(&retMask2);
    // ver si el delta está debajo de la recta del core, sino se aborta la reconstrucción!!!!
    
    if (ImgUtil::DEBUG) {        
        string fileName2 = "OrientationRecontructed_piecewise2Core2Delta";
        ImgUtil::writeOrImg(xCosDeltaComb, ySinDeltaComb, orW, orH, fileName2);
    }
//////////    
    MemUtil::freeMatrix(&OdeltaComb);
    OdeltaComb = averageOrientations(&xCosDeltaComb, &ySinDeltaComb, maskAverageZone);    
    if (ImgUtil::DEBUG) {        
        string fileName2 = "OrientationRecontructed_piecewise2Core2DeltaAveraged";
        
        
        ImgUtil::writeOrImg(xCosDeltaComb, ySinDeltaComb, orW, orH, fileName2);
    }
    
    
    MemUtil::freeMatrix(&maskUpperZoneReconstruct);
    MemUtil::freeMatrix(&maskAverageZone);
    *Onew = OdeltaComb;
    *xCosNew = xCosDeltaComb;
    *ySinNew = ySinDeltaComb;
    
    return 0;    
}

float **OrientationReconstructionLi::reconstructPieceWiseDelta(int posI, int posJ, float ***xCosNew, float ***ySinNew, int ***returnMask) {
    
    // pasar posI y posJ a bloques de Li blockW
    posI = (posI / blockW);
    posJ = (posJ / blockW);
    
    float m30 = tan(30. * PI / 180.);
    float m150 = tan(150. * PI / 180.);
    
    float b30 = (orH - posI - 1) - m30 * posJ; // equivale a y0 - m * x0 = b!
    float b150 = (orH - posI - 1) - m150 * posJ; // equivale a y0 - m * x0 = b!
    
    // Partir en cuatro cuadrantes, y sacaar la varianza en cada uno
    
    // Sacar la media de la orientacion en cada cuadrante
    float *meanOcos = MemUtil::createVector<float>(3);    
    float *meanOsin = MemUtil::createVector<float>(3);    
    float *meanO = MemUtil::createVector<float>(3);
    
    for (int i = max(0, posI - influenceAreaLi); i < min(orH, posI + influenceAreaLi + 1); i++) {
        for (int j = max(0, posJ - influenceAreaLi); j < min(orW, posJ + influenceAreaLi + 1); j++) {            
            float y30 = (m30 * (float)j + b30) - (orH - i - 1);
            float y150 = (m150 * (float)j + b150) - (orH - i - 1);
            
            
            if ((y30 <= 0) && (y150 <= 0)) {
                // arriba
                meanOcos[0] += cos(2. * O[i][j]);
                meanOsin[0] += sin(2. * O[i][j]);
            }
            else if ((y150 > 0) && (j <= posJ)) {
               // abajo izq
                meanOcos[1] += cos(2. * O[i][j]);
                meanOsin[1] += sin(2. * O[i][j]);                   
            }
            else {
               // abajo der
                meanOcos[2] += cos(2. * O[i][j]);
                meanOsin[2] += sin(2. * O[i][j]);                   
            }
            
        }
    }
    for (int i = 0; i < 3; i++) {
        meanO[i] = 0.5 * atan2(meanOsin[i], meanOcos[i]);
//        if (ImgUtil::DEBUG)
//                cout << "mean del cuadrante " << i << " = " << meanO[i] * 180. / PI << endl;
    }
    
    *xCosNew = 0;
    *ySinNew = 0;
    
    
    bool **mask01, **mask02, **mask12;
    float **Onew = reconstructDeltaFromAsymptotes(posI, posJ, meanO[0], meanO[2], 1, 1, xCosNew, ySinNew, &mask02);

    if (ImgUtil::DEBUG) {        
        string fileName2 = "OrientationRecontructed_delta_piecewise02";
        ImgUtil::writeOrImg(*xCosNew, *ySinNew, orW, orH, fileName2);      
    }
    MemUtil::freeMatrix(&Onew);
    
    Onew = reconstructDeltaFromAsymptotes(posI, posJ, meanO[1], meanO[2], 1, -1, xCosNew, ySinNew, &mask12);

    if (ImgUtil::DEBUG) {         
        string fileName2 = "OrientationRecontructed_delta_piecewise12";
        ImgUtil::writeOrImg(*xCosNew, *ySinNew, orW, orH, fileName2);      
    }
    MemUtil::freeMatrix(&Onew);
    
    Onew = reconstructDeltaFromAsymptotes(posI, posJ, meanO[0], meanO[1], -1, -1, xCosNew, ySinNew, &mask01);

    if (ImgUtil::DEBUG) {        
        string fileName2 = "OrientationRecontructed_delta_piecewise01";
        ImgUtil::writeOrImg(*xCosNew, *ySinNew, orW, orH, fileName2);      
    }
    //MemUtil::freeMatrix(&Onew);
    
    int **maskCombined = MemUtil::create2DMatrix<int>(orH, orW);
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {
            Onew[i][j] = 0.5 * atan2((*ySinNew)[i][j], (*xCosNew)[i][j]);
            
            if (Onew[i][j] <= (-PI / 2.))
                Onew[i][j] += PI;
            else if (Onew[i][j] > (PI / 2.))
                Onew[i][j] -= PI;
            
            // combinar máscaras
            if (mask02[i][j])
                maskCombined[i][j] = 0;
            else if (mask12[i][j])
                maskCombined[i][j] = 2;
            else  if (mask01[i][j])
                maskCombined[i][j] = 1;
            else                
                maskCombined[i][j] = -1;
        }
    }
    
    MemUtil::freeMatrix(&meanOcos);
    MemUtil::freeMatrix(&meanOsin);
    MemUtil::freeMatrix(&meanO);
    
    
    MemUtil::freeMatrix(&mask02);
    MemUtil::freeMatrix(&mask12);
    MemUtil::freeMatrix(&mask01);
    
    *returnMask = maskCombined;
    
    return Onew;
}


float **OrientationReconstructionLi::reconstructDeltaFromAsymptotes(int posI, int posJ, float O1, float O2, float sign1, float sign2, float ***xCosNew, float ***ySinNew, bool ***returnMask) {
    
    float A, B, C, D;
    A = 1;
    C = 1;
    
    B = -A / tan(O1);
    D = -C / tan(O2);
    if (tan(O1) == 0)
        B = 0;
    if (tan(O2) == 0)
        D = 1000;
    
//    if (ImgUtil::DEBUG) {
//        cout << " asintota 1 = " << A << "(x - " << posJ << ") + " << B << "(y - " << (orH - posI - 1) << ") = 0" << endl;
//        cout << " asintota 2 = " << C << "(x - " << posJ << ") + " << D << "(y - " << (orH - posI - 1) << ") = 0" << endl;
//    }
    
    A *= sign1;
    B *= sign1;
    C *= sign2;
    D *= sign2;
    
    
    
    int lMask = 2 * influenceAreaLi + 1;
    bool **mask = MemUtil::create2DMatrix<bool>(lMask, lMask);
    for (int i = max(0, posI - influenceAreaLi); i < min(orH, posI + influenceAreaLi + 1); i++) {
        for (int j = max(0, posJ - influenceAreaLi); j < min(orW, posJ + influenceAreaLi + 1); j++) {
            
            float asin1 = A * (j - posJ) + B * (posI - i);
            float asin2 = C * (j - posJ) + D * (posI - i);
            
            int i2 = (i - (posI - influenceAreaLi));
            int j2 = (j - (posJ - influenceAreaLi));
            
            if ((asin1 > 0) && (asin2 > 0))
                mask[i2][j2] = true;
            else
                mask[i2][j2] = false;
        }
    }    
       
//    if (ImgUtil::DEBUG) {
//        cout << "Máscara: " << endl;
//        MemUtil::showMatrix(mask, lMask, lMask);
//    }
    
    float P, Q;
    float theta;
    
    theta = atan2( B * D - A * C + sqrt((B * D - A * C) * (B * D - A * C) + (B * C + A * D) * (B * C + A * D)), B * C + A * D);
    
    
    float sinT, cosT;
    sinT = sin(theta);
    cosT = cos(theta);    
    
    P = (A * cosT + B * sinT) * (C * cosT + D * sinT);
    Q = (B * cosT - A * sinT) * (D * cosT - C * sinT);
        
    float a, b, c, d;
    a = (Q - P) * sinT * cosT;
    b = -Q * cosT * cosT - P * sinT * sinT;
    c = Q * sinT * sinT + P * cosT * cosT;
    d = (P - Q) * sinT * cosT;
    
//    if (ImgUtil::DEBUG)
//        cout << "a = " << a << " b = " << b << " c = " << c  << " d = " << d << endl;
    
    float norm = sqrt(a*a+b*b+c*c+d*d);
    
    a /= norm;
    b /= norm;
    c /= norm;
    d /= norm;
    
//    if (ImgUtil::DEBUG) {
//        cout << "a = " << a << " b = " << b << " c = " << c  << " d = " << d << endl;
//        cout << " are parameters OK? sum to 1? = " << sqrt(a*a+b*b+c*c+d*d) << endl;
//    }
        
    MatrixXd L4(4, 1);
    L4 << a, b, c, d;
    
    VectorXd coefficients(4);
    coefficients << L4.col(0);
    
    MemUtil::freeMatrix(&mask);
    
    mask = MemUtil::create2DMatrix<bool>(orH, orW);
    int limits[4] = {0, orH, 0, orW}; 
    for (int i = limits[0]; i < limits[1]; i++) {
        for (int j = limits[2]; j < limits[3]; j++) {
            
            float asin1 = A * (j - posJ) + B * (posI - i);
            float asin2 = C * (j - posJ) + D * (posI - i);
            
            if ((asin1 > 0) && (asin2 > 0))
                mask[i][j] = true;
            else
                mask[i][j] = false;
        }
    }  
//    if (ImgUtil::DEBUG) {
//        cout << "Máscara: " << endl;
//        MemUtil::showMatrix(mask, limits[1], limits[3]);
//    }
    
    float **Onew = reconstructSPZoneOrientationsFromFirstOrderCoefficientsABCD(posI * blockW, posJ * blockW, coefficients, xCosNew, ySinNew, mask, limits);

    *returnMask = mask;
    // MemUtil::freeMatrix(&mask);
    
    return Onew;
    
}


float **OrientationReconstructionLi::averageOrientations(float ***xCosNew, float ***ySinNew, bool **maskAverageZone) {
    float **Onew = MemUtil::create2DMatrix<float>(orH, orW);
    
    float **xCosAux = MemUtil::copy2DMatrix<float>(*xCosNew, orH, orW);
    float **ySinAux = MemUtil::copy2DMatrix<float>(*ySinNew, orH, orW);
    
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {
            
            if (maskAverageZone[i][j]) {
                // do average
                xCosAux[i][j] = 0;
                ySinAux[i][j] = 0;
                for (int u = max(i - averageWLi, 0); u < min(i + averageWLi + 1, orH); u++) {
                    for (int v = max(j - averageWLi, 0); v < min(j + averageWLi + 1, orW); v++) {
                        xCosAux[i][j] += (*xCosNew)[u][v];
                        ySinAux[i][j] += (*ySinNew)[u][v];
                    }
                }
            }
                
            Onew[i][j] = 0.5 * atan2(ySinAux[i][j], xCosAux[i][j]);
            if (Onew[i][j] > PI / 2.)
                Onew[i][j] -= PI;
            else if (Onew[i][j] <= PI / 2.)
                Onew[i][j] += PI;
            
            xCosAux[i][j] = cos(2.0 * Onew[i][j]);
            ySinAux[i][j] = sin(2.0 * Onew[i][j]);            
            
        }
    }
    
    MemUtil::freeMatrix(xCosNew);
    MemUtil::freeMatrix(ySinNew);
    
    *xCosNew = xCosAux;
    *ySinNew = ySinAux;
    
    
    return Onew;
    
}

void OrientationReconstructionLi::formTrainingData(float **xCosPred, float **ySinPred, float **Opred, float **coherenceTrain) {
    
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {
            
            if (coherence[i][j] >= 0.5) {
                Opred[i][j] = O[i][j];
                xCosPred[i][j] = xCos[i][j];
                ySinPred[i][j] = ySin[i][j];
            }
            else
                coherenceTrain[i][j]  = 0.3; // coherenceTrain es coherence al inicio! con las otras ocurre lo contrario           
        }
    }
    
    if (ImgUtil::DEBUG) {        
        string fileName2 = "OrientationTrain";
        ImgUtil::writeOrImg(xCosPred, ySinPred, orW, orH, fileName2);      
    }
    
}


float **OrientationReconstructionLi::imageAlignement(float **matrix, int h, int w, vector<float> coreAngles){
//    if (ImgUtil::DEBUG)
//        cout << "En IMAGE_ALIGNEMENT " << endl;
    
    int indexCore = extractCenterCore();
    vector<int> centro;
    float angAGirar = /*3.14;*/ ((PI / 2.) + coreAngles[indexCore]);

    if (ImgUtil::DEBUG){
        cout << "DIRECCION DEL CENTRO = " << coreAngles[indexCore] * 180. / PI << endl;
        cout << "ANGULO A GIRAR INICIAL = " << angAGirar * 180. / PI << endl;
    }
    
    if (coreAngles[indexCore] > 0){
        angAGirar =  angAGirar - PI;
    }
    
    
    float **imgRot = MemUtil::create2DMatrix<float>(h + extendedBorderLi, w + extendedBorderLi);
    
    centro.push_back( (cores[2 * indexCore ] + extendedBorderLi/2) );
    centro.push_back( (cores[2 * indexCore + 1] + extendedBorderLi/2) );
    
    if (ImgUtil::DEBUG){
        cout << "ANGULO A GIRAR = " << angAGirar * 180. / PI << endl;
        cout << "Elegido core " << indexCore + 1 << " con coordenadas: " << endl;
        cout << centro[0] << endl;
        cout << centro[1] << endl;
    }
    
    float indexI;
    float indexJ;
    
    for (int i = 0; i < h + extendedBorderLi; i++){
        for (int j = 0; j < w + extendedBorderLi; j++){
            indexI = cos(angAGirar) * (i - centro[0]) - sin(angAGirar) * (j - centro[1]) + centro[0];
            indexJ = sin(angAGirar) * (i - centro[0]) + cos(angAGirar) * (j - centro[1]) + centro[1];
            vector<int> vecinos = extractNeighbours(indexI, indexJ, h, w);
            vector<float> pesos = extractWeights(indexI, indexJ, vecinos);
            
            //
            imgRot[i][j] = calculateRotation(matrix, vecinos, pesos);
            //
            
            vecinos.clear();
            pesos.clear();
        }
    }
    imgRot[centro[0]][centro[1]] = matrix[centro[0]][centro[1]];
    
    if (ImgUtil::DEBUG) {
        string fileName2 = "RotatedImage";
        ImgUtil::writeGrayScaleImage(imgRot, h + extendedBorderLi, w + extendedBorderLi, fileName2, 1);
    }
    
    // Rotar cores y deltas
    int indI, indJ;

    // Rotar cores y deltas
     angAGirar = -angAGirar;
     for (int i = 0; i < cores.size() / 2; i++) {
         indI = round(cos(angAGirar) * (cores[i * 2] - centro[0]) - sin(angAGirar) * (cores[i * 2 + 1] - centro[1]) + centro[0]);
         indJ = round(sin(angAGirar) * (cores[i * 2] - centro[0]) +cos(angAGirar) * (cores[i * 2 + 1] - centro[1]) + centro[1]);
         cores[i * 2] = indI;
         cores[i * 2 + 1] = indJ;
         if ( (cores[i * 2] < 0) || (cores[i * 2] >= h) || (cores[i * 2 + 1] < 0) || (cores[i * 2 + 1] >= w) ){
             cores.erase(cores.begin() + i * 2 + 1);
             cores.erase(cores.begin() + i * 2);
             i--;
         }
     }
     for (int i = 0; i < deltas.size() / 2; i++) {
         indI = round(cos(angAGirar) * (deltas[i * 2] - centro[0]) - sin(angAGirar) * (deltas[i * 2 + 1] - centro[1]) + centro[0]);
         indJ = round(sin(angAGirar) * (deltas[i * 2] - centro[0]) + cos(angAGirar) * (deltas[i * 2 + 1] - centro[1]) + centro[1]);
         deltas[i * 2] = indI;
         deltas[i * 2 + 1] = indJ;
         if ( (deltas[i * 2] < 0) || (deltas[i * 2] >= h) || (deltas[i * 2 + 1] < 0) || (deltas[i * 2 + 1] >= w) ){
             deltas.erase(deltas.begin() + i * 2 + 1);
             deltas.erase(deltas.begin() + i * 2);
             i--;
         }
     }
    
    if (ImgUtil::DEBUG) {
        string fileName = "RotatedImagecoresAndDeltas";
        ImgUtil::writeSPImage(imgRot, h + extendedBorderLi, w + extendedBorderLi, deltas, cores, 1, fileName);
    }
    
    float **imgCrop = imageCropping(imgRot, h + extendedBorderLi, w + extendedBorderLi, centro, croppingDimHLi, croppingDimWLi);
    
    MemUtil::freeMatrix(&imgRot);
    
    centro.clear();
    
    return imgCrop;
}

float OrientationReconstructionLi::calculateRotation(float **imagen, vector<int> vecinos, vector<float> pesos){
    
//    if (ImgUtil::DEBUG)
//        cout << "En CALCULATE_ROTATION " << endl;
    
    float media = 0;
    
    if (vecinos[0] == -1)
        return 0;
    
    for (int i = 0; i < vecinos.size()/2; i++){
        media += imagen[vecinos[2 * i]][vecinos[2 * i + 1]] * pesos[i];
    }
    return media;
}

vector<float> OrientationReconstructionLi::extractWeights(float indexI, float indexJ, vector<int> vecinos){

//    if (ImgUtil::DEBUG)
//        cout << "En EXTRACT_WEIGHTS " << endl;
    
    vector<float> pesosAux;
    vector<float> pesos;
    float total = 0;
    
//    if (( (vecinos[0] == indexI) && (vecinos[1] == indexJ)) || (vecinos[0] == -1)) {
//        pesos.push_back(1);
//        return pesos;
//    }
    
    
    for (int i = 0; i < vecinos.size()/2; i++){     
        float dist = sqrt((indexI - vecinos[i * 2])*(indexI - vecinos[i * 2])+(indexJ - vecinos[i * 2 + 1])*(indexJ - vecinos[i * 2 + 1]));       
        if (dist == 0)
            dist = 10000000;
        total += 1./dist; 
        pesosAux.push_back( 1./dist );
    }  
  
    //Normalized weights
    for (int i = 0; i < pesosAux.size(); i++)
        pesos.push_back( pesosAux[i]/total );
    
    pesosAux.clear();
   
    return pesos;
}

vector<int> OrientationReconstructionLi::extractNeighbours(float indexI, float indexJ, int h, int w){
    
//    if (ImgUtil::DEBUG)
//        cout << "En EXTRACT_NEIGHBOURS " << endl;
    
    vector<int> vecinos;
    
    if ( (indexI < 0) || (indexJ < 0) || (indexI >= h) || (indexJ >= w)) {
        vecinos.push_back(-1);
        vecinos.push_back(-1);
    }
    else{
        vecinos.push_back((int)indexI);
        vecinos.push_back((int)indexJ);

        if ((indexI + 1 < h) && (indexJ + 1 < w)) {
            vecinos.push_back((int)indexI + 1);
            vecinos.push_back((int)indexJ + 1);
        }
        if (indexJ + 1 < w){
            vecinos.push_back((int)indexI);
            vecinos.push_back((int)indexJ + 1);
        }
        if (indexI + 1 < h){
            vecinos.push_back((int)indexI + 1);
            vecinos.push_back((int)indexJ);
        }
    }
    return vecinos;
}

int OrientationReconstructionLi::extractCenterCore(){
    
//    if (ImgUtil::DEBUG)
//        cout << "En EXTRACT_CENTER_CORE " << endl;
    
    if (cores.size() == 2) // 1 core
        return 0;
    else{ // 2 cores
        if (deltas.size() == 2){ // 1 delta
            float dist1 = sqrt((cores[0] - deltas[0])*(cores[0] - deltas[0])+(cores[1] - deltas[1])*(cores[1] - deltas[1]));
            float dist2 = sqrt((cores[2] - deltas[0])*(cores[2] - deltas[0])+(cores[3] - deltas[1])*(cores[3] - deltas[1]));
            return (dist1 > dist2) ? 1 : 0;
        }
        else if (deltas.size() == 4) { // 2 deltas
            float dist11 = sqrt((cores[0] - deltas[0])*(cores[0] - deltas[0])+(cores[1] - deltas[1])*(cores[1] - deltas[1]));
            float dist21 = sqrt((cores[2] - deltas[0])*(cores[2] - deltas[0])+(cores[3] - deltas[1])*(cores[3] - deltas[1]));
            float dist12 = sqrt((cores[0] - deltas[2])*(cores[0] - deltas[2])+(cores[1] - deltas[3])*(cores[1] - deltas[3]));
            float dist22 = sqrt((cores[2] - deltas[2])*(cores[2] - deltas[2])+(cores[3] - deltas[3])*(cores[3] - deltas[3]));
            return ((dist11 + dist12)/2. > (dist21 + dist22)/2.) ? 1 : 0;
        }
        else {
            return cores[0] > cores[2] ? 0 : 1;
        }
            
    }
}

float **OrientationReconstructionLi::imageCropping(float **img, int h, int w, vector<int> centro, int hC, int wC){
    
    float **imgCrop = MemUtil::create2DMatrix<float>(hC, wC);
    int ajusteH = (hC % 2 == 0) ? -1 : 0;
    int ajusteW = (wC % 2 == 0) ? -1 : 0;
    int newCenterH = centro[0] + ((hC / 3.) / 2.); 
    int newCenterW = centro[1]; 

    
    for (int i = 0; i < hC; i++)
        for (int j = 0; j < wC; j++){
            if ((newCenterH - (hC / 2 + ajusteH) + i < 0) || (newCenterH - (hC / 2 + ajusteH) + i >= h)
                    || (newCenterW - (wC / 2 + ajusteW) + j < 0) || (newCenterW - (wC / 2 + ajusteW) + j >= w))
                imgCrop[i][j] = 0;
            else
                imgCrop[i][j] = img[newCenterH - (hC / 2 + ajusteH) + i][newCenterW - (wC / 2 + ajusteW) + j];
        }
    
    if (ImgUtil::DEBUG){
        string fileName2 = "imageCropped";
        ImgUtil::writeGrayScaleImage(imgCrop, hC, wC, fileName2, 1);
    }
    
    for (int i = 0; i < cores.size() / 2; i++) {
        cores[i * 2] = cores[i * 2] - (newCenterH - (hC / 2 + ajusteH));
        cores[i * 2 + 1] = cores[i * 2 + 1] - (newCenterW - (wC / 2 + ajusteW)); 
        if ( (cores[i * 2] < 0) || (cores[i * 2] >= hC) || (cores[i * 2 + 1] < 0) || (cores[i * 2 + 1] >= wC) ){
            cores.erase(cores.begin() + i * 2 + 1);
            cores.erase(cores.begin() + i * 2);
            i--;
        }
    }
    for (int i = 0; i < deltas.size() / 2; i++) {
        deltas[i * 2] = deltas[i * 2] - (newCenterH - (hC / 2 + ajusteH));
        deltas[i * 2 + 1] = deltas[i * 2 + 1] - (newCenterW - (wC / 2 + ajusteW)); 
        if ( (deltas[i * 2] < 0) || (deltas[i * 2] >= hC) || (deltas[i * 2 + 1] < 0) || (deltas[i * 2 + 1] >= wC) ){
            deltas.erase(deltas.begin() + i * 2 + 1);
            deltas.erase(deltas.begin() + i * 2);
            i--;
        }
    }
    
    if (ImgUtil::DEBUG) {
        string fileName = "CropedImagecoresAndDeltas";
        ImgUtil::writeSPImage(imgCrop, hC, wC, deltas, cores, 1, fileName);
    }
    
    
    return imgCrop;
}

float OrientationReconstructionLi::orientationDifference(float **Oorig, float **Omodel, float **coherence, int orH, int orW, int smallH, int smallW, int centralH, int centralW){
    
    float diff = 0;
    float cont = 0;
    int ajusteH = (smallH % 2 == 0) ? -1 : 0; // Para centrar bien las orientaciones con dimensiones smallH y smallW pares
    int ajusteW = (smallW % 2 == 0) ? -1 : 0;
    
    for (int i = 0; i < smallH; i++){
        for (int j = 0; j < smallW; j++){
            
            int a = centralH - (smallH / 2 + ajusteH) + i;
            int b = centralW - (smallW / 2 + ajusteW) + j;
            
            if ((a >= 0) && (a < orH) && (b >= 0) && (b < orW)){

                float dif = min(abs(Oorig[a][b] - Omodel[a][b]), PI - abs(Oorig[a][b] - Omodel[a][b]));
            	diff += sin(dif) * sin(dif) * (1 - coherence[a][b]);                

                //diff += sin(Oorig[a][b] - Omodel[a][b]) * sin(Oorig[a][b] - Omodel[a][b]) * (1 - coherence[a][b]);
                cont++;
            }

            
        }
    }
    diff /= cont;
    
    return diff;
}

float OrientationReconstructionLi::meanCoherence(float **coherence, int orH, int orW){
    
    float mean = 0;
    
    for (int i = 0; i < orH; i++){
        for (int j = 0; j < orW; j++){
            mean += coherence[i][j];
        }
    }
    mean /= orH * orW;
    
    return mean;
}

float *OrientationReconstructionLi::createFeatureVector(VectorXd coefficients, int order, float magCore, float magDelta){
    
    float coreRes = (cores.size() / 2) * magCore;
    float deltaRes = (deltas.size() / 2) * magDelta;
    int lenght = (order + 1) * (order + 2);
    int primaryCore = extractCenterCore();
    
    dimFV = lenght - 2 + 4;
    
    float *fv = MemUtil::createVector<float>(lenght - 2 + 4);

    for (int i = 1; i < lenght / 2; i++){ //No incluimos el primero en el fv ni el de en medio
        fv[i - 1] = coefficients(i);
        fv[(i - 1) + lenght / 2 - 1] = coefficients(i + lenght / 2);
    }
    
    fv[lenght - 2] = coreRes;
    fv[lenght - 1] = deltaRes;
    
    if (deltas.size() == 2){       
        fv[lenght] = (PI / 2.) + atan2(cores[2 * primaryCore] - deltas[0], cores[2 * primaryCore + 1] - deltas[1]);
        fv[lenght +1] = 0;
    }
    else if (deltas.size() == 4){
        fv[lenght] = (PI / 2.) + atan2(cores[2 * primaryCore] - deltas[0], cores[2 * primaryCore + 1] - deltas[1]);
        fv[lenght + 1] = (PI / 2.) + atan2(cores[2 * primaryCore] - deltas[2], cores[2 * primaryCore + 1] - deltas[3]);
    }
    else{       
        fv[lenght] = 0;
        fv[lenght + 1] = 0;
    }
    
    return fv;
}