/* 
 * \file        PCASYSManager.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * Header file for PCASYSManager  class
 *
 * 
 */

#ifndef PCASYSMANAGER_H
#define	PCASYSMANAGER_H


#include "ImgUtil.h"

using namespace std;

/**
 * \class PCASYSManager
 *
 * The PCASYSManager class manages the conexion with PCASYS from NBIS Package. 
 * It enables to segment, enhance, registrate and extract the orientation image and the corresponding feature vectors.
 */
class PCASYSManager {
    
public:
    
    /** Default constructor */
    PCASYSManager();
    
    /** Constructor */
    PCASYSManager(int w, int h);
    
    /**
     * Copy constructor
     * \param other Object to copy from
     */
    PCASYSManager(const PCASYSManager& orig);
    
    /** Default destructor */
    ~PCASYSManager();
    
     /**
     * It returns the width of the segmented image
     * \return the width of the segmented image
     */
    int getSW() const;
    
    /**
     * It returns the height of the segmented image
     * \return the height of the segmented image
     */
    int getSH() const;
    
     /**
     * It returns the width of the segmented foreground image
     * \return the width of the segmented foreground image
     */
    int getSfgW() const;
    
    /**
     * It returns the height of the segmented foreground image
     * \return the height of the segmented foreground image
     */
    int getSfgH() const;
    
    /**
     * It returns the width of the averaged orientation image
     * \return the width of the averaged orientation image
     */
    int getOrW() const;
    
    /**
     * It returns the height of the averaged orientation image
     * \return the height of the averaged orientation image
     */
    int getOrH() const;
    
    /**
     * It returns the width of the averaged orientation image for pseudoridge analyzer!
     * \return the width of the  orientation image for pseudoridge analyzer!
     */
    int getOrW2() const;
    
    /**
     * It returns the height of the averaged orientation image for pseudoridge analyzer!
     * \return the height of the orientation image for pseudoridge analyzer!
     */
    int getOrH2() const;
    
    /**
     * It returns the width of the transformation matrix
     * \return the width of the transformation matrix
     */
    int getTfW() const;
    
    /**
     * It returns the height of the transformation matrix
     * \return the height of the transformation matrix
     */
    int getTfH() const;
    
    /**
     * It segments a FingerprintImg
     * \param image Fingerprint image to be segmented
     * \param w width of the image
     * \param h height of the image
     * \return the segmented image
     */
    unsigned char **segment(unsigned char *image, int w, int h, unsigned char ***segmentedFg);
    
    /**
     * It enhances a FingerprintImg
     * \param image Fingerprint image to be enhanced
     * \param w width of the image
     * \param h height of the image
     * \return the segmented image
     */
    unsigned char **enhance(unsigned char **image, int w, int h);
    
    /**
     * It computes the orientation image of a FingerprintImg
     * \param image Fingerprint image to extract the orientation image
     * \param w width of the image
     * \param h height of the image
     * \param avOrX average orientations X component (to be returned)
     * \param avOrY average orientations Y component (to be returned)
     * \return the segmented image
     */
    char **computeOrientations(unsigned char **image, int w, int h, float ***avrors_x, float ***avrors_y);
    
    /**
     * It creates the image picture of the orientation image to be shown
     * \param avOrX average orientations X component
     * \param avOrY average orientations Y component
     * \param wIn width of the orientations image to be converted
     * \param hIn height  of the orientations image to be converted
     * \param wOut width of the output orientation image
     * \param HOut height of the output orientation image
     * \return the orientation image picture (by columns)
     */
    unsigned char *getOrientationImagePicture(float **avOrX, float **avOrY, int wIn, int hIn, int *wOut, int *hOut);
    
    /**
     * It computes the core of the fingerprint using r92a algorithm
     * \param avOrX average orientations X component
     * \param avOrY average orientations Y component
     */
    void computeCore_r92a(float **avOrX, float **avOrY, int *coreX, int *coreY);
    
    /**
     * It computes the core of the fingerprint using r92a algorithm
     * \param avOrX average orientations X component
     * \param avOrY average orientations Y component
     * \param wOr
     * \param hOr
     */
    void computeCore_r92a(float **avOrX, float **avOrY, int wOr, int hOr, int *coreX, int *coreY);
    
    /**
     * It computes the core of the fingerprint using r92a algorithm
     * \param avOrX average orientations X component
     * \param avOrY average orientations Y component
     * \param wOr
     * \param hOr
     * \param blockSize used to redimension avOrX and avOrY to a 32 x 32
     */
    void computeCore_r92a(float **avOrX, float **avOrY, int wOr, int hOr, int *coreX, int *coreY, int blockSize);
    
    /**
     * Computes the registered orientation image using the core previously computed
     * that is, it computes, from the pixelwise-orientations made earlier, a new,
     * "registered" 28x30-vector grid of average orientations, by using 
     *  averaging squares that are shifted according to the r92 resultcore.
     * \param pixelOr pixelwise orientation image made earlier by pcasys
     * \param coreX x-position of the core
     * \param coreY y-position of the core
     * \param avOrRegX output average orientaitons registered X
     * \param avOrRegY output average orientations registered Y
     * \return -1 if an error has ocurred. 0 otherwise
     */
    void registerOrientations(char **pixelOr, int coreX, int coreY, float ***avOrRegX, float ***avOrRegY);

    
    /**
     * It stracts the feature vector using the registered orientation image
     * The feature vector has 64 o 128 components depending on the pcasys 
     * configuration parameters (RNN or PNN)
     * \param avOrRegX output average orientaitons registered X
     * \param avOrRegY output average orientations registered Y
     * \return the feature vector
     */
    float *extractOrientationFeatureVector(float **avOrRegX, float **avOrRegY);

    /**
     * It computes the orientation image (x and y) from the original pixel orientation
     * image; in this case, the blocks are overlapped! (it is bigger)
     * \param pixelOr   the previosly extracted pixel orientations
     * \param avOrX2  the average orientations image component X
     * \param avOrY2  the average orientations image component Y
     */
    void computeOrientationsForPseudoRidge(char **pixelOr, float ***avOrX2, float ***avOrY2);
    
    /**
     * It computes the smoothed orientation image (x and y) from the orientations image for pseudo ridges
     * It uses the function for finding upwardsconcave lobes
     * \param segmentedFgPCASYS  the original segmentation of PCASYS
     * \param avOrX2  the average orientations image component X
     * \param avOrY2  the average orientations image component Y
     * \return wether has conup or not
     */
    bool smoothOrientationsForPseudoRidge(unsigned char **segmentedFgPCASYS, float **avOrX2, float **avOrY2);


    /** UNSUPPORTED!
     * It computes the smoothed orientation image (x and y) from the orientations image for pseudo ridges
     * It uses the function for finding upwardsconcave lobes
     * \param segmentedFgPCASYS   the original segmentation of PCASYS
     * \param avOrX2  the average smoothed orientations image component X
     * \param avOrY2  the average smoothed orientations image component Y
     * \return the pseudo ridge image
     */
    unsigned char **getPseudoRidges(unsigned char **segmentedFgPCASYS, float **avOrX2, float **avOrY2);

    /** UNSUPPORTED!**/
    unsigned char **printHasConup(float **us, float **vs, const int maxturn,
         const int max_tilt, const int min_side_turn, const int initi_s,
         const int initi_e, const int initj_s, const int initj_e,
         const int maxsteps_eachdir, char **bad, float **dg, const int w,
         const int h);
    
private:
    /**
     * It converts the X and Y average orientations image to directions and magnitude images
     * \param x average orientations X component
     * \param y average orientations Y component
     * \param dg directions degrees
     * \param mag magnitude of the vectors simbolizing the direction
     * \param w width of the orientation image
     * \param h height of the orientation image
     * \param top_ret maximum magnitude?
     */
    void xy_to_dmt(float **x, float **y, float **dg, float **mag, int w, int h, float *top_ret);
    
    /**
     * It translates teh directions degrees and magnitudes to a image which can be written
     * \param dg direction degrees
     * \param mag magnitudes of the direction bars
     * \param aw orientation image width
     * \param ah orientation image height
     * \param top maximum magnitude?
     * \param pic pointer to the image which is constructed
     * \param w width for the final image
     * \param h heigh for the final image
     */
    void dmt_to_pic(float **dg, float **mag, int aw, int ah, float top, unsigned char *pic, int w, int h);
    
private:
    /* Attributes needed by PCASYS */
    SGMNT_PRS sgmnt_prs;                        //!< Parameters for the segmentation
    ENHNC_PRS enhnc_prs;                        //!< Parameters for the image enhancement (Discrete Fourier Transform)
    int rors_slit_range_thresh;                 //!< Parameter for the orientation maps
    float r92a_discard_thresh;                  //!< Parameters for the core detection using R92a algorithm
    RGAR_PRS rgar_prs;                          //!< Parameters for the orientation registration
    int trnsfrm_nrows_use;                      //!< Size of the feature vector of PCASYS
    int pnn_mlp;                                //!< Type of the classifier to be used
    PNN_PRS pnn_prs;                            //!< Parameters for the Probabilistic Neural Network
    MLP_PARAM mlp_prs;                          //!< Parameters for the Multilayer Neural Network
    PSEUDO_PRS pseudo_prs;                      //!< Parameters for the pseudo-ridge tracer
    float combine_clash_confidence;             //!< Parameter for the combination of the pseudo-ridge tracer and the PNN or MLP
    
    /***** Data: *****/
    float *protos_fvs, *tranmat;                //!< Dummy vector neccesary for PCASYS and the transformation matrix to be used in the PCA
    unsigned char *protos_classes;              //!< Dummy vector neccesary to load PCASYS
    
    FILE *fp_demo_images_list, *fp_out;         //!< Files used by PCASYS
    
   
    /* Processing */
    int sw, sh;                                 //!< Width and Height of the segmented image
    int sfgw, sfgh;                             //!< Width and Height of the segmented foregroun images (a mask to separate the foreground from the background)
    
    int wOr, hOr;                               //!< Width and Height of the averaged orientation maps (non-overlapped)
    int wOr2, hOr2;                             //!< Width and Height of the averaged orientation maps (overlapped)
    
    int wTf, hTf;                               //!< Width and Height of the transformation matrix for the PCA

};

 inline int PCASYSManager::getSW() const{ return sw;}

 inline int PCASYSManager::getSH() const{ return sh;}
 
 inline int PCASYSManager::getSfgW() const{ return sfgw;}

 inline int PCASYSManager::getSfgH() const{ return sfgh;}
 
 inline int PCASYSManager::getOrW() const{ return wOr;}

 inline int PCASYSManager::getOrH() const{ return hOr;}
 
 inline int PCASYSManager::getOrW2() const{ return wOr2;}

 inline int PCASYSManager::getOrH2() const{ return hOr2;}
 
 inline int PCASYSManager::getTfW() const{ return wTf;}

 inline int PCASYSManager::getTfH() const{ return hTf;}

#endif	/* PCASYSMANAGER_H */

