 /* 
 * \file        PseudoridgeTracer.h
 * \author      Carlos Lopez Molina (carlos.lopez@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * Header file of PseudoridgeTracer class
 *
 * 
 */

#ifndef PSEUDORIDGETRACER_H
#define	PSEUDORIDGETRACER_H

const int TURNINGRULES_HONG=1;
const int TURNINGRULES_ZHANG=2;

class PseudoridgeTracer {
public:
    
    /*
     * Default constructor
     */    
    PseudoridgeTracer();
    
    /**
     * Basic constructor. It uses an already computed orientation image (with its dimensions) adn
     *      the block size used for its calculation.
     * @param oriImage  Orientation image
     * @param orH       Number of rows of the orientation image.
     * @param orW       Number of cols of the orientation image.
     * @param blockSize Blocksize used in the computation of the orientation image.
     */
    PseudoridgeTracer(float **oriImage,int orH, int orW, int blockSize);
    
    /**
     * Destructor
     */
    virtual ~PseudoridgeTracer();
    
    //Main functions
    
    /**
     * Setting of the center from which the pseudoridge is to be traced
     * @param posI    Number of row
     * @param posJ    Number of cols
     */
    void setCenter(int posI, int posJ);
    
    /**
     * Main function for the calculation of the pseudoridge
     * @param numOfDirections   Number of discrete directions the circunference is divided into
     * @param maxPRidgeLength   Maximum lengh of the ridges
     * @param turningRules      Turning rules in {TURNINGRULES_HONG,TURNINGRULES_ZHANG}
     */
    void tracePseudoridges(int numOfDirections,int maxPRidgeLength, int turningRules);
    
    
    //Auxiliars
    
    /**
     * Evolves the position (rows) depending upon the direction
     * @param oldJ  Current position (rows)
     * @param dir   Current direction
     * @return      Next position (rows)
     */
    int evolveJ(int oldJ,int dir);
    
    /**
     * Evolves the position (cols) depending upon the direction
     * @param oldJ  Current position (cols)
     * @param dir   Current direction
     * @return      Next position (cols)
     */
    int evolveI(int oldI, int dir);
    
    /**
     * Internal function to create a discrete orientation map from the continuous one.
     * @param numOfOrientations   Number of discrete directions the circunference is divided into
     */
    void discretizeOrientations(int numOfOrientations);
    
    //Retrieving info
    
    /**
     * Gets the center (rows) from which the ridge was traced
     * @return The center (rows) from which the ridge was traced
     */
    int getCenterI();
    
    /**
     * Gets the center (cols) from which the ridge was traced
     * @return The center (cols) from which the ridge was traced
     */
    int getCenterJ();
    
    /**
     * Obtains an image containing 1's at the positions crossed by PRidges, 0 otherwise
     * @return An image containing 1's at the positions crossed by PRidges, 0 otherwise
     */
    int **getPseudoridge();
    
    /**
     * Retrieves the position (rows, cols) of the first end of the pseudoridge
     * @return The position (rows, cols) of the first end of the pseudoridge
     */
    int *getFirstEnd();
    
    /**
     * Retrieves the position (rows, cols) of the second end of the pseudoridge
     * @return The position (rows, cols) of the second end of the pseudoridge
     */
    int *getSecondEnd();
    
    /**
     * Checks whether the first Pridge ended up because of a turn condition
     * @return Whether the first Pridge ended up because of a turn condition 
     */
    bool firstPRidgeTurned();
    
    /**
     * Checks whether the second Pridge ended up because of a turn condition
     * @return Whether the second Pridge ended up because of a turn condition 
     */
    bool secondPRidgeTurned();
    
    /**
     * Checks whether the first Pridge ended up because it got out of the image limits
     * @return Whether the first Pridge ended up because it got out of the image limits
     */
    bool firstPRidgeExited();
    
    /**
     * Checks whether the second Pridge ended up because it got out of the image limits
     * @return Whether the second Pridge ended up because it got out of the image limits
     */
    bool secondPRidgeExited();
    
    /**
     * Sets a segmentation for the fingerprint
     * @param segmentation  The segmentation. True=valid position. Same dimensions as the orientations image.
     */
    void setBlockwiseSegmentation(bool **segmentation);
    
    /**
     * 
     * Returns the times the pseudoridge perfoms complete half-circumference turnss
     * @return The times the pseudoridge perfoms complete half-circumference turnss
     */
    int getNumberOfTurns();
    
    
private:
    
    float **oriImage;               //!< Orientation map
    int orW;                        //!< Width of the orientation map
    int orH;                        //!< Height of the orientation map             
    int centerI;                    //!< Vertical position of the center
    int centerJ;                    //!< Horizontal position of the center
    int blockSize;                  //!< Blocksize used in the computation of the orientation image.
    
    bool **blockwiseSegmentation;      //!< Blockwise segmentation. The pseudoridges stop if they reach such regions.
    
    int numDiscreteDirections;      //!< Number of discrete directions the circunference is divided into
    
    int **discreteOriImage;         //!< Discrete orientation map
    
    int **pseudoRImage;              //!< Pseudoridge binary map
    int *endOne;                    //!< Position of the first end of the pseudoridge
    int *endTwo;                    //!< Position of the second end of the pseudoridge
    
    bool oneTurned;                 //!< Whether the first PRidge ended up turning
    bool twoTurned;                 //!< Whether the second PRidge ended up turning
    bool oneExited;                 //!< Whether the first PRidge ended up out of the image limits
    bool twoExited;                 //!< Whether the second PRidge ended up out of the image limits
    float oneAccTurn;                 //!< Whether the first PRidge ended up turning
    float twoAccTurn;                 //!< Whether the second PRidge ended up turning
    
    
};

#endif	/* PSEUDORIDGETRACER_H */

