/* 
 * \file        SPAnglesNyongesa.cpp
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * The implementation of SPAnglesNyongesa (Nyongesa2004) class
 *
 * 
 */

#include "ImgUtil.h"
#include "MemUtil.hpp"
#include "FingerprintImg.h"

#include "SPAnglesNyongesa.h"
#include "NewPoincareSingularPoints.h"

#include "ParametersNyongesa.h"
#include "ParametersPoincare.h"



SPAnglesNyongesa::SPAnglesNyongesa() {
    img = 0;
    h = -1;
    w = -1;
    
    ySin = 0;
    xCos = 0;
    O = 0;
    
    binarizedImg = 0;
}



SPAnglesNyongesa::SPAnglesNyongesa(float **image, int h, int w) {
    
    img = MemUtil::copy2DMatrix<float>(image, h, w);
        
    this->h = h;
    this->w = w;
    
    ySin = 0;
    xCos = 0;
    O = 0;
    
    binarizedImg = 0;
}

SPAnglesNyongesa::~SPAnglesNyongesa() {
    
    MemUtil::freeMatrix(&img);    
    MemUtil::freeMatrix(&O);    
    MemUtil::freeMatrix(&xCos);
    MemUtil::freeMatrix(&ySin);
    MemUtil::freeMatrix(&binarizedImg);
    
}

float *SPAnglesNyongesa::run(string file, PCASYSManager *pcasys){
    FingerprintImg fi;

    if (ImgUtil::fAdjust != 1){  
        fi.setPCASYSManager(pcasys);     
        fi.readFingerprintImage(file);    
        fi.setSegmentedPCASYS(fi.getImage(), fi.getH(), fi.getW());
    }
    else{
        fi.setPCASYSManager(pcasys);
        fi.readFingerprintImage(file);
        fi.segment();
    }

    h = fi.getH();
    w = fi.getW();
    
    img = MemUtil::copy2DMatrix<float>(fi.getSegmentedImageFloat(), h, w);
    float *fv = computeFeatures();

    dimFV = 7;
    
    return fv;
    
}

float *SPAnglesNyongesa::computeFeatures() {
    
    binarizeImage(binarizationWNyongesa);// * ImgUtil::fAdjust); // Parameter    
    // createOrientationImage(5, 15); // block size = 5 x 5    * ImgUtil::fAdjust
    createOrientationImagePerPixel(blockWPixelOrientationNyongesa, blockWAverageOrientationNyongesa , smoothWOrientationNyongesa); // block size = 5 x 5
    
        
    NewPoincareSingularPoints *npsp = new NewPoincareSingularPoints(O, xCos, ySin, orH, orW);
    npsp->generateSegmentation(img, h, w);    
        
    npsp->computeCoresAndDeltas(curveWNyongesa, smoothWPoincareNyongesa, poincareModeNyongesa, rulesToBeAppliedNyongesa, minNBlocksNyongesa);
    
    int smoothSize = smoothWPoincareNyongesa;
    while( (npsp->getnCores() > 2) || (npsp->getnDeltas() > 2) ) { 
        int nSmooth = npsp->computeCoresAndDeltas(curveWNyongesa, smoothSize, poincareModeNyongesa, rulesToBeAppliedNyongesa, minNBlocksNyongesa);
        
        if (nSmooth >= maxSmoothPoincare)
            smoothSize+=2;
    }
    
    if (ImgUtil::DEBUG) {
        string fileName = "coresDeltasNyongesa";
        ImgUtil::writeSPImage(img, h, w, npsp->getDeltaCentroids(), npsp->getCoreCentroids(), this->blockW, fileName);
    }
    
    float overallO = computeOverallOrientation(16);
    
  //  cout << "overallO = " << overallO * 180. / PI << endl;
    
    vector<float> cores = npsp->getCoreCentroids();
    vector<float> deltas = npsp->getDeltaCentroids();

    float *featureVector = MemUtil::createVector<float>(7, -1);
    featureVector[6] = overallO;
    if ((npsp->getnCores() <= 2) && (npsp->getnDeltas() <= 2)) {
        float alfa = 0, beta = 0, gamma11 = 0, gamma21 = 0, gamma12 = 0, gamma22 = 0;
        
        float x1 = -1, y1 = -1, x2 = -1, y2 = -1;
        if (!cores.empty()) {
            y1 = cores[0];
            x1 = cores[1];
            if (npsp->getnCores() > 1) {
                y2 = cores[2];
                x2 = cores[3];
                alfa = 2. * PI - atan2(y2 - y1, x2 - x1);
            }
        }
        
        float p1 = -1, q1 = -1, p2 = -1, q2 = -1;
        if (!deltas.empty()) {
            q1 = deltas[0]; 
            p1 = deltas[1]; 
            if (npsp->getnDeltas() > 1) {
                q2 = deltas[2];
                p2 = deltas[3];
                beta =  2. * PI - atan2(q2 - q1, p2 - p1);
            }
        }
        
        if ( (x1 != -1) && (p1 != -1) )
            gamma11 =  2. * PI - atan2(q1 - y1, p1 - x1);
        if ( (x1 != -1) && (p2 != -1) )
            gamma12 =  2. * PI - atan2(q2 - y1, p2 - x1);
        if ( (x2 != -1) && (p1 != -1) )
            gamma21 =  2. * PI - atan2(q1 - y2, p1 - x2);
        if ( (x2 != -1) && (p2 != -1) )
            gamma22 =  2. * PI - atan2(q2 - y2, p2 - x2);
        
        featureVector[0] = alfa;
        featureVector[1] = beta;
        featureVector[2] = gamma11;
        featureVector[3] = gamma12;
        featureVector[4] = gamma21;
        featureVector[5] = gamma22;
    }
    if (ImgUtil::DEBUG) {
        for (int i = 0; i < 7; i++) 
            cout << featureVector[i] * 180. / PI << endl;
    }
    
    delete npsp;
    
    // reject si hay más de dos cores o deltas! todo va a -1
    return featureVector;
}

void SPAnglesNyongesa::binarizeImage(int W) {
    
    binarizedImg = MemUtil::create2DMatrix<bool>(h, w);
    
    int halfW = W / 2;
    
    for (int i = 0; i < h; i += W) {
        for (int j = 0; j < w; j += halfW) {
            // Compute mean
            float mean = 0;
            int N = 0;
            for (int u = i; u < min(i + W, h); u++) {
                for (int v = j; v < min(j + W, w); v++) {
                    mean += img[u][v];                    
                    N++;
                }
            }
            mean /= (float)N;
            
            for (int u = i; u < min(i + W, h); u++) 
                for (int v = j; v < min(j + halfW, w); v++) 
                    binarizedImg[u][v] = img[u][v] > mean ? 1 : 0;
        }
    }
    
    if (ImgUtil::DEBUG) {
        string fileName2 = "binarizedImg";
        ImgUtil::writeGrayScaleImage(binarizedImg, h, w, fileName2);
    }
    
}

void SPAnglesNyongesa::createOrientationImage(int blockW, int Wsmooth) {
    
    this->blockW = blockW;
    orH = h / blockW;
    orW = w / blockW;
    
    O = MemUtil::create2DMatrix<float>(orH,orW);
    xCos = MemUtil::create2DMatrix<float>(orH,orW);
    ySin = MemUtil::create2DMatrix<float>(orH,orW);

    int masc[blockW][blockW];
    int halfBlock = blockW / 2;
    for (int i = 0; i < blockW; i++) {
        for (int j = 0; j < blockW; j++) {
            masc[i][j] = -1;
            if (i == halfBlock)
                masc[i][j] = 0;
            else if (j == halfBlock)
                masc[i][j] = 2;
            else if (i == j)
                masc[i][j] = 3;
            else if (i + j == blockW - 1)
                masc[i][j] = 1;
        }
    }
    masc[halfBlock][halfBlock] = -1; // comodín
    
    /* Compute Orientations */
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {
            
            float Pm[4];
            for (int k = 0; k < 4; k++)
                Pm[k] = binarizedImg[i + halfBlock][j + halfBlock];
            
            for (int u = i * blockW; u < (i + 1) * blockW; u++) {
                for (int v = j * blockW; v < (j + 1) * blockW; v++) {
                    int valMasc = masc[u % blockW][v % blockW];
                    if (valMasc != -1) {
                            Pm[valMasc] += binarizedImg[u][v];                        
                    }                    
                }
            }
            for (int k = 0; k < 4; k++) 
                Pm[k] /= (float) blockW;
            
            float D[4];
            for (int k = 0; k < 4; k++)
                D[k] = abs(Pm[k] - binarizedImg[i + halfBlock][j + halfBlock]);
            for (int u = i * blockW; u < (i + 1) * blockW; u++) {
                for (int v = j * blockW; v < (j + 1) * blockW; v++) {
                    int valMasc = masc[u % blockW][v % blockW];
                    if (valMasc != -1) {
                            D[valMasc] += abs(Pm[valMasc] - binarizedImg[u][v]);                        
                    }                    
                }
            }
            
            int ang = 0;
            float minD = D[0];
            bool iguales = false;
            for (int k = 1; k < 4; k++) {
                if (D[k] < minD) {
                    ang = k;
                    minD = D[k];
                    iguales = false;
                }
                else if (D[k] == minD)
                    iguales = true;
            }
            
            O[i][j] = PI / 4. * (float)ang;
            if (O[i][j] > PI / 2.)
                O[i][j] -= PI;
            
            xCos[i][j] = cos(2.0 * O[i][j]);
            ySin[i][j] = sin(2.0 * O[i][j]);
            
            if (iguales) {
                O[i][j] = -1;
                xCos[i][j] = 0;
                ySin[i][j] = 0;
            }                
        }
    }
    
    if (ImgUtil::DEBUG) {
        string fileName2; 
        fileName2 = "OrientationSinCos";
        ImgUtil::writeOrImg(xCos, ySin, orW, orH, fileName2);
    }
    
    ImgUtil::smoothOrientations(O, &xCos, &ySin, orH, orW, Wsmooth);
    
    
    if (ImgUtil::DEBUG) {
        string fileName2; 
        fileName2 = "OrientationSinCosSmooth";
        ImgUtil::writeOrImg(xCos, ySin, orW, orH, fileName2);
    }
    
    
}



void SPAnglesNyongesa::createOrientationImagePerPixel(int blockW, int blockW2, int Wsmooth) {
    
    this->blockW = blockW2;
    orH = h / blockW2;
    orW = w / blockW2;
    
    O = MemUtil::create2DMatrix<float>(orH,orW);
    xCos = MemUtil::create2DMatrix<float>(orH,orW);
    ySin = MemUtil::create2DMatrix<float>(orH,orW);
    
    
    float **pixelO = MemUtil::create2DMatrix<float>(h, w);

    int masc[blockW][blockW];
    int halfBlock = blockW / 2;
    for (int i = 0; i < blockW; i++) {
        for (int j = 0; j < blockW; j++) {
            masc[i][j] = -1;
            if ((i == halfBlock) && (j == halfBlock))
                masc[i][j] = -1;
            if (i == halfBlock)
                masc[i][j] = 0;
            else if (j == halfBlock)
                masc[i][j] = 2;
            else if (i == j)
                masc[i][j] = 3;
            else if (i + j == blockW - 1)
                masc[i][j] = 1;
        }
    }
    
    
    
    /* Compute Orientations */
    float Pm[4];
    float D[4];
    int ang = 0;
    float minD = 1000;
    bool iguales = false;
    for (int i = halfBlock; i < h - halfBlock; i++) {
        for (int j = halfBlock; j < w - halfBlock; j++) {
            
            for (int k = 0; k < 4; k++)
                Pm[k] = binarizedImg[i][j];
            
            for (int u = i - halfBlock; u < i + halfBlock + 1; u++) {
                for (int v = j - halfBlock; v < j + halfBlock + 1; v++) {
                    int valMasc = masc[(u - i + halfBlock)][(v - j + halfBlock)];
                    if (valMasc != -1) {
                            Pm[valMasc] += binarizedImg[u][v];                        
                    }                    
                }
            }
            for (int k = 0; k < 4; k++) {
                Pm[k] /= (float) blockW;
            }
            
            for (int k = 0; k < 4; k++)
                D[k] = abs(Pm[k] - binarizedImg[i][j]);
            for (int u = i - halfBlock; u < i + halfBlock + 1; u++) {
                for (int v = j - halfBlock; v < j + halfBlock + 1; v++) {
                    int valMasc = masc[(u - i + halfBlock)][(v - j + halfBlock)];
                    if (valMasc != -1) {
                            D[valMasc] += abs(Pm[valMasc] - binarizedImg[u][v]);                        
                    }                    
                }
            }
            
            ang = 0;
            minD = D[0];
            iguales = false;
            for (int k = 1; k < 4; k++) {
                if (D[k] < minD) {
                    ang = k;
                    minD = D[k];
                    iguales = false;
                }
                else if (D[k] == minD)
                    iguales = true;
            }
            
            pixelO[i][j] = PI / 4. * (float)ang;
            if (pixelO[i][j] > PI / 2.)
                pixelO[i][j] -= PI;
            
            if (iguales) {
                pixelO[i][j] = -1;
            }
        }
    }
    
    /* Compute Orientations */
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {
            
            for (int u = i * blockW2; u < (i + 1) * blockW2; u++) {
                for (int v = j * blockW2; v < (j + 1) * blockW2; v++) {
                    if (pixelO[u][v] != -1) {
                        xCos[i][j] += cos(2. * pixelO[u][v]);
                        ySin[i][j] += sin(2. * pixelO[u][v]);
                    }
                }
            }
            xCos[i][j] /= (float)(blockW2 * blockW2);
            ySin[i][j] /= (float)(blockW2 * blockW2);
                    
            O[i][j] = 0.5 * atan2(ySin[i][j], xCos[i][j]);
            
            if (O[i][j] > PI / 2.)
                O[i][j] -= PI;
            
    //        xCos[i][j] = cos(2.0 * O[i][j]);
  //          ySin[i][j] = sin(2.0 * O[i][j]);
            
        }
    }
    
    if (ImgUtil::DEBUG) {
        string fileName2; 
        fileName2 = "OrientationSinCosPerPixel";
        ImgUtil::writeOrImg(xCos, ySin, orW, orH, fileName2);
    }
    
    ImgUtil::smoothOrientations(O, &xCos, &ySin, orH, orW, Wsmooth, true);
    
    if (ImgUtil::DEBUG) {
        string fileName2; 
        fileName2 = "OrientationSinCosSmoothPerPixel";
        ImgUtil::writeOrImg(xCos, ySin, orW, orH, fileName2);
    }
    
    MemUtil::freeMatrix(&pixelO);
    
}

float SPAnglesNyongesa::computeOverallOrientation(int overallW) {
    
    int i = orH / 2;
    int j = orW / 2;
    
    int halfW = (overallW - 1) / 2;
    int halfW2 = halfW;
    if (overallW % 2 == 0)
        halfW2++;
    
    float overallXCos = 0;
    float overallYSin = 0;
    float overallO = 0;
    
    for (int u = i - halfW; u < i + halfW2 + 1; u++) {
        for (int v = j - halfW; v < j + halfW2 + 1; v++) {
            overallXCos += xCos[u][v];
            overallYSin += ySin[u][v];
        }
    }
    overallXCos /= (float) (overallW * overallW);
    overallYSin /= (float) (overallW * overallW);
    
    overallO = 0.5 * atan2(overallYSin, overallXCos);
    
    return overallO;
}